
/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include "../gb.h"

/* Include the 21x21 icon pixmap for this widget, to be used in the palette */
#include "../graphics/button.xpm"

/*
 * This is the GbWidget struct for this widget (see ../gbwidget.h).
 * It is initialized in the init() function at the end of this file
 */
static GbWidget gbwidget;

static gchar *Label = "GtkButton::label";
/* This is only used for toolbar buttons. */
static gchar *Icon = "GtkButton::icon";


/******
 * NOTE: To use these functions you need to uncomment them AND add a pointer
 * to the function in the GbWidget struct at the end of this file.
 ******/

/*
 * Creates a new GtkWidget of class GtkButton, performing any specialized
 * initialization needed for the widget to work correctly in this environment.
 * If a dialog box is used to initialize the widget, return NULL from this
 * function, and call data->callback with your new widget when it is done.
 * If the widget needs a special destroy handler, add a signal here.
 */
GtkWidget *
gb_button_new (GbWidgetNewData * data)
{
  GtkWidget *new_widget;

  /* SPECIAL CODE: to handle toolbar buttons. */
  MSG1 ("IN gb_button_new parent:%s",
	data->parent ? gtk_widget_get_name (data->parent) : "NULL");
  if (data->parent && GTK_IS_TOOLBAR (data->parent))
    {
      MSG ("   creating toolbar button...");
      new_widget = gb_widget_new_toolbar_button (data, GTK_TOOLBAR_CHILD_BUTTON);
      if (new_widget)
	return new_widget;
    }

  if (data->action == GB_CREATING)
    new_widget = gtk_button_new_with_label (data->name);
  else
    {
      MSG ("   creating normal button with placeholder...");
      new_widget = gtk_button_new ();
      gtk_container_add (GTK_CONTAINER (new_widget), editor_new_placeholder());
    }
  return new_widget;
}


void
gb_button_add_child(GtkWidget *widget, GtkWidget * child, GbWidgetSetArgData *data)
{
  if (GTK_BUTTON (widget)->child)
    gtk_container_remove (GTK_CONTAINER (widget),
                          GTK_BUTTON (widget)->child);
  gtk_container_add (GTK_CONTAINER (widget), child);
}


/*
 * Creates the components needed to edit the extra properties of this widget.
 */
static void
gb_button_create_properties (GtkWidget * widget, GtkWidget * child, GbWidgetCreateArgData * data)
{
  property_add_text (Label, _("Label:"), _("The text to display"), 2);
  property_add_filename (Icon, _("Icon:"), _("The pixmap filename"));
}



/*
 * Gets the properties of the widget. This is used for both displaying the
 * properties in the property editor, and also for saving the properties.
 */
static void
gb_button_get_properties (GtkWidget * widget, GtkWidget * child, GbWidgetGetArgData * data)
{
  gb_widget_output_child_label (widget, data, Label);
  gb_widget_output_child_icon (widget, data, Icon);
}



/*
 * Sets the properties of the widget. This is used for both applying the
 * properties changed in the property editor, and also for loading.
 */
static void
gb_button_set_properties (GtkWidget * widget, GtkWidget * child, GbWidgetSetArgData * data)
{
  gb_widget_input_child_label (widget, data, Label);
  gb_widget_input_child_icon (widget, data, Icon);
}



/*
 * Adds menu items to a context menu which is just about to appear!
 * Add commands to aid in editing a GtkButton, with signals pointing to
 * other functions in this file.
 */
static void
gb_button_create_popup_menu (GtkWidget * widget, GbWidgetCreateMenuData * data)
{
  /* Add command to remove child label. */
  gb_widget_create_child_label_popup_menu (widget, data);

  /* Add commands for toolbar buttons. */
  gb_widget_create_toolbar_button_popup_menu (widget, data);
}



/*
 * Writes the source code needed to create this widget.
 * You have to output everything necessary to create the widget here, though
 * there are some convenience functions to help.
 */
static void
gb_button_write_source (GtkWidget * widget, GbWidgetWriteSourceData * data)
{
  GtkWidget *child = GTK_BUTTON (widget)->child;
  gchar *label_text;

  if (data->create_widget)
    {
      if (gb_widget_is_toolbar_button (widget))
	{
	  gb_widget_write_toolbar_button_source (widget, data);
	}
      else if (child && GTK_IS_LABEL (child) && !GB_IS_GB_WIDGET (child))
	{
	  gtk_label_get (GTK_LABEL (child), &label_text);
	  source_add (data, "  %s = gtk_button_new_with_label (%s);\n",
		      data->wname,
		      source_make_string (label_text, data->use_gettext));
	}
      else
	{
	  source_add (data, "  %s = gtk_button_new ();\n", data->wname);
	}
    }

  gb_widget_write_standard_source (widget, data);
}


static void
gb_button_destroy (GtkWidget * widget, GbWidgetDestroyData * data)
{
  gb_widget_destroy_child_icon (widget, data);
}


/*
 * Initializes the GbWidget structure.
 * I've placed this at the end of the file so we don't have to include
 * declarations of all the functions.
 */
GbWidget *
gb_button_init ()
{
  /* Initialise the GTK type */
  gtk_button_get_type ();

  /* Initialize the GbWidget structure */
  gb_widget_init_struct (&gbwidget);

  /* Fill in the pixmap struct & tooltip */
  gbwidget.pixmap_struct = button_xpm;
  gbwidget.tooltip = _("Button");

  /* Fill in any functions that this GbWidget has */
  gbwidget.gb_widget_new = gb_button_new;
  gbwidget.gb_widget_add_child = gb_button_add_child;
  gbwidget.gb_widget_create_properties = gb_button_create_properties;
  gbwidget.gb_widget_get_properties = gb_button_get_properties;
  gbwidget.gb_widget_set_properties = gb_button_set_properties;
  gbwidget.gb_widget_create_popup_menu = gb_button_create_popup_menu;
  gbwidget.gb_widget_write_source = gb_button_write_source;
  gbwidget.gb_widget_destroy = gb_button_destroy;

  return &gbwidget;
}
