/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998-1999  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <locale.h>
#include <stdlib.h>

#include <gtk/gtk.h>

#include "gladeconfig.h"

#include "editor.h"
#include "glade_clipboard.h"
#include "gbwidget.h"
#include "load.h"
#include "tree.h"
#include "utils.h"

typedef struct _GladeClipboardItem  GladeClipboardItem;

struct _GladeClipboardItem
{
  GladeProject *project;
  GtkType type;
  gchar *xml_data;
  gboolean names_unique;
};


static GtkWindowClass *parent_class = NULL;

static void glade_clipboard_class_init (GladeClipboardClass * klass);
static void glade_clipboard_init (GladeClipboard * clipboard);

static void glade_clipboard_cut_or_copy (GladeClipboard *clipboard,
					 GladeProject   *project,
					 GtkWidget      *widget,
					 gboolean	 cut);
static void glade_clipboard_add	(GladeClipboard *clipboard,
				 GladeProject   *project,
				 gboolean	 names_unique,
				 GtkWidget	*widget,
				 gchar		*xml_data);
static GladeClipboardItem* glade_clipboard_get_current_item (GladeClipboard *clipboard);


#define GLADE_CUT_BUFFER_INCREMENT	1024
#define GLADE_PASTE_BUFFER_INCREMENT	1024


guint
glade_clipboard_get_type (void)
{
  static guint glade_clipboard_type = 0;

  if (!glade_clipboard_type)
    {
      GtkTypeInfo glade_clipboard_info =
      {
	"GladeClipboard",
	sizeof (GladeClipboard),
	sizeof (GladeClipboardClass),
	(GtkClassInitFunc) glade_clipboard_class_init,
	(GtkObjectInitFunc) glade_clipboard_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
#ifdef GTK_HAVE_FEATURES_1_1_0
	(GtkClassInitFunc) NULL,
#endif
      };

      glade_clipboard_type = gtk_type_unique (gtk_window_get_type (),
					      &glade_clipboard_info);
    }
  return glade_clipboard_type;
}


static void
glade_clipboard_class_init (GladeClipboardClass * klass)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass *) klass;

  parent_class = gtk_type_class (gtk_window_get_type ());
}


static void
glade_clipboard_init (GladeClipboard * clipboard)
{
  GtkWidget *vbox;

  gtk_window_set_title (GTK_WINDOW (clipboard), _ ("Clipboard"));
  gtk_window_set_policy (GTK_WINDOW (clipboard), FALSE, TRUE, FALSE);

  vbox = gtk_vbox_new (FALSE, 4);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (clipboard), vbox);

  clipboard->clist = gtk_clist_new (1);
  gtk_clist_set_row_height (GTK_CLIST (clipboard->clist), 23);
  gtk_widget_set_usize (clipboard->clist, 100, 100);
  gtk_clist_set_column_width (GTK_CLIST (clipboard->clist), 0, 100);
  gtk_clist_set_selection_mode (GTK_CLIST (clipboard->clist),
				GTK_SELECTION_BROWSE);
  gtk_widget_show (clipboard->clist);

  gb_box_clist_auto_policy (vbox, clipboard->clist);

  gtk_signal_connect (GTK_OBJECT (clipboard), "delete_event",
		      GTK_SIGNAL_FUNC (glade_util_close_window_on_delete),
		      NULL);
}


GtkWidget *
glade_clipboard_new (void)
{
  return GTK_WIDGET (gtk_type_new (glade_clipboard_get_type ()));
}


void
glade_clipboard_cut (GladeClipboard *clipboard,
		     GladeProject   *project,
		     GtkWidget      *widget)
{
  glade_clipboard_cut_or_copy (clipboard, project, widget, TRUE);
}


void
glade_clipboard_copy (GladeClipboard *clipboard,
		      GladeProject   *project,
		      GtkWidget      *widget)
{
  glade_clipboard_cut_or_copy (clipboard, project, widget, FALSE);
}


static void
glade_clipboard_cut_or_copy (GladeClipboard *clipboard,
			     GladeProject   *project,
			     GtkWidget      *widget,
			     gboolean	     cut)
{
  GbWidgetGetArgData data;
  gchar *old_locale, *saved_locale;
  GList *selection;

  if (widget == NULL)
    {
      selection = editor_get_selection ();
      if (selection)
	widget = GTK_WIDGET (selection->data);
    }

  if (widget == NULL || GB_IS_PLACEHOLDER (widget))
    return;

  data.project = project;
  data.status = GLADE_STATUS_OK;
  data.filename = NULL;
  data.buffer.pos = 0;
  data.indent = 0;
  gb_init_buffer_struct (0, GLADE_CUT_BUFFER_INCREMENT, &data.buffer);

  old_locale = setlocale (LC_NUMERIC, NULL);
  saved_locale = g_strdup (old_locale);
  setlocale (LC_NUMERIC, "C");

  gb_widget_save (widget, &data);

  setlocale (LC_NUMERIC, saved_locale);
  g_free (saved_locale);

  if (data.status == GLADE_STATUS_OK)
    {
      glade_clipboard_add (GLADE_CLIPBOARD (clipboard), project, cut,
			   widget, data.buffer.ptr);
      if (cut)
	editor_delete_widget (widget);
    }
  else
    g_warning ("Error saving widget to clipboard");

  g_free (data.buffer.ptr);
}


void
glade_clipboard_paste (GladeClipboard *clipboard,
		       GladeProject   *project,
		       GtkWidget      *widget)
{
  GladeClipboardItem *item;
  GbWidgetSetArgData data;
  GtkWidget *parent, *new_widget = NULL;
  gchar *old_locale, *saved_locale, *saved_timezone;
  GList *selection;

  item = glade_clipboard_get_current_item (clipboard);
  if (item == NULL)
    return;

  if (widget == NULL)
    {
      selection = editor_get_selection ();
      if (selection)
	widget = GTK_WIDGET (selection->data);
    }

  /* We can only paste toplevel components, i.e. windows and menus, into the
     project. */
  if (widget == NULL)
    {
      if (!gtk_type_is_a (item->type, gtk_window_get_type ())
	  && !gtk_type_is_a (item->type, gtk_menu_get_type ()))
	return;
    }

  /* SPECIAL CODE: Don't allow pasting into dialog widgets. */
  if (widget)
    {
      gchar *child_name = gtk_object_get_data (GTK_OBJECT (widget),
					       GB_CHILD_NAME_KEY);
      if (child_name)
	{
	  if (!strcmp (child_name, "FileSel:ok_button")
	      || !strcmp (child_name, "FileSel:cancel_button")
	      || !strcmp (child_name, "ColorSel:ok_button")
	      || !strcmp (child_name, "ColorSel:cancel_button")
	      || !strcmp (child_name, "ColorSel:help_button")
	      || !strcmp (child_name, "FontSel:ok_button")
	      || !strcmp (child_name, "FontSel:cancel_button")
	      || !strcmp (child_name, "FontSel:apply_button")
	      || !strcmp (child_name, "Dialog:vbox")
	      || !strcmp (child_name, "Dialog:action_area")
	      || !strcmp (child_name, "InputDialog:save_button")
	      || !strcmp (child_name, "InputDialog:close_button")
	      )
	    {
	      /*g_print ("Not allowing paste into dialog widget\n");*/
	      return;
	    }
	}
    }

  data.project = project;
  data.fp = NULL;
  data.filename = NULL;
  data.xml_buffer = item->xml_data;
  data.xml_buffer_pos = 0;
  data.xml_buffer_len = strlen (data.xml_buffer);

  /* We normally create new names for the widgets pasted, discarding the names
     in the XML data. However, if the widgets were cut to the clipboard, and
     we are pasting into the same project for the first time, then we use
     the original names. */
  data.discard_names = TRUE;
  if (project == item->project)
    {
      if (item->names_unique)
	{
	  data.discard_names = FALSE;
	  item->names_unique = FALSE;
	}
    }

  if (widget && (!GTK_IS_CONTAINER (widget)
		 || GTK_IS_BIN (widget)
		 || GTK_IS_BUTTON (widget)))
    {
      parent = widget->parent;
      data.replacing_widget = widget;
    }
  else
    {
      parent = widget;
      data.replacing_widget = NULL;
    }

  load_init_before_read (GLADE_PASTE_BUFFER_INCREMENT, &data);

  old_locale = setlocale (LC_NUMERIC, NULL);
  saved_locale = g_strdup (old_locale);
  setlocale (LC_NUMERIC, "C");

#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_freeze ();
#endif

  /* Save the current time zone environment variable. */
  saved_timezone = g_strdup (getenv ("TZ"));
  if (putenv ("TZ=GMT") == -1)
    data.status = GLADE_STATUS_ERROR;
  else
    {
      load_token_skip_whitespace (&data);
      if (data.status == GLADE_STATUS_OK)
	{
	  if (!strcmp (data.buffer.ptr + data.token, "widget"))
	    new_widget = gb_widget_load (NULL, &data, parent);
	  else
	    data.status = GLADE_STATUS_INVALID_ENTITY;
	}
    }

#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_thaw ();
#endif

  g_free (data.buffer.ptr);
  gb_free_load_properties (&data.properties);
  gb_free_load_properties (&data.child_properties);

  /* FIXME: do this. */
  /*load_validate_widget_names (XXXX);*/

  /* Now reset the timezone. */
  if (saved_timezone)
    {
      gchar *new_timezone = g_malloc (strlen (saved_timezone) + 5);
      strcpy (new_timezone, "TZ=");
      strcat (new_timezone, saved_timezone);
      putenv (new_timezone);
      g_free (new_timezone);
      g_free (saved_timezone);
    }
  else
    putenv ("TZ=");

  setlocale (LC_NUMERIC, saved_locale);
  g_free (saved_locale);

  if (data.status != GLADE_STATUS_OK)
    g_warning ("Error pasting from clipboard");
}


/* This adds the XML for a widget (and any descendants) to the clipboard.
   If the widget has been cut and is later pasted into the same project,
   then the same widget names can be used. Otherwise new widget names have
   to be created. The widget parameter is used to get the widget's name and
   class and to look up its pixmap to display in the clipboard.
   The xml_data parameter is copied. */
static void
glade_clipboard_add	(GladeClipboard *clipboard,
			 GladeProject   *project,
			 gboolean	 names_unique,
			 GtkWidget	*widget,
			 gchar		*xml_data)
{
  GladeClipboardItem *item;
  GbWidget *gbwidget;
  gchar *class_name, *name;

  class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
  name = gtk_widget_get_name (widget);

  gbwidget = g_hash_table_lookup (gb_widget_table, class_name);
  g_return_if_fail (gbwidget != NULL);

  item = g_new (GladeClipboardItem, 1);
  item->project = project;
  item->type = GTK_OBJECT_TYPE (widget);
  item->xml_data = g_strdup (xml_data);
  item->names_unique = names_unique;

  gtk_clist_insert (GTK_CLIST (clipboard->clist), 0, &name);
  gtk_clist_set_row_data (GTK_CLIST (clipboard->clist), 0, item);

  gtk_clist_set_pixtext (GTK_CLIST (clipboard->clist), 0, 0, name, 3,
			 gbwidget->gdkpixmap, gbwidget->mask);

  gtk_clist_select_row (GTK_CLIST (clipboard->clist), 0, 0);
}


/* This returns the currently-selected GladeClipboardItem, or NULL if no item
   is currently selected (i.e. the clipboard is empty). */
static GladeClipboardItem*
glade_clipboard_get_current_item	(GladeClipboard *clipboard)
{
  GList *selection;

  selection = GTK_CLIST (clipboard->clist)->selection;
  if (selection == NULL)
    return NULL;

  return (GladeClipboardItem*) gtk_clist_get_row_data (GTK_CLIST (clipboard->clist), GPOINTER_TO_INT (selection->data));
}
