/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998-1999  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <gtk/gtk.h>

#include "gladeconfig.h"

#include "editor.h"
#include "glade_menu_editor.h"
#include "glade_project.h"
#include "load.h"
#include "property.h"
#include "save.h"
#include "source.h"
#include "tree.h"
#include "utils.h"

/* FIXME: This is the current project. We only support one open at present. */
GladeProject *current_project = NULL;

/* This is the number of new projects created. The project name defaults to
   "project<project_num>". */
static gint project_num = 1;

gchar *GladeLanguages[] = { "C", "C++", "Ada 95" };
gint GladeNumLanguages = sizeof (GladeLanguages) / sizeof (GladeLanguages[0]);

#ifndef GTK_HAVE_FEATURES_1_1_0
typedef void (*GladeProjectSignal1) (GtkObject * object,
				     GtkObject * arg1,
				     gpointer data);

static void
glade_project_marshal_signal_1 (GtkObject * object,
				GtkSignalFunc func,
				gpointer func_data,
				GtkArg * args);
#endif

static void glade_project_class_init (GladeProjectClass * klass);
static void glade_project_init (GladeProject * dialog);
static void glade_project_destroy (GtkObject *object);

static GladeStatusCode glade_project_write_c_source (GladeProject *project);
static GladeStatusCode glade_project_write_cpp_source (GladeProject *project);
static GladeStatusCode glade_project_write_ada95_source (GladeProject *project);

static void free_key (gchar	*key,
		      gchar	*value,
		      gpointer	 data);

static void on_fake_window_destroy (GtkWidget * fake_window,
				    GtkWidget * component);
static void on_menu_editor_ok (GtkWidget *button,
			       GtkWidget *component);
static void on_menu_editor_apply (GtkWidget *button,
				  GtkWidget *component);
static gint on_menu_editor_close (GtkWidget *menued,
				  gpointer data);

static void glade_project_real_ensure_widgets_named (GtkWidget    *widget,
						     GladeProject *project);

enum
{
  ADD_COMPONENT,
  REMOVE_COMPONENT,
  COMPONENT_CHANGED,
  LAST_SIGNAL
};

static guint glade_project_signals[LAST_SIGNAL] = {0};

static GtkObjectClass *parent_class = NULL;


guint
glade_project_get_type (void)
{
  static guint glade_project_type = 0;

  if (!glade_project_type)
    {
      GtkTypeInfo glade_project_info =
      {
	"GladeProject",
	sizeof (GladeProject),
	sizeof (GladeProjectClass),
	(GtkClassInitFunc) glade_project_class_init,
	(GtkObjectInitFunc) glade_project_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
#ifdef GTK_HAVE_FEATURES_1_1_0
	(GtkClassInitFunc) NULL,
#endif
      };

      glade_project_type = gtk_type_unique (gtk_object_get_type (),
					    &glade_project_info);
    }

  return glade_project_type;
}

static void
glade_project_class_init (GladeProjectClass * klass)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass *) klass;

  parent_class = gtk_type_class (gtk_object_get_type ());

  glade_project_signals[ADD_COMPONENT] =
    gtk_signal_new ("add_component",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GladeProjectClass, add_component),
#ifndef GTK_HAVE_FEATURES_1_1_0
		    glade_project_marshal_signal_1,
#else
		    gtk_marshal_NONE__POINTER,
#endif
		    GTK_TYPE_NONE, 1, GTK_TYPE_OBJECT);
  glade_project_signals[REMOVE_COMPONENT] =
    gtk_signal_new ("remove_component",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GladeProjectClass, remove_component),
#ifndef GTK_HAVE_FEATURES_1_1_0
		    glade_project_marshal_signal_1,
#else
		    gtk_marshal_NONE__POINTER,
#endif
		    GTK_TYPE_NONE, 1, GTK_TYPE_OBJECT);
  glade_project_signals[COMPONENT_CHANGED] =
    gtk_signal_new ("component_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GladeProjectClass, component_changed),
#ifndef GTK_HAVE_FEATURES_1_1_0
		    glade_project_marshal_signal_1,
#else
		    gtk_marshal_NONE__POINTER,
#endif
		    GTK_TYPE_NONE, 1, GTK_TYPE_OBJECT);
  gtk_object_class_add_signals (object_class, glade_project_signals,
				LAST_SIGNAL);

  klass->add_component = NULL;
  klass->remove_component = NULL;
  klass->component_changed = NULL;

  object_class->destroy = glade_project_destroy;
}


static void
glade_project_init (GladeProject * project)
{
  gchar project_name[16];

  project->xml_filename = NULL;

  /* The default project name, used when new projects are created. */
  sprintf (project_name, _("project%i"), project_num++);
  project->name = g_strdup (project_name);
  project->directory = NULL;
  project->source_directory = NULL;
  project->pixmaps_directory = NULL;

  project->language = GLADE_LANGUAGE_C;
  property_show_lang_specific_page (project->language);

  project->changed = FALSE;

  project->unique_id_hash = g_hash_table_new (g_str_hash, g_str_equal);

  project->components = NULL;

  project->associated_windows = NULL;

  project->pixmap_filenames = NULL;

  project->current_directory = NULL;

  project->gettext_support = FALSE;
  project->use_widget_names = FALSE;
  project->main_source_file = g_strdup ("gladesrc.c");
  project->main_header_file = g_strdup ("gladesrc.h");
  project->handler_source_file = g_strdup ("gladesig.c");
  project->handler_header_file = g_strdup ("gladesig.h");
}


GladeProject*
glade_project_new (void)
{
  /* FIXME: Currently we only support one project open, so we have to destroy
     any existing project, and reset everything. */
  if (current_project)
    {
      property_set_widget (NULL);
      editor_clear_selection (NULL);
      gtk_object_destroy (GTK_OBJECT (current_project));
      /* Delete all GbStyles and reset hash table. Do this after all widgets
	 are destroyed. */
      gb_widget_reset_gb_styles ();
    }

  current_project = GLADE_PROJECT (gtk_type_new (glade_project_get_type()));
  return current_project;
}


static void
glade_project_destroy (GtkObject *object)
{
  GladeProject *project;
  GList *tmp_list;

  project = GLADE_PROJECT (object);

  g_free (project->xml_filename);

  g_free (project->name);
  g_free (project->directory);
  g_free (project->source_directory);
  g_free (project->pixmaps_directory);

#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_clear ();
#endif

  /* Destroy all project components. */
  tmp_list = project->components;
  while (tmp_list)
    {
      gtk_widget_destroy (GTK_WIDGET (tmp_list->data));
      tmp_list = tmp_list->next;
    }
  g_list_free (project->components);
 
  /* Destroy the unique id hash. */
  g_hash_table_foreach (project->unique_id_hash, (GHFunc) free_key, NULL);
  g_hash_table_destroy (project->unique_id_hash);

  /* Destroy all windows associated with the project. */
  tmp_list = project->associated_windows;
  while (tmp_list)
    {
      gtk_widget_destroy (GTK_WIDGET (tmp_list->data));
      tmp_list = tmp_list->next;
    }
  g_list_free (project->associated_windows);

  /* Free all pixmap filenames. */
  tmp_list = project->pixmap_filenames;
  while (tmp_list)
    {
      g_free (tmp_list->data);
      tmp_list = tmp_list->next;
    }
  g_list_free (project->pixmap_filenames);

  g_free (project->current_directory);

  g_free (project->main_source_file);
  g_free (project->main_header_file);
  g_free (project->handler_source_file);
  g_free (project->handler_header_file);
}


/* This is called for each id in the unique id hash. We have to free the
   id strings, since they were g_strdup'ed. */
static void
free_key (gchar		*key,
	  gchar		*value,
	  gpointer	 data)
{
  g_free (key);
}


GladeStatusCode
glade_project_open	(gchar	       *xml_filename,
			 GladeProject **project_return,
			 GList	      **errors)
{
  GladeStatusCode status;
  GladeProject *project;

  project = glade_project_new ();
  project->xml_filename = g_strdup (xml_filename);

#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_freeze ();
#endif

  status = load_project_file (project, errors);
  
#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_thaw ();
#endif

  if (status != GLADE_STATUS_OK)
    {
      gtk_object_destroy (GTK_OBJECT (project));
      current_project = NULL;
    }
  else
    {
      *project_return = project;
    }
  return status;
}


GladeStatusCode
glade_project_save (GladeProject *project)
{
  GladeStatusCode status;

  /* Copy any pixmaps to the project's pixmaps directory. */
  glade_project_copy_all_pixmaps (project);

  status = save_project_file (project);
  if (status == GLADE_STATUS_OK)
    glade_project_set_changed (project, FALSE);
  return status;
}


GladeStatusCode
glade_project_write_source (GladeProject *project)
{
  /* First we check that we have a source directory and a pixmaps directory. */
  if (project->source_directory == NULL || project->pixmaps_directory == NULL)
    return GLADE_STATUS_ERROR;

  /* Copy any pixmaps to the project's pixmaps directory. */
  glade_project_copy_all_pixmaps (project);

  /* We call a function according to the project's language. */
  switch (project->language)
    {
    case GLADE_LANGUAGE_C:
      return glade_project_write_c_source (project);
    case GLADE_LANGUAGE_CPP:
      return glade_project_write_cpp_source (project);
    case GLADE_LANGUAGE_ADA95:
      return glade_project_write_ada95_source (project);
    }

  /* Shouldn't get here. */
  return GLADE_STATUS_ERROR;
}


GladeStatusCode
glade_project_write_c_source (GladeProject *project)
{
  return source_write (project);
}


GladeStatusCode
glade_project_write_cpp_source (GladeProject *project)
{
  gchar *command_buffer;
  GladeStatusCode status;
 
  command_buffer = g_malloc (9 + strlen (project->xml_filename));
  sprintf (command_buffer, "glade-- %s", project->xml_filename);
  status = system (command_buffer) ? GLADE_STATUS_ERROR : GLADE_STATUS_OK;
  g_free (command_buffer);
  return status;
}


GladeStatusCode
glade_project_write_ada95_source (GladeProject *project)
{
  gchar *command_buffer;
  GladeStatusCode status;
 
  command_buffer = g_malloc (6 + strlen (project->xml_filename));
  sprintf (command_buffer, "gate %s", project->xml_filename);
  status = system (command_buffer) ? GLADE_STATUS_ERROR : GLADE_STATUS_OK;
  g_free (command_buffer);
  return status;
}


void
glade_project_set_changed		(GladeProject  *project,
					 gboolean	changed)
{
  project->changed = changed;
}


gchar*
glade_project_get_name	(GladeProject  *project)
{
  return project->name;
}


void
glade_project_set_name	(GladeProject  *project,
			 gchar	       *name)
{
  if (glade_util_strings_equivalent (project->name, name))
    return;
  g_free (project->name);
  project->name = g_strdup (name);
  glade_project_set_changed (project, TRUE);
}


gchar*
glade_project_get_xml_filename	(GladeProject  *project)
{
  return project->xml_filename;
}


void
glade_project_set_xml_filename	(GladeProject  *project,
				 gchar	       *filename)
{
  gchar *xml_directory;

  if (glade_util_strings_equivalent (project->xml_filename, filename))
    return;
  g_free (project->xml_filename);
  project->xml_filename = g_strdup (filename);
  xml_directory = glade_util_dirname (filename);

  /* If the project directories are not set, set them to defaults based on
     the directory the XML file is in. */
  if (project->directory == NULL || project->directory[0] == '\0')
    project->directory = g_strdup (xml_directory);
  if (project->source_directory == NULL
      || project->source_directory[0] == '\0')
    project->source_directory = g_strdup (xml_directory);
  if (project->pixmaps_directory == NULL
      || project->pixmaps_directory[0] == '\0')
    {
      project->pixmaps_directory = g_malloc (strlen (xml_directory) + 8);
      strcpy (project->pixmaps_directory, xml_directory);
      strcat (project->pixmaps_directory, "pixmaps");
    }

  g_free (xml_directory);
  glade_project_set_changed (project, TRUE);
}


gchar*
glade_project_get_directory	(GladeProject  *project)
{
  return project->directory;
}


void
glade_project_set_directory	(GladeProject  *project,
				 gchar	       *directory)
{
  if (glade_util_strings_equivalent (project->directory, directory))
    return;
  g_free (project->directory);
  project->directory = g_strdup (directory);
  glade_project_set_changed (project, TRUE);
}


gchar*
glade_project_get_source_directory	(GladeProject  *project)
{
  return project->source_directory;
}


void
glade_project_set_source_directory	(GladeProject  *project,
					 gchar	       *directory)
{
  gint directory_len;

  if (glade_util_strings_equivalent (project->source_directory, directory))
    return;
  g_free (project->source_directory);
  project->source_directory = g_strdup (directory);

  /* If the pixmaps directory is not set, set it to defaults based on
     the source directory. */
  if (directory && directory[0] != '\0')
    {
      if (project->pixmaps_directory == NULL
	  || project->pixmaps_directory[0] == '\0')
	{
	  directory_len = strlen (directory);
	  project->pixmaps_directory = g_malloc (directory_len + 9);
	  strcpy (project->pixmaps_directory, directory);
	  if (directory[directory_len - 1] != G_DIR_SEPARATOR)
	    strcat (project->pixmaps_directory, G_DIR_SEPARATOR_S);
	  strcat (project->pixmaps_directory, "pixmaps");
	}
    }

  glade_project_set_changed (project, TRUE);
}


gchar*
glade_project_get_pixmaps_directory	(GladeProject  *project)
{
  return project->pixmaps_directory;
}


void
glade_project_set_pixmaps_directory	(GladeProject  *project,
					 gchar	       *directory)
{
  if (glade_util_strings_equivalent (project->pixmaps_directory, directory))
    return;
  g_free (project->pixmaps_directory);
  project->pixmaps_directory = g_strdup (directory);
  glade_project_set_changed (project, TRUE);
}


gint
glade_project_get_language	(GladeProject  *project)
{
  return project->language;
}


void
glade_project_set_language	(GladeProject     *project,
				 GladeLanguageType language)
{
  if (project->language == language)
    return;
  project->language = language;
  property_show_lang_specific_page (language);
  glade_project_set_changed (project, TRUE);
}


gboolean
glade_project_set_language_name	(GladeProject  *project,
				 gchar	       *language_name)
{
  gint language;

  if (language_name == NULL || language_name[0] == '\0')
    return FALSE;

  for (language = 0; language < GladeNumLanguages; language++)
    {
      if (!strcmp (language_name, GladeLanguages[language]))
	{
	  glade_project_set_language (project, language);
	  return TRUE;
	}
    }
  return FALSE;
}


GtkWidget*
glade_project_new_component		(GladeProject  *project,
					 gchar	       *class_name,
					 gchar	       *name)
{
  GtkWidget *component = NULL;
  GbWidget * gbwidget;

  gbwidget = g_hash_table_lookup (gb_widget_table, class_name);
  g_return_val_if_fail (gbwidget != NULL, NULL);

  if (name == NULL)
    component = gb_widget_new (class_name, NULL);
  else
    component = gb_widget_new_full (class_name, FALSE, NULL, NULL, 0, 0,
				    NULL, GB_CREATING, NULL);
  if (component)
    glade_project_add_component (project, component);
  return component;
}


void
glade_project_add_component		(GladeProject  *project,
					 GtkWidget     *component)
{
  project->components = g_list_append (project->components, component);
#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_add_widget (component);
#endif
  gtk_signal_emit (GTK_OBJECT (project),
		   glade_project_signals[ADD_COMPONENT], component);
  glade_project_set_changed (project, TRUE);
}


void
glade_project_show_component		(GladeProject	*project,
					 GtkWidget	*component)
{
  /* Popup menus are shown in the menu editor. */
  if (GTK_IS_MENU (component))
    {
      GladeMenuEditor *menued;

      menued = GLADE_MENU_EDITOR (glade_menu_editor_new ());
      glade_project_add_associated_window (project, GTK_WIDGET (menued));
      glade_menu_editor_set_menu (menued, GTK_MENU_SHELL (component));
      gtk_signal_connect (GTK_OBJECT (menued->ok_button), "clicked",
			  GTK_SIGNAL_FUNC (on_menu_editor_ok),
			  component);
      gtk_signal_connect (GTK_OBJECT (menued->apply_button), "clicked",
			  GTK_SIGNAL_FUNC (on_menu_editor_apply),
			  component);
      gtk_signal_connect_object (GTK_OBJECT (menued->cancel_button), "clicked",
				 GTK_SIGNAL_FUNC (on_menu_editor_close),
				 GTK_OBJECT (menued));
      gtk_signal_connect_object (GTK_OBJECT (menued), "delete_event",
				 GTK_SIGNAL_FUNC (on_menu_editor_close),
				 GTK_OBJECT (menued));
      gtk_widget_show (GTK_WIDGET (menued));
    }
  else if (GTK_IS_WINDOW (component))
    {
      /* GtkWindows and things derived can be shown on their own. */
      gtk_widget_show (component);
      /* This maps the window, which also de-iconifies it according to the
	 ICCCM. */
      gdk_window_show (component->window);
      gdk_window_raise (component->window);
    }
  else
    {
      GtkWidget * fake_window;
      
      /* Non-Window component, make a window especially to show it. 
	 (pardon my grammer). */

      fake_window = component;
      while (GTK_WIDGET (fake_window)->parent)
	fake_window = (GtkWidget *)GTK_WIDGET (fake_window)->parent;

      if (fake_window == component)
	{
	  fake_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	  gtk_window_set_title (GTK_WINDOW (fake_window),
				gtk_widget_get_name (component));
	  gtk_container_border_width (GTK_CONTAINER (fake_window), 5);
	  gtk_container_add (GTK_CONTAINER (fake_window), component);

	  /* FIXME: Make a minimum size so we don't have to hunt for the
	     window if we like derive something from GtkMinisculeDot */

	  /* If we destroy window, don't destroy component. */
	  gtk_signal_connect (GTK_OBJECT (fake_window), "destroy",
			      GTK_SIGNAL_FUNC (on_fake_window_destroy),
			      (gpointer) component);

	  glade_project_add_associated_window (project, fake_window);
	}
      gtk_widget_show (component);
      gtk_widget_show (fake_window);

      /* Show and de-iconize window, as described above. */
      gdk_window_show (fake_window->window);
      gdk_window_raise (fake_window->window);
    }
}


static void
on_fake_window_destroy (GtkWidget * fake_window, GtkWidget * component)
{
  gtk_container_remove (GTK_CONTAINER (fake_window), component);
  glade_project_remove_associated_window (current_project, fake_window);
}


static void
on_menu_editor_ok (GtkWidget *button,
		   GtkWidget *component)
{
  GladeMenuEditor *menued;

  menued = GLADE_MENU_EDITOR (gtk_widget_get_toplevel (button));
  glade_menu_editor_update_menu (menued, GTK_MENU_SHELL (component));
  glade_project_remove_associated_window (current_project,
					  GTK_WIDGET (menued));
  gtk_widget_destroy (GTK_WIDGET (menued));
  glade_project_set_changed (current_project, TRUE);
}

static void
on_menu_editor_apply (GtkWidget *button,
		      GtkWidget *component)
{
  GladeMenuEditor *menued;

  menued = GLADE_MENU_EDITOR (gtk_widget_get_toplevel (button));
  glade_menu_editor_update_menu (menued, GTK_MENU_SHELL (component));
  glade_project_set_changed (current_project, TRUE);
}


static gint
on_menu_editor_close (GtkWidget *menued,
		      gpointer data)
{
  glade_project_remove_associated_window (current_project, menued);
  gtk_widget_destroy (menued);
  return TRUE;
}


void
glade_project_component_changed	(GladeProject  *project,
				 GtkWidget     *component)
{
  gtk_signal_emit (GTK_OBJECT (project),
		   glade_project_signals[COMPONENT_CHANGED], component);
  glade_project_set_changed (project, TRUE);
}


void
glade_project_foreach_component		(GladeProject  *project,
					 GtkCallback    callback,
					 gpointer       callback_data)
{
  GList *tmp_list;

  tmp_list = project->components;
  while (tmp_list)
    {
      (*callback) (GTK_WIDGET (tmp_list->data), callback_data);
      tmp_list = tmp_list->next;
    }
}


void
glade_project_remove_component		(GladeProject  *project,
					 GtkWidget     *component)
{
  project->components = g_list_remove (project->components, component);
  gtk_signal_emit (GTK_OBJECT (project),
		   glade_project_signals[REMOVE_COMPONENT], component);

  /* FIXME: These could be better. */
  property_set_widget (NULL);
  editor_clear_selection (NULL);

  gtk_widget_destroy (component);
  glade_project_set_changed (project, TRUE);
}


void
glade_project_add_associated_window	(GladeProject  *project,
					 GtkWidget     *window)
{
  project->associated_windows = g_list_append (project->associated_windows,
					       window);
}


void
glade_project_remove_associated_window	(GladeProject  *project,
					 GtkWidget     *window)
{
  project->associated_windows = g_list_remove (project->associated_windows,
					       window);
}


/* These add/remove pixmaps to the project. The same filename can appear
   more than once in the project's list of pixmaps, so refcounting isn't
   needed. The filename is copied. */
void
glade_project_add_pixmap		(GladeProject  *project,
					 gchar	       *filename)
{
  project->pixmap_filenames = g_list_prepend (project->pixmap_filenames,
					      g_strdup (filename));
  glade_project_set_changed (project, TRUE);
}


void
glade_project_remove_pixmap		(GladeProject  *project,
					 gchar	       *filename)
{
  GList *element;

  element = project->pixmap_filenames;
  while (element)
    {
      if (!strcmp (element->data, filename))
	{
	  project->pixmap_filenames = g_list_remove_link (project->pixmap_filenames, element);
	  g_free (element->data);
	  g_list_free (element);
	  break;
	}
      element = element->next;
    }

  glade_project_set_changed (project, TRUE);
}


/* This ensures that all pixmaps are in the project's pixmaps directory,
   by copying them if necessary. Note that it doesn't change the pixmap
   filenames in the list. They will be updated the next time the project is
   opened. */
void
glade_project_copy_all_pixmaps (GladeProject  *project)
{
  gchar *pixmaps_dir, *filename, *basename, *new_filename = NULL;
  gint pixmaps_dir_len, new_filename_len, new_filename_space = 0;
  GList *element;
  gboolean copied;
  struct stat filestat;

  pixmaps_dir = glade_project_get_pixmaps_directory (project);
  g_return_if_fail (pixmaps_dir != NULL);
  g_return_if_fail (pixmaps_dir[0] != '\0');

  pixmaps_dir_len = strlen (pixmaps_dir);

  element = project->pixmap_filenames;
  while (element)
    {
      filename = (gchar*) element->data;

      /* If the start of the pixmap filename doesn't match the pixmaps
	 directory, we make sure it is copied there. */
      if (strncmp (pixmaps_dir, filename, pixmaps_dir_len)
	  || (pixmaps_dir[pixmaps_dir_len - 1] != G_DIR_SEPARATOR
	      && filename[0] != G_DIR_SEPARATOR))
	{
	  basename = g_basename (filename);
	  new_filename_len = pixmaps_dir_len + 1 + strlen (basename) + 1;
	  if (new_filename == NULL || new_filename_len > new_filename_space)
	    {
	      g_free (new_filename);
	      new_filename_space = new_filename_len + 50;
	      new_filename = g_malloc (new_filename_space);
	    }

	  strcpy (new_filename, pixmaps_dir);
	  if (pixmaps_dir[pixmaps_dir_len - 1] != G_DIR_SEPARATOR)
	    strcat (new_filename, G_DIR_SEPARATOR_S);
	  strcat (new_filename, basename);

	  /* Check if it already exists, and copy if it doesn't. */
	  if (!glade_util_file_exists (new_filename))
	    {
	      /* Create the pixmaps directory if it doesn't exist. */
	      if (stat (project->pixmaps_directory, &filestat) != 0)
		{
		  if (errno == ENOENT)
		    {
		      if (mkdir (project->pixmaps_directory, 0777) != 0)
			{
			  g_warning ("Couldn't create pixmaps directory");
			  break;
			}
		    }
		  else
		    {
		      g_warning ("Couldn't stat pixmaps directory");
		      break;
		    }
		}
	      else if (!S_ISDIR (filestat.st_mode))
		{
		  g_warning ("Pixmaps directory is not a directory");
		  break;
		}

	      copied = glade_util_copy_file (filename, new_filename);
	      if (!copied)
		g_warning ("Error copying file");
	    }
	}
      element = element->next;
    }

  if (new_filename)
    g_free (new_filename);
}


gchar*
glade_project_get_current_directory	(GladeProject  *project)
{
  return project->current_directory;
}


void
glade_project_set_current_directory	(GladeProject  *project,
					 gchar	       *directory)
{
  g_free (project->current_directory);
  project->current_directory = g_strdup (directory);
}


/*
 * C Output options.
 */
gboolean
glade_project_get_gettext_support	(GladeProject  *project)
{
  return project->gettext_support;
}


void
glade_project_set_gettext_support	(GladeProject  *project,
					 gboolean       gettext_support)
{
  if (project->gettext_support == gettext_support)
    return;
  project->gettext_support = gettext_support;
  glade_project_set_changed (project, TRUE);
}


gboolean
glade_project_get_use_widget_names	(GladeProject  *project)
{
  return project->use_widget_names;
}


void
glade_project_set_use_widget_names	(GladeProject  *project,
					 gboolean       use_widget_names)
{
  if (project->use_widget_names == use_widget_names)
    return;
  project->use_widget_names = use_widget_names;
  glade_project_set_changed (project, TRUE);
}


/* FIXME: These will be removed when the source code output is improved. */
void
glade_project_get_source_files (GladeProject *project,
				gchar	**main_source_file,
				gchar	**main_header_file,
				gchar	**handler_source_file,
				gchar	**handler_header_file)
{
  *main_source_file = project->main_source_file;
  *main_header_file = project->main_header_file;
  *handler_source_file = project->handler_source_file;
  *handler_source_file = project->handler_header_file;
}


void
glade_project_set_source_files (GladeProject *project,
				gchar	 *main_source_file,
				gchar	 *main_header_file,
				gchar	 *handler_source_file,
				gchar	 *handler_header_file)
{
  if (glade_util_strings_equivalent (project->main_source_file,
				     main_source_file)
      && glade_util_strings_equivalent (project->main_header_file,
					main_header_file)
      && glade_util_strings_equivalent (project->handler_source_file,
					handler_source_file)
      && glade_util_strings_equivalent (project->handler_header_file,
					handler_header_file))
    return;

  g_free (project->main_source_file);
  g_free (project->main_header_file);
  g_free (project->handler_source_file);
  g_free (project->handler_header_file);
  project->main_source_file = g_strdup (main_source_file);
  project->main_header_file = g_strdup (main_header_file);
  project->handler_source_file = g_strdup (handler_source_file);
  project->handler_header_file = g_strdup (handler_header_file);
  glade_project_set_changed (project, TRUE);
}


void
glade_project_load_options (GladeProject *project,
			    GbWidgetSetArgData * data)
{
  gchar *element, *cdata;
  gint element_index, cdata_index;
  gchar *base;

  base = g_dirname (project->xml_filename);

  for (;;)
    {
      load_element (data, &element_index, &cdata_index);
      element = data->buffer.ptr + element_index;
      cdata = data->buffer.ptr + cdata_index;

      if (data->status != GLADE_STATUS_OK)
	{
	  if (data->status == GLADE_STATUS_START_TAG_EXPECTED
	      && data->token_type == GB_TOKEN_END_TAG
	      && (!strcmp (data->buffer.ptr + data->token, "project")))
	    {
	      data->status = GLADE_STATUS_OK;
	      break;
	    }
	  MSG ("Load error");
	  break;
	}

      if (!strcmp (element, "name"))
	{
	  g_free (project->name);
	  project->name = glade_util_copy_string (cdata);
	}
      else if ((!strcmp (element, "directory"))
	       || (!strcmp (element, "project_dir"))) /* Old name. */
	{
 	  g_free (project->directory);
 	  project->directory = glade_util_make_absolute_path (base, cdata);
	}
      else if ((!strcmp (element, "source_directory"))
	       || (!strcmp (element, "srcdir"))) /* Old name. */
	{
 	  g_free (project->source_directory);
 	  project->source_directory = glade_util_make_absolute_path (base,
								     cdata);
	}
      else if (!strcmp (element, "pixmaps_directory"))
	{
 	  g_free (project->pixmaps_directory);
 	  project->pixmaps_directory = glade_util_make_absolute_path (base,
								      cdata);
	}
      else if (!strcmp (element, "language"))
	{
	  if (!glade_project_set_language_name (project, cdata))
	    g_warning ("Invalid source language");
	}
      else if ((!strcmp (element, "gettext_support"))
	       || (!strcmp (element, "i18n_strings"))) /* Old name. */
       {
         project->gettext_support = load_parse_bool (data, cdata);
       }
      else if (!strcmp (element, "use_widget_names"))
       {
         project->use_widget_names = load_parse_bool (data, cdata);
       }

      else if (!strcmp (element, "main_source_file"))
	{
	  g_free (project->main_source_file);
	  project->main_source_file = glade_util_copy_string (cdata);
	}
      else if (!strcmp (element, "main_header_file"))
	{
	  g_free (project->main_header_file);
	  project->main_header_file = glade_util_copy_string (cdata);
	}
      else if (!strcmp (element, "handler_source_file"))
	{
	  g_free (project->handler_source_file);
	  project->handler_source_file = glade_util_copy_string (cdata);
	}
      else if (!strcmp (element, "handler_header_file"))
	{
	  g_free (project->handler_header_file);
	  project->handler_header_file = glade_util_copy_string (cdata);
	}

      else
	{
	  g_warning ("Invalid element in project: %s\n", element);
	}

      if (data->status != GLADE_STATUS_OK)
	break;
    }

  g_free (base);
}


void
glade_project_save_options (GladeProject *project,
			    GbWidgetGetArgData * data)
{
  gchar *base_dir, *dir;

  save_tag (data, "project");
  data->indent++;

  save_string (data, "name", project->name);

  /* All directories are saved relative to the xml file's directory. */
  base_dir = glade_util_dirname (project->xml_filename);

  dir = glade_util_make_relative_path (base_dir, project->directory);
  save_string (data, "directory", dir);
  g_free (dir);

  dir = glade_util_make_relative_path (base_dir, project->source_directory);
  save_string (data, "source_directory", dir);
  g_free (dir);

  dir = glade_util_make_relative_path (base_dir, project->pixmaps_directory);
  save_string (data, "pixmaps_directory", dir);
  g_free (dir);

  g_free (base_dir);

  save_string (data, "language", GladeLanguages[project->language]);
  save_bool   (data, "gettext_support", project->gettext_support);
  save_bool   (data, "use_widget_names", project->use_widget_names);

  save_string (data, "main_source_file", project->main_source_file);
  save_string (data, "main_header_file", project->main_header_file);
  save_string (data, "handler_source_file", project->handler_source_file);
  save_string (data, "handler_header_file", project->handler_header_file);

  data->indent--;
  save_tag (data, "/project");
}


#ifndef GTK_HAVE_FEATURES_1_1_0
static void
glade_project_marshal_signal_1 (GtkObject * object,
				GtkSignalFunc func,
				gpointer func_data,
				GtkArg * args)
{
  GladeProjectSignal1 rfunc;

  rfunc = (GladeProjectSignal1) func;

  (*rfunc) (object, GTK_VALUE_OBJECT (args[0]), func_data);
}
#endif



/*
 * Functions to do with ensuring widget names are unique.
 */

/* This returns a unique name for a widget using the given base name, often
   a widget class name. If the base name starts with 'Gtk' or 'Gnome' then
   that is taken off. The base name is converted to lower case and a number is
   added on to the end of it to ensure that no other widget in the project
   uses the same name. */
gchar*
glade_project_new_widget_name (GladeProject *project,
			       gchar	    *base_name)
{
  char new_widget_name[128];
  gint widget_id, i, new_widget_name_len;

  /* Check we won't overflow the buffer. */
  g_return_val_if_fail (strlen (base_name) < 100, base_name);

  /* Skip 'Gtk' or 'Gnome' at the start of all class names. */
  if (!strncmp (base_name, "Gtk", 3))
    base_name += 3;
  else if (!strncmp (base_name, "Gnome", 5))
    base_name += 5;

  strcpy (new_widget_name, base_name);
  /* convert name to lower case (only normal ASCII chars) */
  new_widget_name_len = strlen (new_widget_name);
  for (i = 0; i < new_widget_name_len; i++)
    {
      if ((new_widget_name[i] >= 'A') && (new_widget_name[i] <= 'Z'))
	new_widget_name[i] += 'a' - 'A';
    }

  widget_id = GPOINTER_TO_INT (g_hash_table_lookup (project->unique_id_hash,
						    new_widget_name));
  if (widget_id == 0)
    {
      widget_id = 1;
      g_hash_table_insert (project->unique_id_hash, g_strdup (new_widget_name),
			   GINT_TO_POINTER (widget_id));
    }
  else
    {
      widget_id++;
      /* We don't need to g_strdup new_widget_name since it is already in the
         hash. */
      g_hash_table_insert (project->unique_id_hash, new_widget_name,
			   GINT_TO_POINTER (widget_id));
    }

  /* Add the ID onto the end of the name. */
  sprintf (new_widget_name + strlen (new_widget_name), "%i", widget_id);

  MSG1 ("Allocating new widget name: %s", new_widget_name);
  return g_strdup (new_widget_name);
}


/* This returns the last ID used for the given name. */
gint
glade_project_get_last_id (GladeProject *project,
			   gchar	*name)
{
  return GPOINTER_TO_INT (g_hash_table_lookup (project->unique_id_hash, name));
}


/* This sets the last ID used for the given name. It checks if the key
   already exists, in which case we don't have to g_strdup it. */
void
glade_project_set_last_id (GladeProject *project,
			   gchar	*name,
			   gint		 last_id)
{
  if (g_hash_table_lookup (project->unique_id_hash, name))
    g_hash_table_insert (project->unique_id_hash, name,
			 GINT_TO_POINTER (last_id));
  else
    g_hash_table_insert (project->unique_id_hash, g_strdup (name),
			 GINT_TO_POINTER (last_id));
}


/* Check if the name has a trailing ID number, and if so compare it to the
   current maximum in the ID hash and update if necessary. */
void
glade_project_reserve_name (GladeProject *project,
			    gchar	 *name)
{
  gchar buffer[128];
  gchar *id_start = name + strlen (name) - 1;
  gint id = 0, current_id;

  while (*id_start >= '0' && *id_start <= '9')
    id_start--;
  id_start++;

  /* If there is no trailing number just return. */
  if (*id_start == '\0')
    return;

  /* Make sure we won't overflow the buffer. */
  g_return_if_fail (id_start - name < 127);

  id = atoi (id_start);
  if (id == 0)
    return;

  strncpy (buffer, name, id_start - name);
  buffer[id_start - name] = '\0';

  current_id = GPOINTER_TO_INT (g_hash_table_lookup (project->unique_id_hash,
						     buffer));
  if (current_id == 0)
    {
      g_hash_table_insert (project->unique_id_hash, g_strdup (buffer),
			   GINT_TO_POINTER (id));
    }
  else if (id > current_id)
    {
      /* We don't need to g_strdup buffer since it is already in the hash. */
      g_hash_table_insert (project->unique_id_hash, buffer,
			   GINT_TO_POINTER (id));
    }
}


/* This recursively descends a widget tree, ensuring that all widgets have
   names. If any names are NULL, default names are created for them. */
void
glade_project_ensure_widgets_named (GladeProject *project,
				    GtkWidget    *widget)
{
  /* We have to switch the aguments for the recursive calls to work. */
  glade_project_real_ensure_widgets_named (widget, project);
}


static void
glade_project_real_ensure_widgets_named (GtkWidget    *widget,
					 GladeProject *project)
{
  if (GB_IS_GB_WIDGET (widget) && widget->name == NULL)
    {
      gchar *class = gtk_type_name (GTK_OBJECT_TYPE (widget));
      gtk_widget_set_name (widget,
			   glade_project_new_widget_name (project, class));
      MSG1 ("set default name for widget: %s\n", gtk_widget_get_name (widget));
    }

  gb_widget_children_foreach (widget, (GtkCallback) glade_project_real_ensure_widgets_named, project);
}
