/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>

#include "gladeconfig.h"

#include "glade_project_options.h"
#include "utils.h"

/* These are used when showing the file selection to indicate which option we
   are setting. */
typedef enum
{
  GladeProjectOptionXMLFilename,
  GladeProjectOptionDirectory,
  GladeProjectOptionSourceDirectory,
  GladeProjectOptionPixmapsDirectory
} GladeProjectOptionType;

/* This key is used to save the project option in the file selection. */
const gchar *GladeProjectOptionKey = "GladeProjectOptionKey";

static GtkWindowClass *parent_class = NULL;


static void glade_project_options_class_init (GladeProjectOptionsClass * klass);
static void glade_project_options_init (GladeProjectOptions * project_options);
static void glade_project_options_destroy (GtkObject *object);

static void glade_project_options_set_project (GladeProjectOptions *options,
					       GladeProject *project);
static void glade_project_options_show_file_selection (GtkWidget *widget,
						       gpointer data);
static void glade_project_options_set_filename (GtkWidget *widget,
						GladeProjectOptions *options);
static void glade_project_options_ok (GtkWidget *widget,
				      GladeProjectOptions *options);
static gchar* get_entry_text (GtkWidget *entry);

guint
glade_project_options_get_type (void)
{
  static guint glade_project_options_type = 0;

  if (!glade_project_options_type)
    {
      GtkTypeInfo glade_project_options_info =
      {
	"GladeProjectOptions",
	sizeof (GladeProjectOptions),
	sizeof (GladeProjectOptionsClass),
	(GtkClassInitFunc) glade_project_options_class_init,
	(GtkObjectInitFunc) glade_project_options_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
#ifdef GTK_HAVE_FEATURES_1_1_0
	(GtkClassInitFunc) NULL,
#endif
      };

      glade_project_options_type = gtk_type_unique (gtk_window_get_type (),
						  &glade_project_options_info);
    }
  return glade_project_options_type;
}


static void
glade_project_options_class_init (GladeProjectOptionsClass * klass)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass *) klass;

  parent_class = gtk_type_class (gtk_window_get_type ());

  object_class->destroy = glade_project_options_destroy;
}


static void
glade_project_options_init (GladeProjectOptions * options)
{
  GtkTooltips *tooltips;
  GtkWidget *vbox, *notebook, *table, *label, *button, *eventbox;
  GtkWidget *hbox, *hbbox, *alignment, *menu, *menuitem;
  gint row, language;

  options->filesel = NULL;

  tooltips = gtk_tooltips_new ();

  gtk_container_border_width (GTK_CONTAINER (options), 4);
  gtk_window_position (GTK_WINDOW (options), GTK_WIN_POS_MOUSE);
  /* We want a reasonable size to display directories so we set the usize for
     thw window. All the entries will expand. */
  gtk_widget_set_usize (GTK_WIDGET (options), 500, -1);

  gtk_window_set_title (GTK_WINDOW (options), _("Project Options"));
  gtk_window_set_policy (GTK_WINDOW (options), TRUE, TRUE, FALSE);

  vbox = gtk_vbox_new (FALSE, 4);
  gtk_container_add (GTK_CONTAINER (options), vbox);
  gtk_widget_show (vbox);

  notebook = gtk_notebook_new ();
  gtk_widget_show (notebook);
  gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);


  /*
   * General Page.
   */

  label = gtk_label_new (_("General"));
  gtk_widget_show (label);

  table = gtk_table_new (5, 2, FALSE);
  gtk_widget_show (table);
  gtk_table_set_row_spacings (GTK_TABLE (table), 2);
  gtk_table_set_col_spacings (GTK_TABLE (table), 2);
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), table, label);
  gtk_container_border_width (GTK_CONTAINER (table), 7);

  /* Project Name. */
  row = 0;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("The name of the current project"), NULL);

  label = gtk_label_new (_("Project Name:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->name_entry = gtk_entry_new ();
  gtk_widget_show (options->name_entry);
  gtk_table_attach (GTK_TABLE (table), options->name_entry,
		    1, 3, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  /* Project XML File. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("The project file"), NULL);

  label = gtk_label_new (_("Project File:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->xml_filename_entry = gtk_entry_new ();
  gtk_widget_show (options->xml_filename_entry);
  gtk_table_attach (GTK_TABLE (table), options->xml_filename_entry,
		    1, 2, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  button = gtk_button_new_with_label ("...");
  gtk_widget_show (button);
  gtk_table_attach (GTK_TABLE (table), button, 2, 3, row, row + 1,
		    0, 0, 0, 0);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (glade_project_options_show_file_selection),
		      GINT_TO_POINTER (GladeProjectOptionXMLFilename));

  /* Project Directory. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("The project directory"), NULL);

  label = gtk_label_new (_("Project Directory:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->directory_entry = gtk_entry_new ();
  gtk_widget_show (options->directory_entry);
  gtk_table_attach (GTK_TABLE (table), options->directory_entry,
		    1, 2, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  button = gtk_button_new_with_label ("...");
  gtk_widget_show (button);
  gtk_table_attach (GTK_TABLE (table), button, 2, 3, row, row + 1,
		    0, 0, 0, 0);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (glade_project_options_show_file_selection),
		      GINT_TO_POINTER (GladeProjectOptionDirectory));

  /* Project Source Directory. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("The directory to save generated source code"),
			NULL);

  label = gtk_label_new (_("Source Directory:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->source_directory_entry = gtk_entry_new ();
  gtk_widget_show (options->source_directory_entry);
  gtk_table_attach (GTK_TABLE (table), options->source_directory_entry,
		    1, 2, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  button = gtk_button_new_with_label ("...");
  gtk_widget_show (button);
  gtk_table_attach (GTK_TABLE (table), button, 2, 3, row, row + 1, 0, 0, 0, 0);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (glade_project_options_show_file_selection),
		      GINT_TO_POINTER (GladeProjectOptionSourceDirectory));

  /* Project Pixmaps Directory. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("The directory to store pixmaps"),
			NULL);

  label = gtk_label_new (_("Pixmaps Directory:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->pixmaps_directory_entry = gtk_entry_new ();
  gtk_widget_show (options->pixmaps_directory_entry);
  gtk_table_attach (GTK_TABLE (table), options->pixmaps_directory_entry,
		    1, 2, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  button = gtk_button_new_with_label ("...");
  gtk_widget_show (button);
  gtk_table_attach (GTK_TABLE (table), button, 2, 3, row, row + 1, 0, 0, 0, 0);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (glade_project_options_show_file_selection),
		      GINT_TO_POINTER (GladeProjectOptionPixmapsDirectory));

  /* Project License. */
  /* Take this out until we actually use it. */
#if 0
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("The license which is added at the top of generated files"), NULL);

  label = gtk_label_new (_("License:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->license_option_menu = gtk_option_menu_new ();
  gtk_widget_show (options->license_option_menu);
  gtk_table_attach (GTK_TABLE (table), options->license_option_menu,
		    1, 3, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  menu = gtk_menu_new ();
  menuitem = gtk_menu_item_new_with_label ("GNU General Public License");
  gtk_widget_show (menuitem);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_option_menu_set_menu (GTK_OPTION_MENU (options->license_option_menu),
			    menu);
#endif

  /* Source Language. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("The source code language"), NULL);

  label = gtk_label_new (_("Language:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  alignment = gtk_alignment_new (0, 0.5, 0, 0);
  gtk_widget_show (alignment);
  gtk_table_attach (GTK_TABLE (table), alignment,
		    1, 3, row, row + 1, GTK_FILL, 0, 0, 0);

  options->language_option_menu = gtk_option_menu_new ();
  gtk_widget_show (options->language_option_menu);
  gtk_container_add (GTK_CONTAINER (alignment), options->language_option_menu);

  menu = gtk_menu_new ();
  for (language = 0; language < GladeNumLanguages; language++)
    {
      menuitem = gtk_menu_item_new_with_label (GladeLanguages[language]);
      gtk_widget_set_usize (menuitem, 100, -1);
      gtk_widget_show (menuitem);
      gtk_menu_append (GTK_MENU (menu), menuitem);
    }
  gtk_option_menu_set_menu (GTK_OPTION_MENU (options->language_option_menu),
			    menu);

  /* Gettext Support. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("If strings are marked for translation by gettext"),
			NULL);

  label = gtk_label_new (_("Gettext Support:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  alignment = gtk_alignment_new (0, 0.5, 0, 0);
  gtk_widget_show (alignment);
  gtk_table_attach (GTK_TABLE (table), alignment,
		    1, 3, row, row + 1, GTK_FILL, 0, 0, 0);

  hbox = gtk_hbox_new (TRUE, 16);
  gtk_container_add (GTK_CONTAINER (alignment), hbox);
  gtk_widget_show (hbox);

  options->gettext_support_yes = gtk_radio_button_new_with_label (NULL,
								  _("Yes"));
  gtk_widget_show (options->gettext_support_yes);
  gtk_box_pack_start (GTK_BOX (hbox), options->gettext_support_yes,
		      FALSE, TRUE, 0);

  options->gettext_support_no = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (options->gettext_support_yes), _("No"));
  gtk_widget_show (options->gettext_support_no);
  gtk_box_pack_start (GTK_BOX (hbox), options->gettext_support_no,
		      FALSE, TRUE, 0);

  /* Setting widget names. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox,
			_("If widget names are set in the source code"),
			NULL);

  label = gtk_label_new (_("Set Widget Names:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  alignment = gtk_alignment_new (0, 0.5, 0, 0);
  gtk_widget_show (alignment);
  gtk_table_attach (GTK_TABLE (table), alignment,
		    1, 3, row, row + 1, GTK_FILL, 0, 0, 0);

  hbox = gtk_hbox_new (TRUE, 16);
  gtk_container_add (GTK_CONTAINER (alignment), hbox);
  gtk_widget_show (hbox);

  options->use_widget_names_yes = gtk_radio_button_new_with_label (NULL,
								  _("Yes"));
  gtk_widget_show (options->use_widget_names_yes);
  gtk_box_pack_start (GTK_BOX (hbox), options->use_widget_names_yes,
		      FALSE, TRUE, 0);

  options->use_widget_names_no = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (options->use_widget_names_yes), _("No"));
  gtk_widget_show (options->use_widget_names_no);
  gtk_box_pack_start (GTK_BOX (hbox), options->use_widget_names_no,
		      FALSE, TRUE, 0);


  /*
   * Filenames Page.
   *
   * Note that these will disappear when we update the source code output.
   */

  label = gtk_label_new (_("Filenames"));
  gtk_widget_show (label);

  table = gtk_table_new (4, 2, FALSE);
  gtk_table_set_row_spacings (GTK_TABLE (table), 2);
  gtk_table_set_col_spacings (GTK_TABLE (table), 2);
  gtk_widget_show (table);
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), table, label);
  gtk_container_border_width (GTK_CONTAINER (table), 7);
  gtk_table_set_row_spacings (GTK_TABLE (table), 3);
  gtk_table_set_col_spacings (GTK_TABLE (table), 10);

  /* Main source file. */
  row = 0;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox, _("The filename which the GUI interface generation source code will be stored into"), NULL);

  label = gtk_label_new (_("Main Source Filename:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->main_source_entry = gtk_entry_new ();
  gtk_widget_show (options->main_source_entry);
  gtk_table_attach (GTK_TABLE (table), options->main_source_entry,
		    1, 2, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  /* Main header file. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox, _("The filename which the GUI interface generation source headers will be stored into"), NULL);

  label = gtk_label_new (_("Main Header Filename:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->main_header_entry = gtk_entry_new ();
  gtk_widget_show (options->main_header_entry);
  gtk_table_attach (GTK_TABLE (table), options->main_header_entry,
		    1, 2, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);


  /* Handler source file. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox, _("The filename which the GUI signal handler source code will be stored into"), NULL);

  label = gtk_label_new (_("Handler Source Filename:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->handler_source_entry = gtk_entry_new ();
  gtk_widget_show (options->handler_source_entry);
  gtk_table_attach (GTK_TABLE (table), options->handler_source_entry,
		    1, 2, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);


  /* Handler header file. */
  row++;
  eventbox = gtk_event_box_new ();
  gtk_widget_show (eventbox);
  gtk_table_attach (GTK_TABLE (table), eventbox, 0, 1, row, row + 1,
		    GTK_FILL, 0, 0, 0);
  gtk_tooltips_set_tip (tooltips, eventbox, _("The filename which the GUI signal handler source headers will be stored into"), NULL);

  label = gtk_label_new (_("Handler Header Filename:"));
  gtk_widget_show (label);
  gtk_container_add (GTK_CONTAINER (eventbox), label);
  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

  options->handler_header_entry = gtk_entry_new ();
  gtk_widget_show (options->handler_header_entry);
  gtk_table_attach (GTK_TABLE (table), options->handler_header_entry,
		    1, 2, row, row + 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);


  /* The button box with OK & Cancel buttons. */
  hbbox = gtk_hbutton_box_new ();
  gtk_widget_show (hbbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbbox, TRUE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbbox), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbbox), 8);

  options->ok_button = gtk_button_new_with_label (_("OK"));
  GTK_WIDGET_SET_FLAGS (options->ok_button, GTK_CAN_DEFAULT);
  gtk_widget_show (options->ok_button);
  gtk_container_add (GTK_CONTAINER (hbbox), options->ok_button);
  gtk_widget_grab_default (options->ok_button);
  gtk_signal_connect (GTK_OBJECT (options->ok_button), "clicked",
		      GTK_SIGNAL_FUNC (glade_project_options_ok), options);

  options->cancel_button = gtk_button_new_with_label (_("Cancel"));
  GTK_WIDGET_SET_FLAGS (options->cancel_button, GTK_CAN_DEFAULT);
  gtk_widget_show (options->cancel_button);
  gtk_container_add (GTK_CONTAINER (hbbox), options->cancel_button);
  gtk_signal_connect_object (GTK_OBJECT (options->cancel_button), "clicked",
			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     GTK_OBJECT (options));
}


GtkWidget *
glade_project_options_new (GladeProject *project)
{
  GladeProjectOptions *options;

  g_return_val_if_fail (GLADE_IS_PROJECT (project), NULL);

  options = GLADE_PROJECT_OPTIONS (gtk_type_new (glade_project_options_get_type ()));
  glade_project_options_set_project (options, project);

  return GTK_WIDGET (options);
}


static void
glade_project_options_destroy (GtkObject *object)
{
  GladeProjectOptions *options;

  options = GLADE_PROJECT_OPTIONS (object);

  /* FIXME: Maybe we need to disconnect the destroy signal handler added to
     the project. */

  if (options->filesel)
    gtk_widget_destroy (options->filesel);
}


/* This sets which project which we show the options of. */
static void
glade_project_options_set_project (GladeProjectOptions *options,
				   GladeProject *project)
{
  g_return_if_fail (GLADE_IS_PROJECT (project));

  options->project = project;

  gtk_entry_set_text (GTK_ENTRY (options->name_entry),
		      project->name ? project->name : "");
  gtk_entry_set_text (GTK_ENTRY (options->xml_filename_entry),
		      project->xml_filename ? project->xml_filename : "");
  gtk_entry_set_text (GTK_ENTRY (options->directory_entry),
		      project->directory ? project->directory : "");
  gtk_entry_set_text (GTK_ENTRY (options->source_directory_entry),
		      project->source_directory ? project->source_directory
		                                : "");
  gtk_entry_set_text (GTK_ENTRY (options->pixmaps_directory_entry),
		      project->pixmaps_directory ? project->pixmaps_directory
		                                : "");

  gtk_option_menu_set_history (GTK_OPTION_MENU (options->language_option_menu),
			       project->language);

  if (project->gettext_support)
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (options->gettext_support_yes), TRUE);
  else
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (options->gettext_support_no), TRUE);

  if (project->use_widget_names)
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (options->use_widget_names_yes), TRUE);
  else
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (options->use_widget_names_no), TRUE);

  gtk_entry_set_text (GTK_ENTRY (options->main_source_entry),
		      project->main_source_file ? project->main_source_file
						: "");
  gtk_entry_set_text (GTK_ENTRY (options->main_header_entry),
		      project->main_header_file ? project->main_header_file
						: "");
  gtk_entry_set_text (GTK_ENTRY (options->handler_source_entry),
		      project->handler_source_file
			? project->handler_source_file : "");
  gtk_entry_set_text (GTK_ENTRY (options->handler_header_entry),
		      project->handler_header_file
			? project->handler_header_file : "");

  /* Connect to the project's destroy signal so we can destroy ourself. */
  gtk_signal_connect_object (GTK_OBJECT (project), "destroy",
			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     GTK_OBJECT (options));
}


static void
glade_project_options_show_file_selection (GtkWidget *widget,
					   gpointer   data)
{
  GladeProjectOptions *options;
  gchar *filename;
  GladeProjectOptionType option;

  options = GLADE_PROJECT_OPTIONS (gtk_widget_get_toplevel (widget));

  if (options->filesel == NULL)
    {
      options->filesel = gtk_file_selection_new (_("Select File"));
      gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (options->filesel)->ok_button),
			  "clicked",
			  GTK_SIGNAL_FUNC (glade_project_options_set_filename),
			  options);

      gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (options->filesel)->cancel_button),
				 "clicked",
				 GTK_SIGNAL_FUNC (glade_util_close_window),
				 GTK_OBJECT (options->filesel));
      gtk_signal_connect (GTK_OBJECT (options->filesel), "delete_event",
			  GTK_SIGNAL_FUNC (glade_util_close_window_on_delete),
			  NULL);
    }

  option = GPOINTER_TO_INT (data);
  gtk_object_set_data (GTK_OBJECT (options->filesel), GladeProjectOptionKey,
		       data);
  switch (option)
    {
      case GladeProjectOptionXMLFilename:
	filename = gtk_entry_get_text (GTK_ENTRY (options->xml_filename_entry));
	break;
      case GladeProjectOptionDirectory:
	filename = gtk_entry_get_text (GTK_ENTRY (options->directory_entry));
	break;
      case GladeProjectOptionSourceDirectory:
	filename = gtk_entry_get_text (GTK_ENTRY (options->source_directory_entry));
	break;
      case GladeProjectOptionPixmapsDirectory:
	filename = gtk_entry_get_text (GTK_ENTRY (options->pixmaps_directory_entry));
	break;
    default:
      g_warning ("Invalid project option");
      return;
    }

  gtk_file_selection_set_filename (GTK_FILE_SELECTION (options->filesel),
				   filename);
  gtk_widget_show (options->filesel);
}


static void
glade_project_options_set_filename (GtkWidget *widget,
				    GladeProjectOptions *options)
{
  GladeProjectOptionType option;
  gchar *filename;

  option = GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT (options->filesel),
						 GladeProjectOptionKey));
  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (options->filesel));

  switch (option)
    {
      case GladeProjectOptionXMLFilename:
	gtk_entry_set_text (GTK_ENTRY (options->xml_filename_entry),
			    filename);
	break;
      case GladeProjectOptionDirectory:
	gtk_entry_set_text (GTK_ENTRY (options->directory_entry),
			    filename);
	break;
      case GladeProjectOptionSourceDirectory:
	gtk_entry_set_text (GTK_ENTRY (options->source_directory_entry),
			    filename);
	break;
      case GladeProjectOptionPixmapsDirectory:
	gtk_entry_set_text (GTK_ENTRY (options->pixmaps_directory_entry),
			    filename);
	break;
    default:
      g_warning ("Invalid project option");
    }

  gtk_widget_hide (options->filesel);
}


static void
glade_project_options_ok (GtkWidget *widget,
			  GladeProjectOptions *options)
{
  GladeProject *project;
  GtkWidget *label;
  gchar *xml_filename, *directory, *source_directory, *pixmaps_directory;
  gchar *label_text;

  project = options->project;

  glade_project_set_name (project, get_entry_text (options->name_entry));

  /* We set these in this order so that if a directory isn't set explicitly,
     it will take a default value based on the other general settings. */
  pixmaps_directory = get_entry_text (options->pixmaps_directory_entry);
  if (pixmaps_directory)
    pixmaps_directory = glade_util_make_absolute_path (glade_current_directory,
						       pixmaps_directory);
  glade_project_set_pixmaps_directory (project, pixmaps_directory);
  g_free (pixmaps_directory);

  source_directory = get_entry_text (options->source_directory_entry);
  if (source_directory)
    source_directory = glade_util_make_absolute_path (glade_current_directory,
						      source_directory);
  glade_project_set_source_directory (project, source_directory);
  g_free (source_directory);

  directory = get_entry_text (options->directory_entry);
  if (directory)
    directory = glade_util_make_absolute_path (glade_current_directory,
					       directory);
  glade_project_set_directory (project, directory);
  g_free (directory);

  xml_filename = get_entry_text (options->xml_filename_entry);
  if (xml_filename)
    xml_filename = glade_util_make_absolute_path (glade_current_directory,
						  xml_filename);
  glade_project_set_xml_filename (project, xml_filename);


  label = GTK_BUTTON (options->language_option_menu)->child;
  gtk_label_get (GTK_LABEL (label), &label_text);
  glade_project_set_language_name (project, label_text);

  glade_project_set_gettext_support (project, GTK_TOGGLE_BUTTON (options->gettext_support_yes)->active ? TRUE : FALSE);

  glade_project_set_use_widget_names (project, GTK_TOGGLE_BUTTON (options->use_widget_names_yes)->active ? TRUE : FALSE);

  glade_project_set_source_files (project,
				  get_entry_text (options->main_source_entry),
				  get_entry_text (options->main_header_entry),
				  get_entry_text (options->handler_source_entry),
				  get_entry_text (options->handler_header_entry));

  gtk_widget_destroy (GTK_WIDGET (options));
}


/* This returns a pointer to the text in the given entry, or NULL if the entry
   is empty. */
static gchar*
get_entry_text (GtkWidget *entry)
{
  gchar *text;

  text = gtk_entry_get_text (GTK_ENTRY (entry));
  if (text && text[0] != '\0')
    return text;
  else
    return NULL;
}
