/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <locale.h>

#include <gtk/gtk.h>

#include "gladeconfig.h"

#include "glade_project.h"
#include "gbwidget.h"
#include "save.h"
#include "utils.h"

#define GB_SAVE_BUFFER_INCREMENT 2048


static GladeStatusCode real_save_project_file (GladeProject *project);
static void save_named_style_callback (gchar * name,
				       GbStyle * gbstyle,
				       GbWidgetGetArgData * data);
static void save_named_style (gchar * name,
			      GbStyle * gbstyle,
			      GbWidgetGetArgData * data);
static void save_component (GtkWidget * item,
			    GbWidgetGetArgData * data);
static void save_indent (GbWidgetGetArgData * data);

static void save_buffer_add (GbBuffControl * data, gchar ch);
static void save_buffer_add_string (GbBuffControl * data, gchar * string);
static void save_buffer_add_chars (GbBuffControl * data, gchar * chars);
static void save_buffer_add_char (GbBuffControl * data, gchar ch);


/* We need this to make sure that numbers are output in a portable syntax,
   instead of using the current locale. This code is from glibc info docs. */
GladeStatusCode
save_project_file (GladeProject *project)
{
  gchar *old_locale, *saved_locale;
  GladeStatusCode status;
     
  old_locale = setlocale (LC_NUMERIC, NULL);
  saved_locale = g_strdup (old_locale);
  setlocale (LC_NUMERIC, "C");
  status = real_save_project_file (project);
  setlocale (LC_NUMERIC, saved_locale);
  g_free (saved_locale);
  return status;
}


static GladeStatusCode
real_save_project_file (GladeProject *project)
{
  gint bytes_written;
  GbWidgetGetArgData data;

  MSG ("Saving project");

  data.project = project;
  data.status = GLADE_STATUS_OK;
  data.filename = project->xml_filename;

  data.fp = fopen (data.filename, "w");
  if (data.fp == NULL)
    return GLADE_STATUS_FILE_OPEN_ERROR;

  gb_init_buffer_struct(0, GB_SAVE_BUFFER_INCREMENT, &data.buffer);

  data.indent = 0;

  fprintf (data.fp, "<?xml version=\"1.0\"?>\n<GTK-Interface>\n\n");

  /* Save the general project information. */
  glade_project_save_options (data.project, &data);
  bytes_written = fwrite (data.buffer.ptr, sizeof (gchar), data.buffer.pos,
                          data.fp);
  if (bytes_written != data.buffer.pos)
    {
      MSG2 ("Bytes: %i Written: %i", data.buffer.pos, bytes_written);
      data.status = GLADE_STATUS_FILE_WRITE_ERROR;
    }

  if (data.status == GLADE_STATUS_OK)
    {
      /* Save default gbstyle first. */
      MSG ("Saving styles");
      save_named_style (gb_widget_default_gb_style->name,
			gb_widget_default_gb_style, &data);
      /* Now save all the other named styles. */
      g_hash_table_foreach (gb_style_hash, (GHFunc) save_named_style_callback,
			    &data);
    }

  if (data.status == GLADE_STATUS_OK)
    {
      MSG ("Saving components");
      glade_project_foreach_component (data.project,
				       (GtkCallback) save_component, &data);
    }

  if (data.status == GLADE_STATUS_OK)
    {
      fprintf (data.fp, "\n</GTK-Interface>\n");
    }

  g_free (data.buffer.ptr);

  fclose (data.fp);
  return data.status;
}


/* This outputs one char, converting <>&" to XML entities. */
static void
save_buffer_add (GbBuffControl * data, gchar ch)
{
  if (ch == '<')
    save_buffer_add_chars (data, "&lt;");
  else if (ch == '>')
    save_buffer_add_chars (data, "&gt;");
  else if (ch == '&')
    save_buffer_add_chars (data, "&amp;");
  else if (ch == '"')
    save_buffer_add_chars (data, "&quot;");
  else
    save_buffer_add_char (data, ch);
}


static void
save_buffer_add_string (GbBuffControl * data, gchar * string)
{
  while (*string)
    {
      save_buffer_add (data, *string);
      string++;
    }
}


static void
save_buffer_add_chars (GbBuffControl * data, gchar * chars)
{
  while (*chars)
    {
      save_buffer_add_char (data, *chars);
      chars++;
    }
}


static void
save_buffer_add_char (GbBuffControl * data, gchar ch)
{
  /* Expand the buffer if necessary. */
  if (data->pos == data->space)
    {
      data->space += GB_SAVE_BUFFER_INCREMENT;
      data->ptr = g_realloc (data->ptr, data->space);
    }
  data->ptr[data->pos++] = ch;
}


static void
save_named_style_callback (gchar * name, GbStyle * gbstyle,
			   GbWidgetGetArgData * data)
{
  /* We've already saved the default GbStyle, so skip it. */
  if (gbstyle == gb_widget_default_gb_style)
    return;
  save_named_style (name, gbstyle, data);
}


static void
save_named_style (gchar * name, GbStyle * gbstyle, GbWidgetGetArgData * data)
{
  gint bytes_written;
  gboolean save_all = FALSE;

  data->buffer.pos = 0;
  data->indent = 0;
  MSG1 ("Saving style: %s", name);
  /* If this is the default style, only save it if it is different to the
     GTK default style, and if it is make sure we save everything. */
  if (gbstyle == gb_widget_default_gb_style)
    {
      if (gbstyle->style == gtk_widget_get_default_style ())
	return;
      else
	save_all = TRUE;
    }
  gb_widget_save_style (gbstyle, data, save_all);
  bytes_written = fwrite (data->buffer.ptr, sizeof (gchar), data->buffer.pos,
			  data->fp);
  if (bytes_written != data->buffer.pos)
    {
      MSG2 ("Bytes: %i Written: %i", data->buffer.pos, bytes_written);
      data->status = GLADE_STATUS_FILE_WRITE_ERROR;
    }
}


static void
save_component (GtkWidget * component, GbWidgetGetArgData * data)
{
  gint bytes_written;

  if (GB_IS_COMPONENT (component))
    return;

  data->buffer.pos = 0;
  data->indent = 0;
  gb_widget_save (component, data);

  bytes_written = fwrite (data->buffer.ptr, sizeof (gchar), data->buffer.pos,
			  data->fp);
  if (bytes_written != data->buffer.pos)
    {
      MSG2 ("Bytes: %i Written: %i", data->buffer.pos, bytes_written);
      data->status = GLADE_STATUS_FILE_WRITE_ERROR;
    }
}


static void
save_indent (GbWidgetGetArgData * data)
{
  gint i;
  for (i = 0; i < (data->indent * 2) / 8; i++)
    save_buffer_add_char (&data->buffer, '\t');
  for (i = 0; i < (data->indent * 2) % 8; i++)
    save_buffer_add_char (&data->buffer, ' ');
}


void
save_tag (GbWidgetGetArgData * data, gchar * tag_name)
{
  gchar *tag_name_start = glade_util_find_start_of_tag_name (tag_name);
  save_indent (data);
  save_buffer_add_char (&data->buffer, '<');
  save_buffer_add_string (&data->buffer, tag_name_start);
  save_buffer_add_chars (&data->buffer, ">\n");
}


void
save_end_tag (GbWidgetGetArgData * data, gchar * tag_name)
{
  gchar *tag_name_start = glade_util_find_start_of_tag_name (tag_name);
  save_indent (data);
  save_buffer_add_chars (&data->buffer, "</");
  save_buffer_add_string (&data->buffer, tag_name_start);
  save_buffer_add_chars (&data->buffer, ">\n");
}


void
save_newline (GbBuffControl * data)
{
  save_buffer_add_char (data, '\n');
}




void
save_string (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  gchar *tag_name_start;

  if (tag_value == NULL)
    return;

  tag_name_start = glade_util_find_start_of_tag_name (tag_name);
  save_indent (data);
  save_buffer_add_char (&data->buffer, '<');
  save_buffer_add_string (&data->buffer, tag_name_start);
  save_buffer_add_char (&data->buffer, '>');
  save_buffer_add_string (&data->buffer, tag_value);
  save_buffer_add_chars (&data->buffer, "</");
  save_buffer_add_string (&data->buffer, tag_name_start);
  save_buffer_add_chars (&data->buffer, ">\n");
}


void
save_text (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  save_string (data, tag_name, tag_value);
}


void
save_int (GbWidgetGetArgData * data, gchar * tag_name, gint tag_value)
{
  gchar buf[32];
  sprintf (buf, "%i", tag_value);
  save_string (data, tag_name, buf);
}


void
save_float (GbWidgetGetArgData * data, gchar * tag_name, gfloat tag_value)
{
  gchar buf[32];
  sprintf (buf, "%g", tag_value);
  save_string (data, tag_name, buf);
}


void
save_bool (GbWidgetGetArgData * data, gchar * tag_name, gint tag_value)
{
  save_string (data, tag_name, tag_value ? "True" : "False");
}


void
save_choice (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  save_string (data, tag_name, tag_value);
}


void
save_combo (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  save_string (data, tag_name, tag_value);
}


/* Colors are saved as r,g,b where rgb are 0-255. */
void
save_color (GbWidgetGetArgData * data, gchar * tag_name, GdkColor * tag_value)
{
  gchar buf[32];
  guint16 red = tag_value->red / 256;
  guint16 green = tag_value->green / 256;
  guint16 blue = tag_value->blue / 256;
  sprintf (buf, "%i,%i,%i", red, green, blue);
  save_string (data, tag_name, buf);
}


void
save_bgpixmap (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  save_string (data, tag_name, tag_value);
}


void
save_dialog (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  save_string (data, tag_name, tag_value);
}


void
save_filename (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  save_string (data, tag_name, tag_value);
}


void
save_pixmap_filename (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  gchar *pixmaps_dir, *filename;
  gint pixmaps_dir_len;
  gboolean free_filename = FALSE;

  /* When saving the XML file, we need to convert to relative pathnames.
     If the pixmap file is already in the pixmaps directory, we use the
     relative path. If not we just use the basename of the pixmap file, since
     it will be copied to the pixmaps directory in
     glade_project_copy_all_pixmaps(). */
  if (data->filename)
    {
      pixmaps_dir = glade_project_get_pixmaps_directory (data->project);
      g_return_if_fail (pixmaps_dir != NULL);
      g_return_if_fail (pixmaps_dir[0] != '\0');

      pixmaps_dir_len = strlen (pixmaps_dir);

      if (tag_value == NULL || tag_value[0] == '\0')
	filename = NULL;
      else if (strncmp (pixmaps_dir, tag_value, pixmaps_dir_len)
	       || (pixmaps_dir[pixmaps_dir_len - 1] != G_DIR_SEPARATOR
		   && tag_value[0] != G_DIR_SEPARATOR))
	filename = g_basename (tag_value);
      else
	{
	  filename = glade_util_make_relative_path (pixmaps_dir, tag_value);
	  free_filename = TRUE;
	}
      save_string (data, tag_name, filename);
      if (free_filename)
	g_free (filename);
    }
  else
    {
      /* When saving to the clipboard, we just save the full path.
	 This means it will still work if we paste into other projects. */
      save_string (data, tag_name, tag_value);
    }
}


void
save_font (GbWidgetGetArgData * data, gchar * tag_name, gchar * tag_value)
{
  save_string (data, tag_name, tag_value);
}

/* This saves a date in the RFC1123 format (an update of RFC822),
   e.g. 'Sun, 06 Nov 1994 08:49:37 GMT'. */
void
save_date (GbWidgetGetArgData * data, gchar * tag_name, time_t *time)
{
  struct tm *t;
  gchar buffer[32];

  t = gmtime (time);
  sprintf (buffer, "%s, %02d %s %04d %02d:%02d:%02d GMT",
	   GladeDayNames[t->tm_wday], t->tm_mday, GladeMonthNames[t->tm_mon],
	   t->tm_year + 1900, t->tm_hour, t->tm_min, t->tm_sec);
  save_string (data, tag_name, buffer);
}
