/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include <gtk/gtk.h>

#include "gladeconfig.h"

#include "palette.h"
#include "property.h"
#include "gbwidget.h"
#include "glade.h"
#include "glade_project.h"
#ifdef GTK_HAVE_FEATURES_1_1_0
#include "glade_gtk11lib.h"
#else
#include "glade_gtk10lib.h"
#endif
#include "save.h"
#include "editor.h"
#include "load.h"
#include "source.h"
#include "utils.h"
#include "tree.h"

/*
 * The hash table associates a Gtk widget class name with a GbWidget struct,
 * which contains a table of functions used by the builder, e.g. creating a
 * new widget, saving the widget etc.
 */
GHashTable *gb_widget_table = NULL;

/* These are used to return what is in a table cell */
#define GB_CELL_EMPTY		1
#define GB_CELL_WIDGET		2
#define GB_CELL_PLACEHOLDER	3

/* Tooltips for the created widgets */
static GtkTooltips *gb_widget_tooltips;


/* These are used for outputting signal handler prototypes. */
#define GB_PARAM_INDENT		40
#define GB_PARAM_TYPE_WIDTH	16

static gchar *GbToolbarIconFilenameKey = "GbIconFilename";


static void on_widget_destroy (GtkWidget * widget,
			       gpointer data);
static void gb_widget_free_accelerators (GtkWidget * widget);
static void gb_widget_free_signals (GtkWidget * widget);

static void show_color_properties (GdkColor colors[],
				   gchar * name);
static void show_accelerators (GtkWidget * widget,
			       GbWidgetGetArgData * data);
static void show_signals (GtkWidget * widget,
			  GbWidgetGetArgData * data);

static void set_standard_properties (GtkWidget * widget,
				     GbWidgetSetArgData * data);
static void set_position_properties (GtkWidget * widget,
				     GbWidgetSetArgData * data);
static void set_special_child_properties (GtkWidget * widget,
					  GbWidgetSetArgData * data);
static void set_lang_specific_properties (GtkWidget * widget,
					  GbWidgetSetArgData * data);
static void apply_style (GtkWidget * widget,
			 GbWidgetSetArgData * data);
static gboolean apply_colors (GtkWidget * widget,
			      GbWidgetSetArgData * data,
			      GdkColor colors[],
			      GdkColor new_colors[],
			      gchar * name);
static void apply_accelerators (GtkWidget * widget,
				GbWidgetSetArgData * data);
static void apply_signals (GtkWidget * widget,
			   GbWidgetSetArgData * data);

static void add_standard_top_menu_items (GtkWidget * widget,
					 GbWidgetCreateMenuData * data);
static void add_standard_bottom_menu_items (GtkWidget * widget,
					    GbWidgetCreateMenuData * data);

static void gb_widget_load_properties (GbWidgetSetArgData * data);
static void load_special_child_properties (GbWidgetSetArgData * data);
static void load_accelerator (GbWidgetSetArgData * data);
static void load_signal (GbWidgetSetArgData * data);
static void load_style (GtkWidget * widget,
			GbWidgetSetArgData * data);
static gboolean load_colors (GbWidgetSetArgData * data,
			     GdkColor colors[],
			     gchar * name,
			     gchar * element,
			     gchar * cdata);

static void get_standard_properties (GtkWidget * widget,
				     GbWidgetGetArgData * data);

static void get_position_properties (GtkWidget * widget,
				     GbWidgetGetArgData * data);
static void get_special_child_properties (GtkWidget * widget,
					  GbWidgetGetArgData * data);
static void get_lang_specific_properties (GtkWidget * widget,
					  GbWidgetGetArgData * data);

static void save_style (GtkWidget * widget,
			GbWidgetGetArgData * data);
static void save_colors (GbWidgetGetArgData * data,
			 gboolean save_all,
			 GdkColor colors[],
			 GdkColor default_colors[],
			 gchar * name);
static void save_accelerators (GtkWidget * widget,
			       GbWidgetGetArgData * data);
static gchar* create_modifiers_string (guint8 modifiers);
static void save_signals (GtkWidget * widget,
			  GbWidgetGetArgData * data);


static void gb_widget_write_signals_source (GtkWidget * widget,
					    GbWidgetWriteSourceData * data);
static gchar *get_type_name (GtkType type,
			     gboolean * is_pointer);
static gchar *get_gdk_event (gchar * signal_name);
static gchar **lookup_signal_arg_names (gchar * type,
					gchar * signal_name);
static void gb_widget_write_accelerators_source (GtkWidget * widget,
					    GbWidgetWriteSourceData * data);

static void gb_widget_add_alignment (GtkWidget * menuitem,
				     GtkWidget * widget);
static void gb_widget_remove_alignment (GtkWidget * menuitem,
					GtkWidget * widget);
static void gb_widget_add_event_box (GtkWidget * menuitem,
				     GtkWidget * widget);
static void gb_widget_remove_event_box (GtkWidget * menuitem,
					GtkWidget * widget);


static void split_table_placeholder (GtkWidget * table,
				     GtkWidget * placeholder,
				     gint left,
				     gint right,
				     gint top,
				     gint bottom,
				     gint skip_row,
				     gint skip_col);
static gint is_cell_occupied (GtkWidget * table,
			      gint row,
			      gint col);
static void remove_table_placeholders (GtkWidget * table,
				       gint row,
				       gint col);

static void gb_widget_convert_toolbar_button (GtkWidget * menuitem,
					      GtkWidget * widget);
static GtkWidget* gb_widget_get_toolbar_radio_group_widget (GtkWidget *widget,
							    gboolean skip_widget,
							    gboolean find_default_group);
static void gb_widget_check_toolbar_radio_group (GtkWidget * widget);

static void table_foreach (GtkTable * table,
			   GtkCallback callback,
			   gpointer callback_data);
static void notebook_tabs_foreach (GtkNotebook *notebook,
				   GtkCallback callback,
				   gpointer callback_data);

static GtkNotebookPage *find_notebook_page (GtkWidget * widget,
					    GtkWidget * current_child,
					    gint * pos);
static void update_box_size_callback (GtkWidget * widget,
				      gint data[2]);



/*************************************************************************
 * Initialization functions
 *************************************************************************/
void
gb_free_load_properties (GbLoadPropControl *p)
{
  p->n = 0;
  p->space = 0;
  if (p->ptr)
     g_free (p->ptr);
  p->ptr = NULL;
}

void
gb_init_load_properties (GbLoadPropControl *p)
{
  p->n = 0;
  p->space = GB_LOADED_PROPERTIES_INC;
  p->ptr = g_new (GbLoadedProperty, GB_LOADED_PROPERTIES_INC);
}

void
gb_init_buffer_struct (int pos, int space, GbBuffControl * p)
{
  p->space = space;
  p->ptr = g_new (gchar, p->space);
  if (p->ptr)
  {
    p->ptr[0] = '\0';
  }
  p->pos = pos;  
}

void
gb_widgets_init ()
{
  gint index, j;
  GladeWidgetInitData *gwid;
  GladePaletteSectionData *palsec;
  GbWidget *(*init_func) ();
  GbWidget *gbwidget;

  gb_widget_table = g_hash_table_new (g_str_hash, g_str_equal);

  gb_widget_reset_gb_styles ();

  /* Create tooltips */
  gb_widget_tooltips = gtk_tooltips_new ();

  index = 0;
  while (1)
    {
      j = 0;
      palsec = &sections[index];
      index++;
      if (!palsec->section)
        break;
      while (1)
        {
          gwid = &palsec->widgets[j];
          j++;
          if (!gwid->name)
            break;  
          init_func = gwid->init_func;
          gbwidget = (*init_func) ();
          g_hash_table_insert (gb_widget_table, gwid->name, gbwidget);
        }
    }
}


void
gb_widget_init_struct (GbWidget * gbwidget)
{
  gbwidget->properties_page_number = GB_PROPERTIES_NOT_CREATED;
  gbwidget->pixmap_struct = NULL;
  gbwidget->tooltip = NULL;

  gbwidget->gb_widget_new = NULL;
  gbwidget->gb_widget_add_child = NULL;
  gbwidget->gb_widget_create_properties = NULL;
  gbwidget->gb_widget_get_properties = NULL;
  gbwidget->gb_widget_set_properties = NULL;
  gbwidget->gb_widget_create_popup_menu = NULL;
  gbwidget->gb_widget_write_source = NULL;
  gbwidget->gb_widget_destroy = NULL;
  gbwidget->gb_widget_get_child = NULL;
}

void
gb_widget_init_struct2 (gchar *tooltip,gchar **xpm, GbWidget * gbwidget)
{
   gb_widget_init_struct (gbwidget);
   /* Fill in the pixmap struct & tooltip */
   gbwidget->pixmap_struct = xpm;
   gbwidget->tooltip = tooltip;
}

/*************************************************************************
 * Functions for creating & destroying GbWidgets
 *************************************************************************/

GtkWidget *
gb_widget_new (gchar * class_name, GtkWidget * parent)
{
  return gb_widget_new_full (class_name, TRUE, parent, NULL, 0, 0, NULL,
			     GB_CREATING, NULL);
}


/* Creates a new widget.
 * class_name is the name of the widget class, e.g. 'GtkLabel'
 * create_default_name is TRUE if you want a default name to be created,
 *   e.g. 'label1'.
 * parent is the widget that the new widget will be added beneath, so that
 *   the callback function knows where to put the widget.
 * current_child is the widget that the new widget will replace, or NULL
 *   if the new widget is just being added. It is used when replacing
 *   placeholders.
 * x & y are the coordinates of the new widget if it is being added to a
 *   GtkFixed container.
 * callback is the function to call once the widget is created to actually
 *   add it to the parent. Some widgets require dialog boxes for creating
 *   them (e.g. the dialog box to set the number of rows/cols in a table).
 *   So we need to provide a function to be called after this is done
 *   (we could have possibly chosen to use modal dialogs instead.)
 *   It is also currently used for adding the widget to the tree view, though
 *   I'm not sure it should be done this way.
 * action is either GB_CREATING or GB_LOADING. When loading widgets we don't
 *   want dialog boxes to pop up when they are being created.
 * loading_data is set when action is GB_LOADING, and contains the data used
 *   while loading, so that the GbWidgets can get any properties they need to
 *   create the widget without popping up a dialog box.
 */
GtkWidget *
gb_widget_new_full (gchar * class_name, gboolean create_default_name,
		    GtkWidget * parent, GtkWidget * current_child,
		    gint x, gint y, GbWidgetNewCallback callback,
		    GbWidgetAction action, GbWidgetSetArgData * loading_data)
{
  GbWidgetNewData *data;
  GtkWidget *new_widget;
  GbWidget *gbwidget;
  gint type;

  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
  g_return_val_if_fail (gbwidget != NULL, NULL);

  type = gtk_type_from_name (class_name);
  g_return_val_if_fail (type != 0, NULL);

  data = g_new (GbWidgetNewData, 1);
  /* Don't set data->name to NULL, since many widgets use it to set the label
     of the new widget. */
  data->project = current_project;
  data->name = create_default_name ? glade_project_new_widget_name (data->project, class_name) : g_strdup ("");
  data->type = type;
  data->callback = callback;
  data->parent = parent;
  if (parent)
    gtk_widget_ref (parent);
  data->current_child = current_child;
  if (current_child)
    gtk_widget_ref (current_child);
  data->x = x;
  data->y = y;
  data->widget_data = gb_widget_new_widget_data ();
  data->action = action;
  data->loading_data = loading_data;

  if (gbwidget->gb_widget_new)
    new_widget = (gbwidget->gb_widget_new) (data);
  else
    new_widget = gtk_type_new (data->type);

  /* If the widget has been created immediately, then we can finish it off now,
     and free the GbWidgetNewData struct, otherwise we leave that to the
     dialog. */
  if (new_widget)
    {
      gb_widget_initialize (new_widget, data);
      if (data->callback)
	(*data->callback) (new_widget, data);
      gb_widget_free_new_data (data);
    }

  return new_widget;
}


/* This turns a normal widget into a GbWidget, adding a GbWidgetData struct
   and the necessary signals. The widget must have its parent set already. The
   name will have a unique ID added on to it, e.g. "ok_button1". */
void
gb_widget_create_from (GtkWidget * widget,
		       gchar * name)
{
  GbWidgetData *wdata;

  MSG ("In create_from");
  gtk_widget_set_name (widget, glade_project_new_widget_name (current_project,
							      name));
  wdata = gb_widget_new_widget_data ();
  gb_widget_real_initialize (widget, wdata);
}


GbWidgetData *
gb_widget_new_widget_data ()
{
  GbWidgetData *widget_data = g_new (GbWidgetData, 1);
  widget_data->flags = GB_VISIBLE | GB_SENSITIVE | GB_STYLE_IS_UNNAMED
    | GB_STYLE_PROPAGATE | GB_INITIAL_EXPOSE;
  widget_data->x = 0;
  widget_data->y = 0;
  widget_data->width = 0;
  widget_data->height = 0;
  widget_data->events = 0;
  widget_data->tooltip = NULL;
  widget_data->signals = NULL;
  widget_data->accelerators = NULL;
  widget_data->gbstyle = gb_widget_default_gb_style;
  gb_widget_ref_gb_style (gb_widget_default_gb_style);

  /* C options. */
  widget_data->source_file = NULL;
  widget_data->public_field = 1;

  /* C++ options. */
  widget_data->separate_file = 0;
  widget_data->use_heap = 0;
  widget_data->flatten_hierarchy = 0;

  return widget_data;
}


void
gb_widget_initialize (GtkWidget * widget, GbWidgetNewData * data)
{
  if (data->name && data->name[0] != '\0')
    gtk_widget_set_name (widget, data->name);
  gb_widget_real_initialize (widget, data->widget_data);

  /* Now we set the widget's real style */
  /* FIXME: check if style should be propagated down from an ancestor? */
  if (widget->style != data->widget_data->gbstyle->style)
    {
      gtk_widget_set_style (widget, data->widget_data->gbstyle->style);
    }

  /* FIXME: GTK workarounds to make sure that some widgets have reasonable
     sizes initially. Quite a few widgets default to a width and height of 0,
     which means that if there is little space available they will disappear,
     so we may need to do more here. */
  if (GTK_IS_ARROW (widget))
    gtk_widget_set_usize (widget, 16, 16);

  /* Set this to NULL so we don't try to free it later. */
  data->widget_data = NULL;
}



void
gb_widget_real_initialize (GtkWidget * widget, GbWidgetData * wdata)
{
  gtk_object_set_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY, wdata);
  gtk_signal_connect (GTK_OBJECT (widget), "destroy",
		      GTK_SIGNAL_FUNC (on_widget_destroy), &widget);
  /* We don't add signals to menu items. */
  if (!GTK_IS_MENU_ITEM (widget))
    {
      editor_add_mouse_signals (widget);
      editor_add_draw_signals (widget);
      if (GTK_IS_WINDOW (widget))
	editor_add_key_signals (widget);
    }
  if (!GTK_IS_WINDOW (widget) && !GTK_IS_MENU (widget))
    gtk_widget_show (widget);
}


/* This returns TRUE if it is OK to complete the new() procedure, i.e. that
   the widget to replace or the parent widget still exist. It is used after
   the OK button is pressed in the dialog boxes for creating new tables/boxes.
   FIXME: I'm not too sure what we should do here. */
gboolean
gb_widget_can_finish_new (GbWidgetNewData * data)
{
  if (data->current_child)
    {
      if (!GTK_OBJECT_DESTROYED (data->current_child)
	  && data->current_child->parent)
	return TRUE;
    }
  else if (data->parent)
    {
      if (!GTK_OBJECT_DESTROYED (data->parent)
	  && (GTK_IS_WINDOW (data->parent) || data->parent->parent))
	return TRUE;
    }
  return FALSE;
}


void
gb_widget_free_new_data (GbWidgetNewData * data)
{
  g_free (data->name);
  g_free (data->widget_data);
  if (data->parent)
    gtk_widget_unref (data->parent);
  if (data->current_child)
    gtk_widget_unref (data->current_child);
  g_free (data);
}


static void
on_widget_destroy (GtkWidget * widget, gpointer user_data)
{
  GbWidget *gbwidget;
  GbWidgetData *widget_data;
  GbWidgetDestroyData data;
  gchar *class_name;

  MSG1 ("IN on_widget_destroy widget:%s", gtk_widget_get_name (widget));

  /* If the entire project is being destroyed, we don't need to update the
     selection or the widget tree. */
  if (!GTK_OBJECT_DESTROYED (current_project))
    {
      editor_remove_widget_from_selection (widget);
#ifdef GTK_HAVE_FEATURES_1_1_0
      tree_remove_widget (widget);
#endif
    }

  class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
  g_return_if_fail (gbwidget != NULL);

  /* Call the GbWidget destroy function, if it has one. */
  data.project = current_project;
  if (gbwidget->gb_widget_destroy)
    (gbwidget->gb_widget_destroy) (widget, &data);

  widget_data = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  g_free (widget_data->tooltip);
  gb_widget_free_accelerators (widget);
  gb_widget_free_signals (widget);
  g_free (widget_data->source_file);
  g_free (widget_data);
  g_free (gtk_object_get_data (GTK_OBJECT (widget), GB_CHILD_NAME_KEY));
  MSG1 ("OUT on_widget_destroy widget:%s", gtk_widget_get_name (widget));
}


static void
gb_widget_free_accelerators (GtkWidget * widget)
{
  GbWidgetData *widget_data;
  GList *item;
  GbAccelerator *accelerator;

  widget_data = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  if (widget_data->accelerators)
    {
      item = widget_data->accelerators;
      while (item)
	{
	  accelerator = (GbAccelerator *) item->data;
	  g_free (accelerator->key);
	  g_free (accelerator->signal);
	  item = item->next;
	}
      g_list_free (widget_data->accelerators);
      widget_data->accelerators = NULL;
    }
}


static void
gb_widget_free_signals (GtkWidget * widget)
{
  GbWidgetData *widget_data;
  GList *item;
  GbSignal *signal;

  widget_data = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  if (widget_data->signals)
    {
      item = widget_data->signals;
      while (item)
	{
	  signal = (GbSignal *) item->data;
	  g_free (signal->name);
	  g_free (signal->handler);
	  g_free (signal->object);
	  g_free (signal->data);
	  item = item->next;
	}
      g_list_free (widget_data->signals);
      widget_data->signals = NULL;
    }
}


/*************************************************************************
 * Functions for creating the page of properties specific to this widget
 *************************************************************************/

/* Returns the page number of the new page in the notebook which contains the
   widget's specific properties, or GB_PROPERTIES_NOT_NEEDED if it has none. */
gint
gb_widget_create_properties (GtkWidget * widget)
{
  GtkWidget *page;
  gint page_number;
  GbWidgetCreateArgData data;
  GbWidget *gbwidget;
  gchar *class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);

  g_return_val_if_fail (gbwidget != NULL, GB_PROPERTIES_NOT_NEEDED);
  if (gbwidget->gb_widget_create_properties || GTK_IS_CONTAINER (widget))
    {
      /* Create skeleton of properties page, so gbwidget just has to add
         properties */
      page = gtk_table_new (10, 3, FALSE);
      gtk_widget_show (page);
      page_number = property_add_gbwidget_page (page);
      property_set_table_position (page, 0);

      /* If widget is a container add a border width property */
      if (GTK_IS_CONTAINER (widget))
	{
	  gchar buf[128];
	  sprintf (buf, "%s::border_width", class_name);
	  property_add_int_range (buf, _("Border Width:"),
			     _("The width of the border around the container"),
				  0, 1000, 1, 10, 1);
	}

      data.project = current_project;
      if (gbwidget->gb_widget_create_properties)
	(gbwidget->gb_widget_create_properties) (widget, NULL, &data);
      return page_number;
    }
  else
    {
      return GB_PROPERTIES_NOT_NEEDED;
    }
}


/*************************************************************************
 * Functions for showing the widget's properties
 *************************************************************************/

void
gb_widget_show_properties (GtkWidget * widget)
{
  GbWidgetGetArgData data;
  gchar *class_name;
  GbWidget *gbwidget;
  GbWidgetData *widget_data;
  gint page;

  /* If properties of widget are already shown, just return */
  if (property_get_widget () == widget)
    return;

  /* If widget is a placeholder reset the properties window and return */
  if (GB_IS_PLACEHOLDER (widget))
    {
      property_set_widget (NULL);
      return;
    }

  class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
  g_return_if_fail (gbwidget != NULL);

  /* Turn off auto-apply so we can set properties without the 'changed'
     callbacks calling gb_widget_apply_properties (). */
  property_set_auto_apply (FALSE);

  /* Need this here to make sure properties notebook is sensitive */
  property_set_widget (widget);

  page = gbwidget->properties_page_number;
  /* If widget's properties page hasn't been created, create it now */
  if (page == GB_PROPERTIES_NOT_CREATED)
    {
      page = gb_widget_create_properties (widget);
      gbwidget->properties_page_number = page;
    }

  /* Show the widget's own properties page if it has one.
     Need to show the page before setting properties because of the
     Text widget - it must be realized before setting the text :-( */
  if (page == GB_PROPERTIES_NOT_NEEDED)
    property_hide_gbwidget_page ();
  else
    property_show_gbwidget_page (page);

  widget_data = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  g_return_if_fail (widget_data != NULL);
  data.project = current_project;
  data.action = GB_SHOWING;
  data.widget_data = widget_data;
  get_standard_properties (widget, &data);

  if (gbwidget->gb_widget_get_properties)
    (gbwidget->gb_widget_get_properties) (widget, NULL, &data);

  /* Turn auto-apply back on again */
  property_set_auto_apply (TRUE);
}


/* This is called when the widget's size or position has changed, so that we
   should update the values shown in the properties editor. */
void
gb_widget_show_position_properties (GtkWidget * widget)
{
  GbWidgetGetArgData data;
  GbWidgetData *widget_data;

  /* Make sure this is the widget shown in the properties editor. */
  if (property_get_widget () != widget)
    return;

  widget_data = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  g_return_if_fail (widget_data != NULL);
  data.project = current_project;
  data.action = GB_SHOWING;
  data.widget_data = widget_data;

  property_set_auto_apply (FALSE);
  get_position_properties (widget, &data);
  property_set_auto_apply (TRUE);
}



void
gb_widget_show_style (GtkWidget * widget)
{
  GbWidgetData *wdata;
  GbStyle *gbstyle;
  GtkStyle *style = widget->style;
  gchar buffer[128];
  gint i;

  wdata = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  g_return_if_fail (wdata != NULL);
  gbstyle = wdata->gbstyle;

  property_set_bool (GbStylePropagate, wdata->flags & GB_STYLE_PROPAGATE);

  property_set_dialog (GbStyleName, wdata->flags & GB_STYLE_IS_UNNAMED ?
		       "" : gbstyle->name, NULL);
  property_set_font (GbStyleFont, style->font, gbstyle->xlfd_fontname);

  /* Colors */
  show_color_properties (style->fg, "fg");
  show_color_properties (style->bg, "bg");
  show_color_properties (style->text, "text");
  show_color_properties (style->base, "base");

  /* Background pixmaps */
  for (i = 0; i < GB_NUM_STYLE_STATES; i++)
    {
      sprintf (buffer, "GtkStyle::%s[%s]", GbBgPixmapName, GbStateNames[i]);
      property_set_bgpixmap (buffer, style->bg_pixmap[i],
			     gbstyle->bg_pixmap_filenames[i]);
    }
}


static void
show_color_properties (GdkColor colors[], gchar * name)
{
  GdkColor gdkcolor;
  guchar color[4];
  gint state;
  gchar buf[128];

  for (state = 0; state < GB_NUM_STYLE_STATES; state++)
    {
      sprintf (buf, "GtkStyle::%s[%s]", name, GbStateNames[state]);
      gdkcolor = colors[state];
      color[0] = gdkcolor.red >> 8;
      color[1] = gdkcolor.green >> 8;
      color[2] = gdkcolor.blue >> 8;
      property_set_color (buf, color);
    }
}


static void
show_accelerators (GtkWidget * widget, GbWidgetGetArgData * data)
{
  GList *element = data->widget_data->accelerators;
  property_clear_accelerators ();
  while (element)
    {
      property_add_accelerator ((GbAccelerator *) element->data);
      element = element->next;
    }
}


static void
show_signals (GtkWidget * widget, GbWidgetGetArgData * data)
{
  GList *element = data->widget_data->signals;
  property_clear_signals ();
  MSG1 ("Num signals: %i", g_list_length (element));
  while (element)
    {
      property_add_signal ((GbSignal *) element->data);
      element = element->next;
    }
}



/*************************************************************************
 * Functions for applying properties to a widget
 *************************************************************************/

void
gb_widget_apply_properties (GtkWidget * widget, GtkWidget * property)
{
  GbWidgetSetArgData data;
  GbWidget *gbwidget;
  GbWidgetData *widget_data;
  gchar *class_name;

  MSG1 ("Applying properties: %s", gtk_widget_get_name (widget));

  class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
  g_return_if_fail (gbwidget != NULL);

  widget_data = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  g_return_if_fail (widget_data != NULL);
  data.project = current_project;
  data.action = GB_APPLYING;
  data.widget_data = widget_data;
  data.property_to_apply = property;
  set_standard_properties (widget, &data);

  MSG ("Calling widget's own apply_properties");
  if (gbwidget->gb_widget_set_properties)
    (gbwidget->gb_widget_set_properties) (widget, NULL, &data);
  MSG ("Called widget's own apply_properties");
}


static void
set_standard_properties (GtkWidget * widget, GbWidgetSetArgData * data)
{
  GbWidgetData *wdata = data->widget_data;
  gchar *name, *tooltip, *events_str, *ext_events;
  gboolean visible, sensitive, can_default, has_default, can_focus, has_focus;
  gint events, i;

  /* Properties on widget page */
  /* When pasting, we may want to discard the names from the XML, and generate
     a new name instead. */
  if (data->action == GB_LOADING && data->xml_buffer && data->discard_names)
    {
      gchar *class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
      name = glade_project_new_widget_name (data->project, class_name);
      gtk_widget_set_name (widget, name);
    }
  else
    {
      name = gb_widget_input_string (data, GbName);
      if (data->apply)
	{
#ifdef GTK_HAVE_FEATURES_1_1_0
	  /* If we are loading, the name will be set when the widget is added
	     to the tree, so we don't need to set it here. */
	  if (data->action != GB_LOADING)
	    tree_rename_widget (widget, name);
#endif
	  gtk_widget_set_name (widget, name);
	  /* If widget is a toplevel window/dialog set the component's name in
	     the project window */
	  if (widget->parent == NULL)
	    glade_project_component_changed (data->project, widget);

	  /* If we are loading, we have to check if the name has a trailing ID
	     and if so we reserve it so no other widget can use it. */
	  if (data->action == GB_LOADING)
	    glade_project_reserve_name (data->project, name);
	}
    }

  /* If widget is a container, show the border width */
  if (GTK_IS_CONTAINER (widget))
    {
      gchar buf[128];
      gint border_width;
      gchar *class_name;

      class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
      sprintf (buf, "%s::border_width", class_name);
      border_width = gb_widget_input_int (data, buf);
      if (data->apply && GTK_CONTAINER (widget)->border_width != border_width)
	{
	  gtk_container_border_width (GTK_CONTAINER (widget), border_width);
	  /*if (data->action == GB_APPLYING) */
	  editor_refresh_widget (widget);
	}
    }

  /* Language-specific properties. */
  set_lang_specific_properties (widget, data);

  /* Special child properties page */
  set_special_child_properties (widget, data);

  /* Properties on standard page */
  set_position_properties (widget, data);

  visible = gb_widget_input_bool (data, GbVisible);
  if (data->apply)
    {
      if (visible)
	wdata->flags |= GB_VISIBLE;
      else
	wdata->flags &= ~GB_VISIBLE;
    }

  sensitive = gb_widget_input_bool (data, GbSensitive);
  if (data->apply)
    {
      if (sensitive)
	wdata->flags |= GB_SENSITIVE;
      else
	wdata->flags &= ~GB_SENSITIVE;
    }

  tooltip = gb_widget_input_string (data, GbTooltip);
  if (data->apply)
    {
      g_free (wdata->tooltip);
      if (tooltip && tooltip[0] == '\0')
	tooltip = NULL;
      wdata->tooltip = g_strdup (tooltip);

      /* SPECIAL CODE: toolbar children's tooltips use the toolbar's tooltips
	 object rather than our tooltip object. */
      if (widget->parent && GTK_IS_TOOLBAR (widget->parent))
	gtk_tooltips_set_tip (GTK_TOOLBAR (widget->parent)->tooltips, 
			      widget, tooltip, NULL);
      else
	gtk_tooltips_set_tip (gb_widget_tooltips, widget, tooltip, NULL);
    }

  can_default = gb_widget_input_bool (data, GbCanDefault);
  if (data->apply)
    {
      if (can_default)
	GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_DEFAULT);
      else
	GTK_WIDGET_UNSET_FLAGS (widget, GTK_CAN_DEFAULT);
      if (data->action == GB_APPLYING)
	editor_refresh_widget (widget);
    }

  has_default = gb_widget_input_bool (data, GbHasDefault);
  if (data->apply)
    {
      if (has_default)
	wdata->flags |= GB_GRAB_DEFAULT;
      else
	wdata->flags &= ~GB_GRAB_DEFAULT;
    }

  /* Different widgets have different default values for CAN_FOCUS, so when
     we load an XML file we must make sure that we always set it or unset it.
     Also, since we don't save the can_focus flag if it is false, we must make
     sure that we apply it anyway when loading. */
  can_focus = gb_widget_input_bool (data, GbCanFocus);
  if (!data->apply)
    can_focus = FALSE;
  if (data->apply || data->action == GB_LOADING)
    {
      if (can_focus)
	GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_FOCUS);
      else
	GTK_WIDGET_UNSET_FLAGS (widget, GTK_CAN_FOCUS);
    }
  has_focus = gb_widget_input_bool (data, GbHasFocus);
  if (data->apply)
    {
      if (has_focus)
	wdata->flags |= GB_GRAB_FOCUS;
      else
	wdata->flags &= ~GB_GRAB_FOCUS;
    }

  /* Events & ext events. */
  if (!GTK_WIDGET_NO_WINDOW (widget))
    {
      if (data->action == GB_APPLYING)
	{
	  events = property_events_string_to_value (gb_widget_input_string (data,
								 GbEvents));
	  if (data->apply)
	    wdata->events = events;
	}
      else
	{
	  events_str = gb_widget_input_string (data, GbEvents);
	  if (data->apply)
	    {
	      for (i = 0; i < GB_EVENT_MASKS_COUNT; i++)
		{
		  if (glade_util_strstr (events_str, GbEventMaskSymbols[i]))
		    wdata->events |= GbEventMaskValues[i];
		}
	    }
	}

      ext_events = gb_widget_input_choice (data, GbExtEvents);
      if (data->apply)
	{
	  for (i = 0; GbExtensionModeChoices[i]; i++)
	    {
	      if (!strcmp (ext_events, GbExtensionModeChoices[i])
		  || !strcmp (ext_events, GbExtensionModeSymbols[i]))
		{
		  gtk_widget_set_extension_events (widget, GbExtensionModeValues
						   [i]);
		  break;
		}
	    }
	}
    }

  if (data->action == GB_APPLYING)
    {
      apply_style (widget, data);
      apply_accelerators (widget, data);
      apply_signals (widget, data);
    }
  else
    {
      load_style (widget, data);
      data->widget_data->signals = data->signals;
      data->widget_data->accelerators = data->accelerators;
    }
}


static void
set_special_child_properties (GtkWidget * widget,
			      GbWidgetSetArgData * data)
{
  GtkWidget *parent = widget->parent;

  if (!parent)
    return;

  /* When pasting a widget to replace an existing widget, the child properties
     will already have been set, so we just return. */
  if (data->action == GB_LOADING && data->xml_buffer && data->replacing_widget)
    return;

  /* Tell the load functions to use the child properties array. */
  data->loading_type = GB_CHILD_PROPERTIES;

  if (GTK_IS_TABLE (parent))
    {
      GtkTableChild *tchild = glade_util_find_table_child (GTK_TABLE (parent),
							   widget);
      gint xpad, ypad, left_attach, right_attach, top_attach, bottom_attach;
      gint old_left_attach, old_right_attach, old_top_attach, old_bottom_attach;
      gint xexpand, yexpand, xshrink, yshrink, xfill, yfill;
      gboolean myApply;

      g_return_if_fail (tchild != NULL);
      old_left_attach = tchild->left_attach;
      old_right_attach = tchild->right_attach;
      old_top_attach = tchild->top_attach;
      old_bottom_attach = tchild->bottom_attach;

      xpad = gb_widget_input_int (data, GbXPad);
      myApply = data->apply;
      ypad = gb_widget_input_int (data, GbYPad);
      myApply |= data->apply;
      xexpand = (gb_widget_input_bool (data, GbXExpand)) ? GTK_EXPAND : 0;
      myApply |= data->apply;
      yexpand = (gb_widget_input_bool (data, GbYExpand)) ? GTK_EXPAND : 0;
      myApply |= data->apply;
      xshrink = (gb_widget_input_bool (data, GbXShrink)) ? GTK_SHRINK : 0;
      myApply |= data->apply;
      yshrink = (gb_widget_input_bool (data, GbYShrink)) ? GTK_SHRINK : 0;
      myApply |= data->apply;
      xfill = (gb_widget_input_bool (data, GbXFill)) ? GTK_FILL : 0;
      myApply |= data->apply;
      yfill = (gb_widget_input_bool (data, GbYFill)) ? GTK_FILL : 0;
      myApply |= data->apply;

      if (data->action == GB_APPLYING)
	{
	  left_attach = gb_widget_input_int (data, GbCellX);
	  myApply |= data->apply;
	  top_attach = gb_widget_input_int (data, GbCellY);
	  myApply |= data->apply;
	  right_attach = gb_widget_input_int (data, GbColSpan) + left_attach;
	  myApply |= data->apply;
	  bottom_attach = gb_widget_input_int (data, GbRowSpan) + top_attach;
	  myApply |= data->apply;
	}
      else
	{
	  left_attach = gb_widget_input_int (data, "GtkWidget::left_attach");
	  myApply |= data->apply;
	  top_attach = gb_widget_input_int (data, "GtkWidget::top_attach");
	  myApply |= data->apply;
	  right_attach = gb_widget_input_int (data, "GtkWidget::right_attach");
	  myApply |= data->apply;
	  bottom_attach = gb_widget_input_int (data, "GtkWidget::bottom_attach");
	  myApply |= data->apply;
	  if (right_attach <= left_attach)
	    right_attach = left_attach + 1;
	  if (bottom_attach <= top_attach)
	    bottom_attach = top_attach + 1;
	}

      if (myApply)
	{
	  if (xpad != tchild->xpadding || ypad != tchild->ypadding
	   || (xexpand && !tchild->xexpand) || (!xexpand && tchild->xexpand)
	   || (yexpand && !tchild->yexpand) || (!yexpand && tchild->yexpand)
	   || (xshrink && !tchild->xshrink) || (!xshrink && tchild->xshrink)
	   || (yshrink && !tchild->yshrink) || (!yshrink && tchild->yshrink)
	      || (xfill && !tchild->xfill) || (!xfill && tchild->xfill)
	      || (yfill && !tchild->yfill) || (!yfill && tchild->yfill)
	      || left_attach != tchild->left_attach
	      || right_attach != tchild->right_attach
	      || top_attach != tchild->top_attach
	      || bottom_attach != tchild->bottom_attach)
	    {

#ifndef GTK_HAVE_FEATURES_1_1_4
	      gtk_container_block_resize (GTK_CONTAINER (parent));
#endif
	      gtk_widget_ref (widget);
	      gtk_container_remove (GTK_CONTAINER (parent), widget);
	      gtk_table_attach (GTK_TABLE (parent), widget,
		       left_attach, right_attach, top_attach, bottom_attach,
		       xexpand | xshrink | xfill, yexpand | yshrink | yfill,
				xpad, ypad);
	      gtk_widget_unref (widget);
	      if (data->action == GB_APPLYING)
		gb_widget_update_table_placeholders (parent, -1, -1);
#ifndef GTK_HAVE_FEATURES_1_1_4
	      gtk_container_unblock_resize (GTK_CONTAINER (parent));
#endif
	    }
	}
    }
  else if (GTK_IS_BOX (parent) && !GTK_IS_BUTTON_BOX (parent))
    {
      gint padding;
      gboolean expand, fill, pack, myApply;

      padding = gb_widget_input_int (data, GbPadding);
      myApply = data->apply;
      expand = gb_widget_input_bool (data, GbExpand);
      myApply |= data->apply;
      fill = gb_widget_input_bool (data, GbFill);
      myApply |= data->apply;

      if (data->action == GB_APPLYING)
	{
	  pack = gb_widget_input_bool (data, GbPack);
	}
      else
	{
	  gchar *pack_symbol = gb_widget_input_string (data, GbPack);
	  pack = (!data->apply || (!strcmp (pack_symbol, "GTK_PACK_START")));
	}
      myApply |= data->apply;
      if (myApply)
	gtk_box_set_child_packing (GTK_BOX (parent), widget, expand, fill, padding,
				   pack ? GTK_PACK_START : GTK_PACK_END);
    }

  else if (GTK_IS_TOOLBAR (parent))
    {
      gboolean new_group = gb_widget_input_bool (data, GbNewToolbarGroup);
      if (data->apply)
	gb_widget_set_new_toolbar_group (parent, widget, new_group);
    }

#ifdef GTK_HAVE_FEATURES_1_1_0
  else if (GTK_IS_PACKER (parent))
    {
      gchar *side_str, *anchor_str;
      GtkSideType side = GTK_SIDE_TOP;
      GtkAnchorType anchor = GTK_ANCHOR_CENTER;
      gboolean expand, fillx, filly, myApply;
      gint border, padx, pady, ipadx, ipady, i;

      side_str = gb_widget_input_choice (data, GbPackerSide);
      myApply = data->apply;
      anchor_str = gb_widget_input_choice (data, GbPackerAnchor);
      myApply |= data->apply;
      expand = (gb_widget_input_bool (data, GbPackerExpand)) ? GTK_PACK_EXPAND : 0;
      myApply |= data->apply;
      fillx = (gb_widget_input_bool (data, GbPackerFillX)) ? GTK_FILL_X : 0;
      myApply |= data->apply;
      filly = (gb_widget_input_bool (data, GbPackerFillY)) ? GTK_FILL_Y : 0;
      myApply |= data->apply;
      border = gb_widget_input_int (data, GbPackerBorder);
      myApply |= data->apply;
      padx = gb_widget_input_int (data, GbPackerPadX);
      myApply |= data->apply;
      pady = gb_widget_input_int (data, GbPackerPadY);
      myApply |= data->apply;
      ipadx = gb_widget_input_int (data, GbPackerIPadX);
      myApply |= data->apply;
      ipady = gb_widget_input_int (data, GbPackerIPadY);
      myApply |= data->apply;

      if (myApply)
	{
	  for (i = 0; GbPackerSideChoices[i]; i++)
	    {
	      if (!strcmp (side_str, GbPackerSideChoices[i])
		  || !strcmp (side_str, GbPackerSideSymbols[i]))
		{
		  side = GbPackerSideValues[i];
		  break;
		}
	    }

	  for (i = 0; GbPackerAnchorChoices[i]; i++)
	    {
	      if (!strcmp (anchor_str, GbPackerAnchorChoices[i])
		  || !strcmp (anchor_str, GbPackerAnchorSymbols[i]))
		{
		  anchor = GbPackerAnchorValues[i];
		  break;
		}
	    }

	  /* FIXME: only set if changed. */
	  gtk_packer_configure (GTK_PACKER (parent), widget, side, anchor,
				expand | fillx | filly,
				border, padx, pady, ipadx, ipady);
	}
    }
#endif

  data->loading_type = GB_STANDARD_PROPERTIES;
}


/* This ensures that placeholders are placed in every unoccupied cell.
   If rows and cols are not -1, then they are the new dimensions of the table.
 */
void
gb_widget_update_table_placeholders (GtkWidget * table, gint rows, gint cols)
{
  gint row, col, cell_contents;

  if (rows == -1)
    rows = GTK_TABLE (table)->nrows;
  if (cols == -1)
    cols = GTK_TABLE (table)->ncols;

  for (row = 0; row < rows; row++)
    {
      for (col = 0; col < cols; col++)
	{
	  /* Find out what is in the cell */
	  cell_contents = is_cell_occupied (table, row, col);
	  if (cell_contents == GB_CELL_WIDGET)
	    {
	      /* If cell is occupied, delete any placeholders there. If a placeholder
	         occupies the cell but spans multiple rows/cols split it into single
	         cells but without a placeholder in this cell */
	      remove_table_placeholders (table, row, col);

	    }
	  else if (cell_contents == GB_CELL_EMPTY)
	    {
	      /* If the cell is empty, put a placeholder in it */
	      gtk_table_attach_defaults (GTK_TABLE (table),
					 editor_new_placeholder (),
					 col, col + 1, row, row + 1);
	    }
	}
    }
}


/* Finds out if cell is occupied by a real widget. If not, returns whether
   the cell is empty or whether a placeholder is in it */
static gint
is_cell_occupied (GtkWidget * table, gint row, gint col)
{
  GList *children;
  GtkTableChild *child;
  gint return_val = GB_CELL_EMPTY;

  children = GTK_TABLE (table)->children;
  while (children)
    {
      child = children->data;
      if (child->top_attach <= row && child->bottom_attach > row
	  && child->left_attach <= col && child->right_attach > col)
	{
	  if (GB_IS_PLACEHOLDER (child->widget))
	    return_val = GB_CELL_PLACEHOLDER;
	  else
	    return GB_CELL_WIDGET;
	}
      children = children->next;
    }
  return return_val;
}


static void
remove_table_placeholders (GtkWidget * table, gint row, gint col)
{
  GList *children, *next;
  GtkTableChild *child;
  gint left, right, top, bottom;

  children = GTK_TABLE (table)->children;
  while (children)
    {
      next = children->next;
      child = children->data;
      left = child->left_attach;
      right = child->right_attach;
      top = child->top_attach;
      bottom = child->bottom_attach;

      if (top <= row && bottom > row && left <= col && right > col)
	{
	  /* If the widget is a placeholder remove it */
	  if (GB_IS_PLACEHOLDER (child->widget))
	    {
	      if (bottom - top > 1 || right - left > 1)
		{
		  split_table_placeholder (table, child->widget,
					   left, right, top, bottom,
					   row, col);
		}
	      else
		{
		  gtk_container_remove (GTK_CONTAINER (table), child->widget);
		}
	    }
	}
      children = next;
    }
}


static void
split_table_placeholder (GtkWidget * table, GtkWidget * placeholder,
			 gint left, gint right, gint top, gint bottom,
			 gint skip_row, gint skip_col)
{
  gint row, col;

  gtk_container_remove (GTK_CONTAINER (table), placeholder);
  for (row = top; row < bottom; row++)
    {
      for (col = left; col < right; col++)
	{
	  if (!(row == skip_row && col == skip_col))
	    {
	      gtk_table_attach_defaults (GTK_TABLE (table),
					 editor_new_placeholder (),
					 col, col + 1, row, row + 1);
	    }
	}
    }
}


static void
set_position_properties (GtkWidget * widget,
			 GbWidgetSetArgData * data)
{
  GbWidgetData *wdata = data->widget_data;
  gint x, y, w, h;
  gboolean applyX, applyY, applyWidth, applyHeight;
  gboolean set_usize = FALSE, set_upos = FALSE;

  x = gb_widget_input_int (data, GbX);
  applyX = data->apply;
  y = gb_widget_input_int (data, GbY);
  applyY = data->apply;
  w = gb_widget_input_int (data, GbWidth);
  applyWidth = data->apply;
  h = gb_widget_input_int (data, GbHeight);
  applyHeight = data->apply;

  /* When loading we need to remember which values have been set explicitly. */
  if (data->action == GB_LOADING)
    {
      if (applyX)
	wdata->flags |= GB_X_SET;
      if (applyY)
	wdata->flags |= GB_Y_SET;
      if (applyWidth)
	wdata->flags |= GB_WIDTH_SET;
      if (applyHeight)
	wdata->flags |= GB_HEIGHT_SET;
    }

#if 0
  g_print ("In set_position_properties X:%i Y:%i W:%i H:%i\n", x, y, w, h);
#endif
  if (GTK_IS_WINDOW (widget))
    {
      if (applyX)
	wdata->x = x;
      if (applyY)
	wdata->y = y;

      if (applyWidth && wdata->width != w)
	{
	  wdata->width = w;
	  set_usize = TRUE;
	}
      if (applyHeight && wdata->height != h)
	{
	  wdata->height = h;
	  set_usize = TRUE;
	}

      if (set_usize)
	{
	  gint w = wdata->flags & GB_WIDTH_SET ? wdata->width : -1;
	  gint h = wdata->flags & GB_HEIGHT_SET ? wdata->height : -1;
	  gb_widget_set_usize (widget, w, h);
	}
    }
  else if (widget->parent && GTK_IS_FIXED (widget->parent))
    {
      if (applyWidth && wdata->width != w)
	{
	  wdata->width = w;
	  set_usize = TRUE;
	}
      if (applyHeight && wdata->height != h)
	{
	  wdata->height = h;
	  set_usize = TRUE;
	}
      if (set_usize)
	gb_widget_set_usize (widget, wdata->width, wdata->height);

      if (applyX && wdata->x != x)
	{
	  wdata->x = x;
	  set_upos = TRUE;
	}
      if (applyY && wdata->y != y)
	{
	  wdata->y = y;
	  set_upos = TRUE;
	}
      if (set_upos)
	{
	  /* FIXME: Do we need both? Widget with windows aren't moved if we
	     only use gtk_fixed_move(). */
	  gtk_fixed_move (GTK_FIXED (widget->parent), widget,
			  wdata->x, wdata->y);
	  gtk_widget_set_uposition (widget, wdata->x, wdata->y);
	}
    }
  else
    {
      /* X & Y aren't applicable to widgets in standard containers, so we
	 make sure they aren't set. */
      wdata->flags &= ~(GB_X_SET | GB_Y_SET);

      if (applyWidth && wdata->width != w)
	{
	  wdata->width = w;
	  set_usize = TRUE;
	}
      if (applyHeight && wdata->height != h)
	{
	  wdata->height = h;
	  set_usize = TRUE;
	}

      if (set_usize)
	{
	  gint w = wdata->flags & GB_WIDTH_SET ? wdata->width : -1;
	  gint h = wdata->flags & GB_HEIGHT_SET ? wdata->height : -1;
	  gb_widget_set_usize (widget, w, h);
	}
      MSG2 ("*** Width set:%i Height set:%i", wdata->flags & GB_WIDTH_SET,
	    wdata->flags & GB_HEIGHT_SET);
    }
}


static void
apply_style (GtkWidget * widget,
	     GbWidgetSetArgData * data)
{
  GbWidgetData *wdata = data->widget_data;
  GtkStyle *style = widget->style, *old_style = NULL;
  GbStyle *gbstyle = wdata->gbstyle, *new_gbstyle;
  GdkFont *font = NULL;
  gchar *style_name, *xlfd_fontname;
  GdkColor fg[GB_NUM_STYLE_STATES];
  GdkColor bg[GB_NUM_STYLE_STATES];
  GdkColor text[GB_NUM_STYLE_STATES];
  GdkColor base[GB_NUM_STYLE_STATES];
  GdkPixmap *bg_pixmap[GB_NUM_STYLE_STATES];
  gchar *bg_pixmap_filenames[GB_NUM_STYLE_STATES];
  gint recreate = FALSE, redraw = FALSE, i;
  gchar buf[128], *filename;
  gboolean named_style;

  style_name = gb_widget_input_dialog (data, GbStyleName);
  named_style = (style_name[0] == '\0') ? FALSE : TRUE;
  if (data->apply)
    {
      if (named_style)
	{
	  new_gbstyle = (GbStyle *) g_hash_table_lookup (gb_style_hash, style_name);
	  g_return_if_fail (new_gbstyle != NULL);
	  if (new_gbstyle != gbstyle)
	    {
	      gbstyle = new_gbstyle;
	      gb_widget_set_gb_style (widget, gbstyle);
	      wdata->flags &= ~GB_STYLE_IS_UNNAMED;
	      redraw = TRUE;
	    }
	}
      else
	{
	  wdata->flags |= GB_STYLE_IS_UNNAMED;
	}
    }

  font = gb_widget_input_font (data, GbStyleFont, &xlfd_fontname);
  if (data->apply)
    {
      if (font != style->font)
	recreate = TRUE;
    }

  recreate |= apply_colors (widget, data, style->fg, fg, "fg");
  recreate |= apply_colors (widget, data, style->bg, bg, "bg");
  recreate |= apply_colors (widget, data, style->text, text, "text");
  recreate |= apply_colors (widget, data, style->base, base, "base");

  /* Background pixmaps */
  for (i = 0; i < GB_NUM_STYLE_STATES; i++)
    {
      sprintf (buf, "GtkStyle::%s[%s]", GbBgPixmapName, GbStateNames[i]);
      bg_pixmap[i] = gb_widget_input_bgpixmap (data, buf, &filename);
      bg_pixmap_filenames[i] = filename;
      if (data->apply)
	{
	  if (bg_pixmap[i] != style->bg_pixmap[i])
	    recreate = TRUE;
	}
    }

  if (recreate)
    {
      old_style = style;

      /* If the widget is supposedly using an unnamed GbStyle, but currently is
         actually using a named GbStyle (for convenience), then we need to
         create a copy of the GbStyle and place our new style in it. */
      if ((wdata->flags & GB_STYLE_IS_UNNAMED) && gbstyle->name)
	{
	  gbstyle = gb_widget_copy_gb_style (gbstyle);
	  g_free (gbstyle->name);
	  gbstyle->name = NULL;
	}

      style = gtk_style_new ();
      for (i = 0; i < GB_NUM_STYLE_STATES; i++)
	{
	  style->fg[i] = fg[i];
	  style->bg[i] = bg[i];
	  style->text[i] = text[i];
	  style->base[i] = base[i];
	  style->bg_pixmap[i] = bg_pixmap[i];
	  if (bg_pixmap[i])
	    gdk_pixmap_ref (bg_pixmap[i]);

	  if (gbstyle->bg_pixmap_filenames[i] != bg_pixmap_filenames[i])
	    {
	      g_free (gbstyle->bg_pixmap_filenames[i]);
	      gbstyle->bg_pixmap_filenames[i] = g_strdup (bg_pixmap_filenames
							  [i]);
	    }
	}
      if (font)
	{
	  gdk_font_unref (style->font);
	  style->font = font;
	  gdk_font_ref (style->font);
	}
      if (strcmp (gbstyle->xlfd_fontname, xlfd_fontname))
	{
	  g_free (gbstyle->xlfd_fontname);
	  gbstyle->xlfd_fontname = g_strdup (xlfd_fontname);
	}

      gbstyle->style = style;
      gtk_style_ref (style);
      gb_widget_set_gb_style (widget, gbstyle);
    }


  /* If a named style has been changed/recreated we have to update all
     widget's that use it. */
  if (recreate || redraw)
    {
      if (named_style)
	{
	  gb_widget_update_gb_styles (gbstyle, gbstyle);
	}
      else
	{
	  editor_refresh_widget (widget);
	}
    }

  if (old_style)
    gtk_style_unref (old_style);
}


/* This makes sure a widget's gbstyle & its style are up to date, and
   if the propagate flag is set it also updates any children.
   But it won't change a descendant's style if it has been set explicitly.
   FIXME: only propagates one level at present, and always sets child's style,
   even if it has a different GbStyle! */
void
gb_widget_set_gb_style (GtkWidget * widget,
			GbStyle * gbstyle)
{
  GbWidgetData *wdata;

  if (!GB_IS_PLACEHOLDER (widget))
    {
      if (widget->style != gbstyle->style)
	gtk_widget_set_style (widget, gbstyle->style);
    }

  wdata = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  if (wdata)
    {
      if (wdata->gbstyle != gbstyle)
	{
	  gb_widget_unref_gb_style (wdata->gbstyle);
	  wdata->gbstyle = gbstyle;
	  gb_widget_ref_gb_style (gbstyle);
	}
      /* If propagate style flags is set, propagate style to children */
      if (wdata->flags & GB_STYLE_PROPAGATE)
	gb_widget_children_foreach (widget, (GtkCallback) gb_widget_set_gb_style,
			   gbstyle);
    }
}


static gboolean
apply_colors (GtkWidget * widget, GbWidgetSetArgData * data, GdkColor colors[],
	      GdkColor new_colors[], gchar * name)
{
  gint state;
  gchar buf[128];
  guchar *color;
  GdkColor *old_gdkcolor, *new_gdkcolor;
  gboolean need_redraw = FALSE;

  for (state = 0; state < GB_NUM_STYLE_STATES; state++)
    {
      sprintf (buf, "GtkStyle::%s[%s]", name, GbStateNames[state]);
      old_gdkcolor = &colors[state];
      new_gdkcolor = &new_colors[state];
      new_gdkcolor->red = old_gdkcolor->red;
      new_gdkcolor->green = old_gdkcolor->green;
      new_gdkcolor->blue = old_gdkcolor->blue;
      new_gdkcolor->pixel = old_gdkcolor->pixel;

      color = gb_widget_input_color (data, buf);
      if (data->apply)
	{
	  new_gdkcolor->red = color[0] << 8;
	  new_gdkcolor->green = color[1] << 8;
	  new_gdkcolor->blue = color[2] << 8;
	  if (!gdk_color_equal (new_gdkcolor, old_gdkcolor))
	    need_redraw = TRUE;
	}
    }
  return need_redraw;
}


/* Currently this frees all the GbAccelerators and creates them from scratch */
static void
apply_accelerators (GtkWidget * widget, GbWidgetSetArgData * data)
{
  if (data->property_to_apply == NULL
      || property_is_accel_clist (data->property_to_apply))
    {
      gb_widget_free_accelerators (widget);
      data->widget_data->accelerators = property_get_accelerators ();
    }
}


/* Currently this frees all the GbSignals and creates them from scratch. */
static void
apply_signals (GtkWidget * widget, GbWidgetSetArgData * data)
{
  if (data->property_to_apply == NULL
      || property_is_signal_clist (data->property_to_apply))
    {
      gb_widget_free_signals (widget);
      data->widget_data->signals = property_get_signals ();
    }
}



/*************************************************************************
 * Functions for showing the popup context-sensitive menu of a widget
 *************************************************************************/

void
gb_widget_show_popup_menu (GtkWidget * widget,
			   GdkEventButton * event)
{
  static GtkWidget *menu = NULL;

  GbWidget *gbwidget;
  gchar *class_name, *name;
  GtkWidget *menuitem, *ancestor, *submenu, *child;
  GbWidgetCreateMenuData data;

  class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
  g_return_if_fail (gbwidget != NULL);

  name = gtk_widget_get_name (widget);
  if (GB_IS_PLACEHOLDER (widget))
    name = "Placeholder";

  /* FIXME: bug workaround - delete last popup menu */
  if (menu)
    gtk_widget_destroy (menu);

  menu = gtk_menu_new ();
  /* doesn't call menuitem's 'activate' handler if we do this
     gtk_signal_connect_object (GTK_OBJECT (menu), "deactivate",
     GTK_SIGNAL_FUNC(gtk_widget_destroy),
     GTK_OBJECT(menu));
   */
  menuitem = gtk_menu_item_new_with_label (name);
  gtk_widget_show (menuitem);
  gtk_widget_set_sensitive (menuitem, FALSE);
  gtk_menu_append (GTK_MENU (menu), menuitem);

  data.project = current_project;
  data.menu = menu;
  data.child = NULL;
  add_standard_top_menu_items (widget, &data);

  if (gbwidget->gb_widget_create_popup_menu)
    (gbwidget->gb_widget_create_popup_menu) (widget, &data);

  add_standard_bottom_menu_items (widget, &data);

  child = widget;
  ancestor = widget->parent;
  while (ancestor)
    {
      name = gtk_widget_get_name (ancestor);
      if (GB_IS_PLACEHOLDER (ancestor))
	name = "Placeholder";

      /* Skip widgets which aren't GbWidgets */
      if (GB_IS_GB_WIDGET (ancestor))
	{
	  /* Add a separator */
	  menuitem = gtk_menu_item_new ();
	  gtk_menu_append (GTK_MENU (menu), menuitem);
	  gtk_widget_show (menuitem);

	  menuitem = gtk_menu_item_new_with_label (name);
	  gtk_widget_show (menuitem);
	  gtk_menu_append (GTK_MENU (menu), menuitem);

	  /* Create submenu */
	  submenu = gtk_menu_new ();
	  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), submenu);

	  data.menu = submenu;
	  data.child = child;
	  add_standard_top_menu_items (ancestor, &data);

	  /* Call ancestors function to add any menu items */
	  class_name = gtk_type_name (GTK_OBJECT_TYPE (ancestor));
	  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
	  if (gbwidget != NULL && gbwidget->gb_widget_create_popup_menu)
	    (gbwidget->gb_widget_create_popup_menu) (ancestor, &data);

	  add_standard_bottom_menu_items (ancestor, &data);
	}
      child = ancestor;
      ancestor = ancestor->parent;
    }

  MSG ("showing popup menu");
  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
		  event->button, event->time);
}


static void
add_standard_top_menu_items (GtkWidget * widget, GbWidgetCreateMenuData * data)
{
  GtkWidget *menuitem;

  menuitem = gtk_menu_item_new_with_label (_("Select"));
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC (editor_on_select_activate), widget);
  gtk_widget_show (menuitem);
  gtk_menu_append (GTK_MENU (data->menu), menuitem);
}


static void
add_standard_bottom_menu_items (GtkWidget * widget, GbWidgetCreateMenuData * data)
{
  GtkWidget *menuitem;

  if (!GTK_IS_WINDOW (widget)
      && !GTK_IS_MISC (widget)
      && !GTK_IS_ALIGNMENT (widget)
      && !GB_IS_PLACEHOLDER (widget))
    {
      if (GTK_IS_ALIGNMENT (widget->parent))
	{
	  menuitem = gtk_menu_item_new_with_label (_("Remove Alignment"));
	  gtk_widget_show (menuitem);
	  gtk_menu_append (GTK_MENU (data->menu), menuitem);
	  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC (gb_widget_remove_alignment), widget);
	}
      else
	{
	  menuitem = gtk_menu_item_new_with_label (_("Add Alignment"));
	  gtk_widget_show (menuitem);
	  gtk_menu_append (GTK_MENU (data->menu), menuitem);
	  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
			 GTK_SIGNAL_FUNC (gb_widget_add_alignment), widget);
	}
    }

  if (GTK_WIDGET_NO_WINDOW (widget)
      && !GTK_IS_EVENT_BOX (widget)
      && !GB_IS_PLACEHOLDER (widget))
    {
      if (GTK_IS_EVENT_BOX (widget->parent))
	{
	  menuitem = gtk_menu_item_new_with_label (_("Remove Event Box"));
	  gtk_widget_show (menuitem);
	  gtk_menu_append (GTK_MENU (data->menu), menuitem);
	  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC (gb_widget_remove_event_box), widget);
	}
      else
	{
	  menuitem = gtk_menu_item_new_with_label (_("Add Event Box"));
	  gtk_widget_show (menuitem);
	  gtk_menu_append (GTK_MENU (data->menu), menuitem);
	  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
			 GTK_SIGNAL_FUNC (gb_widget_add_event_box), widget);
	}
    }

  menuitem = gtk_menu_item_new_with_label (_("Cut"));
  gtk_widget_show (menuitem);
  gtk_menu_append (GTK_MENU (data->menu), menuitem);
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC (editor_on_cut_activate), widget);
 
  menuitem = gtk_menu_item_new_with_label (_("Copy"));
  gtk_widget_show (menuitem);
  gtk_menu_append (GTK_MENU (data->menu), menuitem);
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC (editor_on_copy_activate), widget);
 
  menuitem = gtk_menu_item_new_with_label (_("Paste"));
  gtk_widget_show (menuitem);
  if (!widget->parent)
    gtk_widget_set_sensitive (menuitem, FALSE);
  gtk_menu_append (GTK_MENU (data->menu), menuitem);
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC (editor_on_paste_activate), widget);
 
  /* This is Elliot's code to convert to/from components but its not finished
     so I've taken it out for now. */
#if 0
  if (!GB_IS_COMPONENT(widget))
    {
      menuitem = gtk_menu_item_new_with_label (_("Convert to Component"));
      gtk_widget_show (menuitem);
      if (!widget->parent)
        gtk_widget_set_sensitive (menuitem, FALSE);
      gtk_menu_append (GTK_MENU (data->menu), menuitem);
      gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
                      GTK_SIGNAL_FUNC (editor_on_convert_component), widget);
    }

  if (GB_IS_COMPONENT(widget))
    {
      menuitem = gtk_menu_item_new_with_label (_("Convert from Component"));
      gtk_widget_show (menuitem);
      if (!widget->parent)
        gtk_widget_set_sensitive (menuitem, FALSE);
      gtk_menu_append (GTK_MENU (data->menu), menuitem);
      gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
                      GTK_SIGNAL_FUNC (editor_on_convert_from_component),
		      widget);
    }
#endif

  menuitem = gtk_menu_item_new_with_label (_("Delete"));
  gtk_widget_show (menuitem);
  gtk_menu_append (GTK_MENU (data->menu), menuitem);
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC (editor_on_delete_activate), widget);
}



/*************************************************************************
 * Functions for loading project files
 *************************************************************************/

/* This reads in a widget, and is recursively called to read it's children.
   It assumes a '<widget>' has just been read, and reads up to and including
   '</widget>'. Note that if any gbwidgets have their own gb_widget_load
   function they are responsible for calling functions to set the standard
   properties, add the widget to its parent, and load any children. */
GtkWidget*
gb_widget_load (GtkWidget * widget, GbWidgetSetArgData * data, GtkWidget * parent)
{
  gchar *class_name, *child_name, *ancestor_class_name, *component_key;
  GbWidget *gbwidget, *ancestor_gbwidget;
  GtkWidget *ancestor;
  GbWidgetData *wdata;

  data->action = GB_LOADING;
  data->loading_type = GB_STANDARD_PROPERTIES;
  gb_widget_load_properties (data);

  if (data->status != GLADE_STATUS_OK)
    {
      MSG ("Load error");
      return NULL;
    }

#ifdef GLD_DEBUG
  {
    gint i;
    for (i = 0; i < data->properties.n; i++)
      {
	MSG3 ("Line number: %i Property: %s Value: %s",
	      data->properties.ptr[i].line_number,
	      data->buffer.ptr + data->properties.ptr[i].tag_index,
	      data->buffer.ptr + data->properties.ptr[i].cdata_index);

      }
  }
#endif
  class_name = load_get_value (data, "class");
  MSG1 ("Reading Class: %s", class_name);
  if (!class_name)
    {
      MSG ("Load error");
      data->status = GLADE_STATUS_CLASS_NAME_MISSING;
      return NULL;
    }

  child_name = load_get_value (data, "child_name");
  /* SPECIAL CODE: if a toolbar button is pasted, but not into a toolbar, we
     need to clear the child_name. This is a kludge. */
  if (data->xml_buffer && data->replacing_widget
      && child_name && !strcmp (child_name, "Toolbar:button"))
    {
      if (parent == NULL || !GTK_IS_TOOLBAR (parent))
	child_name = NULL;
    }

  if (!g_strcasecmp (class_name, "placeholder"))
    {
      MSG ("found placeholder");
      widget = editor_new_placeholder ();
      gb_widget_add_child (parent, data, widget);
      if (child_name)
	gtk_object_set_data (GTK_OBJECT (widget), GB_CHILD_NAME_KEY,
			     g_strdup (child_name));
      return NULL;
    }

  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
  if (gbwidget == NULL)
    {
      MSG ("Load error");
      data->status = GLADE_STATUS_CLASS_UNKNOWN;
      load_add_error_message_with_tag (data, GLADE_LINE_PROPERTY,
				       glade_get_error_message (data->status),
				       "class", class_name);
      return NULL;
    }

  component_key = load_get_value (data, "component_key");
  
  /* If this is a special child of a widget, step through the ancestors of
     the widget and try to find it. Note that some special child widgets
     have to be created normally, and then added by the parent container,
     e.g. notebook tabs and clist titles - see gb_widget_add_child(). */
  if (child_name)
    {
      MSG1 ("Child name: %s", child_name);
      ancestor = parent;
      while (ancestor)
	{
	  ancestor_class_name = gtk_type_name (GTK_OBJECT_TYPE (ancestor));
	  MSG1 ("Ancestor type: %s", ancestor_class_name);
	  ancestor_gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table,
						       ancestor_class_name);
	  if (ancestor_gbwidget && ancestor_gbwidget->gb_widget_get_child)
	    {
	      widget = (ancestor_gbwidget->gb_widget_get_child) (ancestor,
								 child_name);
	      if (widget)
		break;
	    }

	  ancestor = ancestor->parent;
	}

#if 0
      if (!widget)
	g_print ("Child widget %s not found - may be a problem\n", child_name);
#endif
    }

  /* If this is a standard widget, we need to create it and add it to its
     parent. If the widget has already been created by its parent, we can
     just set the properties. */
  if (widget == NULL)
    {
      MSG ("widget == NULL, has not been created by parent.");
      widget = gb_widget_new_full (class_name, FALSE, parent, NULL,
				   0, 0, NULL, GB_LOADING, data);
      wdata = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
      data->widget_data = wdata;
      gb_widget_add_child (parent, data, widget);
      /* This must come after add_child so we can set special child properties,
         and also we may need to realize the widget. */
      set_standard_properties (widget, data);

      if (parent == NULL)
	glade_project_add_component (data->project, widget);
#ifdef GTK_HAVE_FEATURES_1_1_0
      tree_add_widget (widget);
#endif
    }
  else
    {
      MSG ("widget != NULL, has been created by parent.");
      wdata = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
      if (wdata == NULL)
	g_warning ("Null widget_data\n");
      data->widget_data = wdata;
      set_standard_properties (widget, data);
#ifdef GTK_HAVE_FEATURES_1_1_0
      tree_add_widget (widget);
#endif
    }

  if (child_name)
    gtk_object_set_data (GTK_OBJECT (widget), GB_CHILD_NAME_KEY,
			 g_strdup (child_name));

  if (component_key)
    gtk_object_set_data (GTK_OBJECT (widget), GB_COMPONENT_KEY, "True");

  if (data->status != GLADE_STATUS_OK)
    {
      MSG ("Load error");
      return NULL;
    }

  MSG ("Calling widgets set_properties()");
  if (gbwidget->gb_widget_set_properties)
    (gbwidget->gb_widget_set_properties) (widget, NULL, data);
  MSG ("Called widgets set_properties()");

  if (data->status != GLADE_STATUS_OK)
    {
      MSG ("Load error");
      return NULL;
    }

  /* All the necessary properties should have been copied from the buffer
     by now, se we can clear it. Note: actually we use it below but that
     should be OK?? */
  load_buffer_release (data);

  /* When pasting, we only ever replace the top widget. All children are added
     as normal. */
  data->replacing_widget = FALSE;

  MSG ("Loading children");

  /* load children. */
  while (data->token_type == GB_TOKEN_START_TAG)
    {
      if (!strcmp (data->buffer.ptr + data->token, "widget"))
	{
	  /* We have to reset the widget_data since this may have been changed
	     while loading the last child (and its children) */
	  data->widget_data = wdata;
	  gb_widget_load (NULL, data, widget);
	  if (data->status != GLADE_STATUS_OK)
	    {
	      MSG ("Load error");
	      return NULL;
	    }
	}
      else
	{
	  break;
	}
      load_token_skip_whitespace (data);
      if (data->status != GLADE_STATUS_OK)
	{
	  MSG ("Load error");
	  return NULL;
	}
    }

  MSG ("Loaded children");

  if (data->token_type != GB_TOKEN_END_TAG
      || strcmp (data->buffer.ptr + data->token, "widget"))
    {
      data->status = GLADE_STATUS_END_TAG_EXPECTED;
      load_add_error_message (data, data->line_number, _("End tag expected"),
			      "widget");
      MSG ("Load error");
      return NULL;
    }

  /* If the widget is a table, we make sure all empty cells have placeholders
     in them. */
  if (GTK_IS_TABLE (widget))
    {
      gb_widget_update_table_placeholders (widget, -1, -1);
    }

  return widget;
}


/* Adds a widget to a parent, when loading, using code for each type of
   container. If the parent is NULL, then this is a toplevel window/dialog
   etc. and so we add it to the component list.
   If we are pasting from the clipboard, the top widget has to be treated
   specially - if the given parent widget is a container (but not a simple
   bin or button) we add the new widget as a child as normal. If it isn't
   we replace the widget. */
void
gb_widget_add_child (GtkWidget * parent,
		     GbWidgetSetArgData * data,
		     GtkWidget * child)
{
  gchar *class_name;
  GbWidget *gbwidget;


  MSG2 ("Adding %s to %s", gtk_widget_get_name (child),
	parent ? gtk_widget_get_name (parent) : "NULL");

  /* If we are pasting, the top widget replaces the parent rather than being
     added to it. */
  if (data->xml_buffer && data->replacing_widget)
    {
      gb_widget_replace_child (parent, data->replacing_widget, child);
      return;
    }

  if (parent == NULL)
      return;

  class_name = gtk_type_name (GTK_OBJECT_TYPE (parent));
  gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
  if (gbwidget->gb_widget_add_child)
    {
      gbwidget->gb_widget_add_child (parent, child, data);
    }
  else if (GTK_IS_BIN (parent))
    {
      if (GTK_BIN (parent)->child)
	gtk_container_remove (GTK_CONTAINER (parent), GTK_BIN (parent)->child);
      gtk_container_add (GTK_CONTAINER (parent), child);
    }
  else if (GTK_IS_CONTAINER (parent))
    {
      gtk_container_add (GTK_CONTAINER (parent), child);
    }
}


static void
gb_realloc_properties_if_needed (GbLoadPropControl *p)
{
  if (p->n == p->space)
    {
      p->space += GB_LOADED_PROPERTIES_INC;
      p->ptr = g_realloc (p->ptr, sizeof(GbLoadedProperty) * p->space);
    }
}

static void
gb_add_properties (gint tag_index, gint cdata_index,
		   gint line_number, GbLoadPropControl *p)
{
    gb_realloc_properties_if_needed (p);
    p->ptr[p->n].tag_index   = tag_index;
    p->ptr[p->n].cdata_index = cdata_index;
    p->ptr[p->n].line_number = line_number;
    p->n++;
}


/* This will load all the properties/accelerators/signals of a widget.
   It assumes a <widget> has just been read and reads up to a <widget>,
   </widget>. */
static void
gb_widget_load_properties (GbWidgetSetArgData * data)
{
  gint property;

  /* Initialise the widget's properties. */
  data->properties.n = 0;
  data->child_properties.n = 0;

  data->accelerators = NULL;
  data->signals = NULL;
  data->gbstyle = NULL;

  for (;;)
    {
      load_token_skip_whitespace (data);
      if (data->status != GLADE_STATUS_OK)
	{
	  MSG ("Load error");
	  return;
	}

      if (!strcmp (data->buffer.ptr + data->token, "widget"))
	return;

      if (data->token_type != GB_TOKEN_START_TAG)
	{
	  data->status = GLADE_STATUS_START_TAG_EXPECTED;
	  MSG ("Load error");
	  return;
	}

      if (!strcmp (data->buffer.ptr + data->token, "style"))
	gb_widget_load_style (data);
      else if (!strcmp (data->buffer.ptr + data->token, "accelerator"))
	load_accelerator (data);
      else if (!strcmp (data->buffer.ptr + data->token, "signal"))
	load_signal (data);
      else if (!strcmp (data->buffer.ptr + data->token, "child"))
	load_special_child_properties (data);
      else
	{
	  property = data->token;
	  load_token (data);
	  if (data->status != GLADE_STATUS_OK)
	    {
	      MSG ("Load error");
	      return;
	    }

	  /* Remember the property could be empty, so we may get the end tag.*/
	  if (data->token_type == GB_TOKEN_START_TAG)
	    {
	      data->status = GLADE_STATUS_DATA_EXPECTED;
	      load_add_error_message (data, data->line_number,
				      _("Data expected"),
				      data->buffer.ptr + property);
	      MSG ("Load error");
	      return;
	    }

	  if (data->token_type == GB_TOKEN_DATA)
	    {
              gb_add_properties (property, data->token, data->line_number,
				 &data->properties);

	      load_token (data);
	      if (data->status != GLADE_STATUS_OK)
		{
		  MSG ("Load error");
		  return;
		}
	      if (data->token_type != GB_TOKEN_END_TAG
		  || strcmp (data->buffer.ptr + data->token,
			     data->buffer.ptr + property))
		{
		  data->status = GLADE_STATUS_END_TAG_EXPECTED;
		  load_add_error_message (data, data->line_number,
					  _("End tag expected"),
					  data->buffer.ptr + property);
		  MSG ("Load error");
		  return;
		}
	    }
	  else
	    /* An empty tag may have been read. Check the start and end tags
	       match. */
	    {
	      if (strcmp (data->buffer.ptr + data->token, data->buffer.ptr + property))
		{
		  data->status = GLADE_STATUS_END_TAG_EXPECTED;
		  load_add_error_message (data, data->line_number,
					  _("End tag expected"),
					  data->buffer.ptr + property);
		  MSG ("Load error");
		  return;
		}
              gb_add_properties (property, 0 ,data->line_number,
				 &data->properties); 
	    }
	}
    }
}


/* This will load all the child properties of a widget, e.g. its packing
   options. It assumes a <child> has just been read, and reads up to to
   </child>. */
static void
load_special_child_properties (GbWidgetSetArgData * data)
{
  gint element_index, cdata_index;

  for (;;)
    {
      load_element (data, &element_index, &cdata_index);

      if (data->status != GLADE_STATUS_OK)
	{
	  if (data->status == GLADE_STATUS_START_TAG_EXPECTED
	      && data->token_type == GB_TOKEN_END_TAG
	      && (!strcmp (data->buffer.ptr + data->token, "child")))
	    {
	      data->status = GLADE_STATUS_OK;
	    }
	  return;
	}

      gb_add_properties (element_index, cdata_index, data->line_number,
			 &data->child_properties);
    }
}


/* This will load all the accelerators of a widget. It assumes a <accelerator>
   has just been read and reads up to to </accelerator>. */
static void
load_accelerator (GbWidgetSetArgData * data)
{
  GbAccelerator *accel;
  gint element_index, cdata_index;
  gchar *element, *cdata;

  accel = g_new (GbAccelerator, 1);
  accel->modifiers = 0;
  accel->key = NULL;
  accel->signal = NULL;

  for (;;)
    {
      load_element (data, &element_index, &cdata_index);
      element = data->buffer.ptr + element_index;
      cdata = data->buffer.ptr + cdata_index;

      if (data->status != GLADE_STATUS_OK)
	{
	  if (data->status == GLADE_STATUS_START_TAG_EXPECTED
	      && data->token_type == GB_TOKEN_END_TAG
	      && (!strcmp (data->buffer.ptr + data->token, "accelerator")))
	    {
	      data->accelerators = g_list_append (data->accelerators, accel);
	      data->status = GLADE_STATUS_OK;
	    }
	  else
	    {
	      g_free (accel->key);
	      g_free (accel->signal);
	      g_free (accel);
	    }
	  return;
	}

      if (!strcmp (element, "modifiers"))
	{
	  /* We simple scan for the flags in the cdata, although they should
	     be separated by '|'. */
	  if (glade_util_strstr (cdata, "GDK_CONTROL_MASK"))
	    accel->modifiers |= GDK_CONTROL_MASK;
	  if (glade_util_strstr (cdata, "GDK_SHIFT_MASK"))
	    accel->modifiers |= GDK_SHIFT_MASK;
	  if (glade_util_strstr (cdata, "GDK_MOD1_MASK"))
	    accel->modifiers |= GDK_MOD1_MASK;
	}
      else if (!strcmp (element, "key"))
	{
	  if ((strlen (cdata) < 5) || strncmp (cdata, "GDK_", 4))
	    {
	      load_add_error_message_with_tag (data, GLADE_LINE_PROPERTY,
					       _("Invalid accelerator key"),
					       element, cdata);
	    }
	  else
	    {
	      g_free (accel->key);
	      /* The initial 'GDK_' is removed. */
	      accel->key = g_strdup (cdata + 4);
	    }
	}
      else if (!strcmp (element, "signal"))
	{
	  g_free (accel->signal);
	  accel->signal = g_strdup (cdata);
	}
      else
	load_add_error_message_with_tag (data, GLADE_LINE_PROPERTY,
					 _("Unknown tag"), element, cdata);
    }
}


/* This will load all the accelerators of a widget. It assumes a <signal>
   has just been read and reads up to to </signal>. */
static void
load_signal (GbWidgetSetArgData * data)
{
  GbSignal *signal;
  gint element_index, cdata_index;
  gchar *element, *cdata;

  signal = g_new (GbSignal, 1);
  signal->name = NULL;
  signal->handler = NULL;
  signal->object = NULL;
  signal->after = FALSE;
  signal->data = NULL;
  signal->last_modification_time = 0;

  for (;;)
    {
      load_element (data, &element_index, &cdata_index);
      element = data->buffer.ptr + element_index;
      cdata = data->buffer.ptr + cdata_index;

      if (data->status != GLADE_STATUS_OK)
	{
	  if (data->status == GLADE_STATUS_START_TAG_EXPECTED
	      && data->token_type == GB_TOKEN_END_TAG
	      && (!strcmp (data->buffer.ptr + data->token, "signal")))
	    {
	      data->signals = g_list_append (data->signals, signal);
	      data->status = GLADE_STATUS_OK;
	      MSG ("Signal loaded OK");
	    }
	  else
	    {
	      g_free (signal->name);
	      g_free (signal->handler);
	      g_free (signal->object);
	      g_free (signal->data);
	      g_free (signal);
	      MSG1 ("ERROR loading signal: %i", data->status);
	    }
	  return;
	}

      MSG2 ("Signal element:%s cdata:%s", element, cdata);

      if (!strcmp (element, "name"))
	{
	  g_free (signal->name);
	  signal->name = g_strdup (cdata);
	}
      else if (!strcmp (element, "handler"))
	{
	  g_free (signal->handler);
	  signal->handler = g_strdup (cdata);
	}
      else if (!strcmp (element, "object"))
	{
	  g_free (signal->object);
	  signal->object = g_strdup (cdata);
	}
      else if (!strcmp (element, "after"))
	{
	  signal->after = load_parse_bool (data, cdata);
	  if (data->status == GLADE_STATUS_INVALID_VALUE)
	    {
	      signal->after = FALSE;
	      load_add_error_message_with_tag (data, GLADE_LINE_CURRENT,
					       _("Invalid boolean value"),
					       element, cdata);
	      data->status = GLADE_STATUS_OK;
	    }
	}
      else if (!strcmp (element, "data"))
	{
	  g_free (signal->data);
	  signal->data = g_strdup (cdata);
	}
      else if (!strcmp (element, "last_modification_time"))
	{
	  signal->last_modification_time = load_parse_date (data, cdata);
	  if (data->status == GLADE_STATUS_INVALID_VALUE)
	    {
	      signal->last_modification_time = 0;
	      load_add_error_message_with_tag (data, GLADE_LINE_CURRENT,
					       _("Invalid date"), element,
					       cdata);
	      data->status = GLADE_STATUS_OK;
	    }
	}
      else
	load_add_error_message (data, data->line_number, _("Unknown tag"),
				element);
    }
}


/* This will load a style. It assumes a <style> has just been read and reads
   up to to </style>. If the style is named it is added to the GbStyle hash.
   The created GbStyle is returned in data->gbstyle. */
void
gb_widget_load_style (GbWidgetSetArgData * data)
{
  GbStyle *gbstyle;
  GtkStyle *style;
  gchar *element, *cdata;
  gint element_index, cdata_index;
  gboolean loading_default_style = FALSE;

  gbstyle = gb_widget_new_gb_style ();
  /* We copy the default style, since only changed values will appear in the
     input file. */
  style = gbstyle->style = gtk_style_copy (gb_widget_default_gb_style->style);
  gtk_style_ref (style);

  for (;;)
    {
      load_element (data, &element_index, &cdata_index);
      element = data->buffer.ptr + element_index;
      cdata = data->buffer.ptr + cdata_index;

      if (data->status != GLADE_STATUS_OK)
	{
	  if (data->status == GLADE_STATUS_START_TAG_EXPECTED
	      && data->token_type == GB_TOKEN_END_TAG
	      && (!strcmp (data->buffer.ptr + data->token, "style")))
	    {
	      /* Make sure the new GbStyle has a font name set. */
	      if (!gbstyle->xlfd_fontname)
		{
		  gbstyle->xlfd_fontname
		    = g_strdup (gb_widget_default_gb_style->xlfd_fontname);
		}

	      /* If we are loading the Default style, replace current one. */
	      if (loading_default_style)
		{
		  gb_widget_destroy_gb_style (gb_widget_default_gb_style, TRUE);
		  gb_widget_default_gb_style = gbstyle;
		}

	      /* If the style is named, add it to the hash. */
	      if (gbstyle->name)
		g_hash_table_insert (gb_style_hash, gbstyle->name, gbstyle);

	      data->status = GLADE_STATUS_OK;
	      return;
	    }
	  MSG ("Load error");
	  return;
	}

      if (!strcmp (element, "style_name"))
	{
	  g_free (gbstyle->name);
	  gbstyle->name = g_strdup (cdata);
	  if (!strcmp (gbstyle->name, GB_STYLE_DEFAULT))
	    loading_default_style = TRUE;
	  MSG1 ("Loading named style:%s", gbstyle->name);
	}
      else if (!strcmp (element, "style_font"))
	{
	  GdkFont *font;
	  g_free (gbstyle->xlfd_fontname);
	  gbstyle->xlfd_fontname = g_strdup (cdata);
	  font = gdk_font_load (cdata);
	  if (font)
	    {
	      gdk_font_unref (style->font);
	      style->font = font;
	      gdk_font_ref (font);
	    }
	  else
	    load_add_error_message_with_tag (data, GLADE_LINE_CURRENT,
					     _("Couldn't load font"),
					     element, cdata);
	}
      else
	{
	  gboolean found;

	  found = load_colors (data, style->fg, "fg", element, cdata);
	  if (!found)
	    found = load_colors (data, style->bg, "bg", element, cdata);
	  if (!found)
	    found = load_colors (data, style->text, "text", element, cdata);
	  if (!found)
	    found = load_colors (data, style->base, "base", element, cdata);

	  /* Check if element is a color or bgpixmap of one of the states. */
	  if (!found)
	    {
	      gint state;
	      gchar buf[128];
	      GdkPixmap *gdkpixmap;

	      for (state = 0; state < GB_NUM_STYLE_STATES; state++)
		{
		  /* Remember the tag will be converted to lowercase. */
		  sprintf (buf, "GtkStyle::%s-%s", GbBgPixmapName,
			   GbStateNames[state]);
		  g_strdown (buf);
		  if (!strcmp (element,
			       glade_util_find_start_of_tag_name (buf))
		      && cdata[0] != '\0')
		    {
		      g_free (gbstyle->bg_pixmap_filenames[state]);
		      gbstyle->bg_pixmap_filenames[state] = g_strdup (cdata);

		      gdkpixmap = gdk_pixmap_create_from_xpm (NULL,
							      NULL,
					       &style->bg[GTK_STATE_NORMAL],
							      cdata);
		      if (gdkpixmap)
			{
			  style->bg_pixmap[state] = gdkpixmap;
			}
		      else
			load_add_error_message_with_tag (data,
							 GLADE_LINE_CURRENT,
							 _("Couldn't load pixmap"),
							 element, cdata);

		      found = TRUE;
		    }
		}
	    }

	  if (!found)
	    load_add_error_message_with_tag (data, GLADE_LINE_CURRENT,
					     _("Unknown tag"), element, cdata);
	}
    }
}


static gboolean
load_colors (GbWidgetSetArgData * data, GdkColor colors[], gchar * name,
	     gchar * element, gchar * cdata)
{
  gint state;
  gchar buf[128];
  guchar *color;

  for (state = 0; state < GB_NUM_STYLE_STATES; state++)
    {
      /* Remember the tag will be converted to lowercase. */
      sprintf (buf, "GtkStyle::%s-%s", name, GbStateNames[state]);
      g_strdown (buf);
      if (!strcmp (element, glade_util_find_start_of_tag_name (buf)))
	{
	  color = load_parse_color (data, cdata);
	  if (color)
	    {
	      MSG3 ("RGB: %i,%i,%i", color[0], color[1], color[2]);
	      colors[state].red = color[0] << 8;
	      colors[state].green = color[1] << 8;
	      colors[state].blue = color[2] << 8;
	      return TRUE;
	    }
	  else
	    {
	      load_add_error_message_with_tag (data, GLADE_LINE_PROPERTY,
					       _("Invalid color"), element,
					       cdata);
	      return TRUE;
	    }
	}
    }
  return FALSE;
}


/* This applies the loaded style to the widget. */
static void
load_style (GtkWidget * widget, GbWidgetSetArgData * data)
{
  gchar *gbstyle_name;

  /* Check if its using a named style. */
  gbstyle_name = load_get_value (data, GbStyleName);
  if (gbstyle_name)
    {

    }

}


/*************************************************************************
 * Functions for saving project files
 *************************************************************************/

void
gb_widget_save (GtkWidget * widget,
		GbWidgetGetArgData * data)
{
  GbWidget *gbwidget;
  GbWidgetData *widget_data;
  gchar *class_name, *child_name, *component_key;

  child_name = gtk_object_get_data (GTK_OBJECT (widget), GB_CHILD_NAME_KEY);

  /* Don't save a placeholder if its parent is a table, since we don't really
     want placeholders in the XML when the interface is finished, but it
     is quite possible that some table cells will be left blank. */
  if (GB_IS_PLACEHOLDER (widget))
    {
      if (!GTK_IS_TABLE (widget->parent))
	{
	  save_newline (&data->buffer);
	  save_tag (data, "widget");
	  data->indent++;
	  save_string (data, "class", "Placeholder");
	  if (child_name)
	    save_string (data, "child_name", child_name);
	  data->indent--;
	  save_end_tag (data, "widget");
	}
      return;
    }

  class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));

  /* Write out widget data and any child widgets */
  widget_data = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  /* If this isn't a gbwidget, skip it. */
  if (widget_data)
    {
      data->action = GB_SAVING;
      data->widget_data = widget_data;

      save_newline (&data->buffer);
      save_tag (data, "widget");
      data->indent++;
      save_string (data, "class", class_name);
      if (child_name)
	save_string (data, "child_name", child_name);
      component_key = gtk_object_get_data (GTK_OBJECT (widget),
				           GB_COMPONENT_KEY);
      if (component_key)
        save_string (data, "component_key", component_key);

      get_standard_properties (widget, data);

      /* Call gbwidgets save function for any extra properties */
      gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
      g_return_if_fail (gbwidget != NULL);
      if (gbwidget->gb_widget_get_properties)
	(gbwidget->gb_widget_get_properties) (widget, NULL, data);
    }

  /* Recursively save children. */
  gb_widget_children_foreach (widget, (GtkCallback) gb_widget_save, data);

  if (widget_data)
    {
      data->indent--;
      save_end_tag (data, "widget");
    }
}


/* This outputs all the standard widget properties, for showing or saving.
   Note that when saving we try not to save default values. */
static void
get_standard_properties (GtkWidget * widget,
			 GbWidgetGetArgData * data)
{
  gboolean can_default, has_default, can_focus, has_focus, visible, sensitive;
  GbWidgetData *wdata = data->widget_data;
  gchar *name, *class_name;
  GbWidgetAction action = data->action;

  /* Save widget's name if it has been set */
  name = gtk_widget_get_name (widget);
  class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));
  if (action != GB_SAVING || strcmp (name, class_name))
    gb_widget_output_string (data, GbName, name);

  /* If widget is a container, save the border width */
  if (GTK_IS_CONTAINER (widget))
    {
      gint border_width;
      gchar buf[128];
      border_width = GTK_CONTAINER (widget)->border_width;
      if (action != GB_SAVING || border_width > 0)
	{
	  sprintf (buf, "%s::border_width", class_name);
	  gb_widget_output_int (data, buf, border_width);
	}
    }

  get_special_child_properties (widget, data);
  get_position_properties (widget, data);
  get_lang_specific_properties (widget, data);

  visible = wdata->flags & GB_VISIBLE;
  if (action != GB_SAVING || !visible)
    gb_widget_output_bool (data, GbVisible, visible);
  sensitive = wdata->flags & GB_SENSITIVE;
  if (action != GB_SAVING || !sensitive)
    gb_widget_output_bool (data, GbSensitive, sensitive);

  /* Only widgets with windows can have tooltips at present. */
  if (!GTK_WIDGET_NO_WINDOW (widget))
    {
      if (action != GB_SAVING || wdata->tooltip)
	gb_widget_output_string (data, GbTooltip, wdata->tooltip);
      if (action == GB_SHOWING)
	property_set_sensitive (GbTooltip, TRUE);
    }
  else if (action == GB_SHOWING)
    {
      /* N/A stands for 'Not Applicable'. It is used when a standard widget
	 property does not apply to the current widget. e.g. widgets without
	 windows can't use the Events property. This appears in the property
	 editor and so should be a short abbreviation. */
      gb_widget_output_string (data, GbTooltip, _("N/A"));
      property_set_sensitive (GbTooltip, FALSE);
    }

  can_default = GTK_WIDGET_CAN_DEFAULT (widget);
  if (action != GB_SAVING || can_default)
    gb_widget_output_bool (data, GbCanDefault, can_default);
  has_default = wdata ? wdata->flags & GB_GRAB_DEFAULT : 0;
  if (action != GB_SAVING || has_default)
    gb_widget_output_bool (data, GbHasDefault, has_default);

  can_focus = GTK_WIDGET_CAN_FOCUS (widget);
  if (action != GB_SAVING || can_focus)
    gb_widget_output_bool (data, GbCanFocus, can_focus);
  has_focus = wdata ? wdata->flags & GB_GRAB_FOCUS : 0;
  if (action != GB_SAVING || has_focus)
    gb_widget_output_bool (data, GbHasFocus, has_focus);

  if (!GTK_WIDGET_NO_WINDOW (widget))
    {
      GdkExtensionMode ext_mode = gtk_widget_get_extension_events (widget);
      gint i;

      if (action == GB_SAVING)
	{
	  if (wdata && wdata->events)
	    {
	      gchar buffer[1024];
	      gboolean written_first = FALSE;

	      buffer[0] = '\0';
	      for (i = 0; i < GB_EVENT_MASKS_COUNT; i++)
		{
		  if (wdata->events & GbEventMaskValues[i])
		    {
		      if (written_first)
			strcat (buffer, " | ");
		      strcat (buffer, GbEventMaskSymbols[i]);
		      written_first = TRUE;
		    }
		}

	      gb_widget_output_string (data, GbEvents, buffer);
	    }
	}
      else
	{
	  gb_widget_output_string (data, GbEvents,
			   property_events_value_to_string (wdata->events));
	}

      /* Don't save default extension mode ('None') */
      if (action != GB_SAVING || ext_mode != GDK_EXTENSION_EVENTS_NONE)
	{
	  for (i = 0; GbExtensionModeChoices[i]; i++)
	    {
	      if (GbExtensionModeValues[i] == ext_mode)
		gb_widget_output_choice (data, GbExtEvents, i,
					 GbExtensionModeSymbols[i]);
	    }
	}
      if (action == GB_SHOWING)
	{
	  property_set_sensitive (GbEvents, TRUE);
	  property_set_sensitive (GbExtEvents, TRUE);
	}
    }
  else if (action == GB_SHOWING)
    {
      gb_widget_output_dialog (data, GbEvents, _("N/A"), NULL);
      property_set_sensitive (GbEvents, FALSE);
      gb_widget_output_choice (data, GbExtEvents, 0, GbExtensionModeSymbols[0]);
      property_set_sensitive (GbExtEvents, FALSE);
    }

  if (action == GB_SHOWING)
    {
      gb_widget_show_style (widget);
      show_accelerators (widget, data);
      show_signals (widget, data);
    }
  else
    {
      save_style (widget, data);
      save_accelerators (widget, data);
      save_signals (widget, data);
    }
}


static void
get_special_child_properties (GtkWidget * widget, GbWidgetGetArgData * data)
{
  GtkWidget *parent = widget->parent;
#ifdef GTK_HAVE_FEATURES_1_1_0
  gboolean expand, fill;
  guint padding;
#else
  gint expand, fill, padding;
#endif
  GtkPackType pack_type;

  if (!parent)
    {
      if (data->action == GB_SHOWING)
	property_hide_special_child_page ();
      return;
    }

  /* If the widget is not the child of a table or a box, hide the special
     child properties (if we're showing the properties) and return. */
  if (!(GTK_IS_TABLE (parent)
	|| (GTK_IS_BOX (parent) && !GTK_IS_BUTTON_BOX (parent))
	|| GTK_IS_TOOLBAR (parent)
#ifdef GTK_HAVE_FEATURES_1_1_0
	|| GTK_IS_PACKER (parent)
#endif
      ))
    {
      if (data->action == GB_SHOWING)
	property_hide_special_child_page ();
      return;
    }

  /* If we're saving, start the 'child' section. The toolbar outputs this
     itself if needed so we skip it here. */
  if (data->action == GB_SAVING && !GTK_IS_TOOLBAR (parent))
    {
      save_tag (data, "child");
      data->indent++;
    }

  if (GTK_IS_TABLE (parent))
    {
      GtkTableChild *table_child = glade_util_find_table_child (GTK_TABLE (parent),
								widget);
      g_return_if_fail (table_child != NULL);

      if (data->action == GB_SAVING)
	{
	  /* We use left/right/top/bottom_attach here as they are used in GTK */
	  gb_widget_output_int (data, "GtkWidget::left_attach",
				table_child->left_attach);
	  gb_widget_output_int (data, "GtkWidget::right_attach",
				table_child->right_attach);
	  gb_widget_output_int (data, "GtkWidget::top_attach",
				table_child->top_attach);
	  gb_widget_output_int (data, "GtkWidget::bottom_attach",
				table_child->bottom_attach);
	}
      else
	{
	  gb_widget_output_int (data, GbCellX, table_child->left_attach);
	  gb_widget_output_int (data, GbCellY, table_child->top_attach);
	  gb_widget_output_int (data, GbColSpan, table_child->right_attach
				- table_child->left_attach);
	  gb_widget_output_int (data, GbRowSpan, table_child->bottom_attach
				- table_child->top_attach);
	  property_show_special_child_page (0);
	}
      gb_widget_output_int (data, GbXPad, table_child->xpadding);
      gb_widget_output_int (data, GbYPad, table_child->ypadding);
      gb_widget_output_bool (data, GbXExpand, table_child->xexpand);
      gb_widget_output_bool (data, GbYExpand, table_child->yexpand);
      gb_widget_output_bool (data, GbXShrink, table_child->xshrink);
      gb_widget_output_bool (data, GbYShrink, table_child->yshrink);
      gb_widget_output_bool (data, GbXFill, table_child->xfill);
      gb_widget_output_bool (data, GbYFill, table_child->yfill);

    }
  else if (GTK_IS_BOX (parent) && !GTK_IS_BUTTON_BOX (parent))
    {
      gtk_box_query_child_packing (GTK_BOX (parent), widget,
				   &expand, &fill, &padding, &pack_type);
      gb_widget_output_int (data, GbPadding, padding);
      gb_widget_output_bool (data, GbExpand, expand);
      gb_widget_output_bool (data, GbFill, fill);

      if (data->action == GB_SAVING)
	{
	  /* Save pack type as an enum symbol rather than a bool */
	  if (pack_type == GTK_PACK_END)
	    {
	      gb_widget_output_string (data, GbPack, "GTK_PACK_END");
	    }
	}
      else
	{
	  gb_widget_output_bool (data, GbPack, pack_type == GTK_PACK_START);
	  property_show_special_child_page (1);
	}
    }

  else if (GTK_IS_TOOLBAR (parent))
    {
      GtkToolbarChild *child;
      gint pos = 0;

      child = glade_util_find_toolbar_child (parent, widget, &pos, NULL);
      if (pos == 0)
	{
	  if (data->action == GB_SHOWING)
	    {
	      gb_widget_output_bool (data, GbNewToolbarGroup, FALSE);
	      property_set_sensitive (GbNewToolbarGroup, FALSE);
	    }
	}
      else
	{
	  gboolean new_group = gb_widget_get_new_toolbar_group (parent,
								widget);
	  if (data->action == GB_SHOWING)
	    {
	      property_set_sensitive (GbNewToolbarGroup, TRUE);
	      gb_widget_output_bool (data, GbNewToolbarGroup, new_group);
	    }
	  else if (new_group && data->action == GB_SAVING)
	    {
	      save_tag (data, "child");
	      data->indent++;
	      gb_widget_output_bool (data, GbNewToolbarGroup, new_group);
	      data->indent--;
	      save_end_tag (data, "child");
	    }
	}
      if (data->action == GB_SHOWING)
	property_show_special_child_page (2);
    }

#ifdef GTK_HAVE_FEATURES_1_1_0
  else if (GTK_IS_PACKER (parent))
    {
      gint i;
      GtkPackerChild *pchild = glade_util_find_packer_child (GTK_PACKER (parent),
							     widget);
      g_return_if_fail (pchild != NULL);

      for (i = 0; GbPackerSideChoices[i]; i++)
	{
	  if (GbPackerSideValues[i] == pchild->side)
	    gb_widget_output_choice (data, GbPackerSide, i,
				     GbPackerSideSymbols[i]);
	}
      for (i = 0; GbPackerAnchorChoices[i]; i++)
	{
	  if (GbPackerAnchorValues[i] == pchild->anchor)
	    gb_widget_output_choice (data, GbPackerAnchor, i,
				     GbPackerAnchorSymbols[i]);
	}

      gb_widget_output_bool (data, GbPackerExpand,
			     pchild->options & GTK_PACK_EXPAND);
      gb_widget_output_bool (data, GbPackerFillX,
			     pchild->options & GTK_FILL_X);
      gb_widget_output_bool (data, GbPackerFillY,
			     pchild->options & GTK_FILL_Y);
      gb_widget_output_int (data, GbPackerBorder, pchild->border_width);
      gb_widget_output_int (data, GbPackerPadX, pchild->pad_x);
      gb_widget_output_int (data, GbPackerPadY, pchild->pad_y);
      gb_widget_output_int (data, GbPackerIPadX, pchild->i_pad_x);
      gb_widget_output_int (data, GbPackerIPadY, pchild->i_pad_y);

      if (data->action == GB_SAVING)
	{
	}
      else
	{
	  property_show_special_child_page (3);
	}
    }
#endif

  if (data->action == GB_SAVING && !GTK_IS_TOOLBAR (parent))
    {
      data->indent--;
      save_end_tag (data, "child");
    }
}


/* This is used when saving or displaying properties. */
static void
get_lang_specific_properties (GtkWidget * widget, GbWidgetGetArgData * data)
{
  GbWidgetData *wdata = data->widget_data;
  GbWidgetAction action = data->action;

  /* Note that when saving we only save the property if it is not the
     default value. Otherwise we get lots of unnecessary XML output. */

  /* C options. */
  if (action != GB_SAVING || wdata->source_file)
    gb_widget_output_filename (data, GbCSourceFile, wdata->source_file);
  if (action != GB_SAVING || wdata->public_field == 0)
    gb_widget_output_bool (data, GbCPublic, wdata->public_field);
  if (data->action == GB_SHOWING)
    {
      /* We only want the source file set for toplevel widgets. */
      if (GTK_IS_WINDOW (widget) || GTK_IS_MENU (widget))
	property_set_sensitive (GbCSourceFile, TRUE);
      else
	property_set_sensitive (GbCSourceFile, FALSE);
    }

  /* C++ options. */
  if (action != GB_SAVING || wdata->separate_file == 1)
    gb_widget_output_bool (data, GbCppSeparateFile, wdata->separate_file);
  if (action != GB_SAVING || wdata->use_heap == 1)
    gb_widget_output_bool (data, GbCppUseHeap, wdata->use_heap);
  if (action != GB_SAVING || wdata->flatten_hierarchy == 1)
    gb_widget_output_bool (data, GbCppFlatten, wdata->flatten_hierarchy);
}


/* This is used when loading or applying properties. */
static void
set_lang_specific_properties (GtkWidget * widget, GbWidgetSetArgData * data)
{
  GbWidgetData *wdata = data->widget_data;
  gchar *filename;
  gboolean public_field;
  gboolean separate_file, use_heap, flatten_hierarchy;

  /* C options. */
  filename = gb_widget_input_filename (data, GbCSourceFile);
  if (data->apply)
    {
      g_free (wdata->source_file);
      wdata->source_file = g_strdup (filename);
    }
  public_field = gb_widget_input_bool (data, GbCPublic);
  if (data->apply)
    wdata->public_field = public_field ? 1 : 0;

  /* C++ options. */
  separate_file = gb_widget_input_bool (data, GbCppSeparateFile);
  if (data->apply)
    wdata->separate_file = separate_file ? 1 : 0;
  use_heap = gb_widget_input_bool (data, GbCppUseHeap);
  if (data->apply)
    wdata->use_heap = use_heap ? 1 : 0;
  flatten_hierarchy = gb_widget_input_bool (data, GbCppFlatten);
  if (data->apply)
    wdata->flatten_hierarchy = flatten_hierarchy ? 1 : 0;
}


static void
get_position_properties (GtkWidget * widget,
			 GbWidgetGetArgData * data)
{
  GbWidgetData *wdata = data->widget_data;
  gboolean xy_applicable = TRUE, wh_applicable = TRUE;
  gboolean show_xy_buttons = TRUE, show_wh_buttons = TRUE;

  /* Don't bother if widget hasn't been positioned yet, unless it is a
     window. */
  if (data->action == GB_SHOWING && (wdata->flags & GB_INITIAL_EXPOSE)
      && !GTK_IS_WINDOW (widget))
      return;

  MSG3 ("In get_position_properties: %s Width set:%s Height set:%s",
	gtk_widget_get_name (widget),
	wdata->flags & GB_WIDTH_SET ? "1" : "0",
	wdata->flags & GB_HEIGHT_SET ?  "1" : "0");

  if (GTK_IS_MENU (widget))
    {
      if (data->action == GB_SHOWING)
	{
	  gb_widget_output_int (data, GbX, widget->allocation.x);
	  gb_widget_output_int (data, GbY, widget->allocation.y);
	  gb_widget_output_int (data, GbWidth, widget->requisition.width);
	  gb_widget_output_int (data, GbHeight, widget->requisition.height);
	}
      xy_applicable = FALSE;
      wh_applicable = FALSE;
    }
  else if (GTK_IS_WINDOW (widget))
    {
      /* Toplevel window widgets */
      if (data->action == GB_SHOWING)
	{
	  gb_widget_output_int (data, GbX, wdata->x);
	  gb_widget_output_int (data, GbY, wdata->y);
	  gb_widget_output_int (data, GbWidth, wdata->width);
	  gb_widget_output_int (data, GbHeight, wdata->height);
	}
      else
	{
	  if (wdata->flags & GB_X_SET)
	    gb_widget_output_int (data, GbX, wdata->x);
	  if (wdata->flags & GB_Y_SET)
	    gb_widget_output_int (data, GbY, wdata->y);
	  if (wdata->flags & GB_WIDTH_SET)
	    gb_widget_output_int (data, GbWidth, wdata->width);
	  if (wdata->flags & GB_HEIGHT_SET)
	    gb_widget_output_int (data, GbHeight, wdata->height);
	}
      if (GTK_WINDOW (widget)->position != GTK_WIN_POS_NONE)
	xy_applicable = FALSE;
    }
  else if (widget->parent && GTK_IS_FIXED (widget->parent))
    {
      /* Widgets in fixed containers */
      gb_widget_output_int (data, GbX, wdata->x);
      gb_widget_output_int (data, GbY, wdata->y);
      gb_widget_output_int (data, GbWidth, wdata->width);
      gb_widget_output_int (data, GbHeight, wdata->height);
      show_xy_buttons = FALSE;
      show_wh_buttons = FALSE;
    }
  else
    {
      /* Widgets in standard containers */
      if (data->action == GB_SHOWING)
	{
	  gb_widget_output_int (data, GbX, widget->allocation.x);
	  gb_widget_output_int (data, GbY, widget->allocation.y);

	  /* If the width or height has been set explicitly we show the size
	     set, else we show the current requisition. We always remember
	     what we have shown in wdata->width & height so we know if the
	     user changes it. */
	  if (!(wdata->flags & GB_WIDTH_SET))
	    wdata->width = widget->requisition.width;
	  gb_widget_output_int (data, GbWidth, wdata->width);

	  if (!(wdata->flags & GB_HEIGHT_SET))
	    wdata->height = widget->requisition.height;
	  gb_widget_output_int (data, GbHeight, wdata->height);

	}
      else
	{
	  /* Only save if user has set it explicitly. */
	  if (wdata->flags & GB_WIDTH_SET)
	    gb_widget_output_int (data, GbWidth, wdata->width);
	  if (wdata->flags & GB_HEIGHT_SET)
	    gb_widget_output_int (data, GbHeight, wdata->height);
	}
      xy_applicable = FALSE;
    }

  /* Show the buttons for setting the size & positions explicitly, if
     applicable, and set the values sensitive if they are currently set. */
  if (data->action == GB_SHOWING)
    {
      if (xy_applicable)
	{
	  gboolean xsensitive = (wdata->flags & GB_X_SET) ? TRUE : FALSE;
	  gboolean ysensitive = (wdata->flags & GB_Y_SET) ? TRUE : FALSE;

	  property_set_sensitive_full (GbX, TRUE, xsensitive, show_xy_buttons);
	  property_set_sensitive_full (GbY, TRUE, ysensitive, show_xy_buttons);
	}
      else
	{
	  property_set_sensitive_full (GbX, FALSE, FALSE, FALSE);
	  property_set_sensitive_full (GbY, FALSE, FALSE, FALSE);
	}

      if (wh_applicable)
	{
	  gboolean wsensitive = (wdata->flags & GB_WIDTH_SET) ? TRUE : FALSE;
	  gboolean hsensitive = (wdata->flags & GB_HEIGHT_SET) ? TRUE : FALSE;

	  property_set_sensitive_full (GbWidth, TRUE, wsensitive,
				       show_wh_buttons);
	  property_set_sensitive_full (GbHeight, TRUE, hsensitive,
				       show_wh_buttons);
	}
      else
	{
	  property_set_sensitive_full (GbWidth, FALSE, FALSE, FALSE);
	  property_set_sensitive_full (GbHeight, FALSE, FALSE, FALSE);
	}
    }
}


static void
save_style (GtkWidget * widget, GbWidgetGetArgData * data)
{
  GbWidgetData *wdata = data->widget_data;
  g_return_if_fail (wdata != NULL);

  /* Don't save the style if it's the default. */
  if (wdata->gbstyle != gb_widget_default_gb_style)
    {
      /* If the widget is using a named style, just save the name. */
      if (!(wdata->flags & GB_STYLE_IS_UNNAMED))
	{
	  save_string (data, GbStyleName, wdata->gbstyle->name);
	}
      else
	{
	  gb_widget_save_style (wdata->gbstyle, data, FALSE);
	}
    }
  /* The default propagate value is TRUE. */
  if (!(wdata->flags & GB_STYLE_PROPAGATE))
    save_bool (data, GbStylePropagate, wdata->flags & GB_STYLE_PROPAGATE);
}


void
gb_widget_save_style (GbStyle * gbstyle, GbWidgetGetArgData * data,
		      gboolean save_all)
{
  GtkStyle *style = gbstyle->style;
  GtkStyle *default_style = gb_widget_default_gb_style->style;
  gchar buffer[128];
  gint i;

  save_tag (data, "style");
  data->indent++;

  if (gbstyle->name)
    save_string (data, GbStyleName, gbstyle->name);
  if (save_all
      || strcmp (gbstyle->xlfd_fontname, gb_widget_default_gb_style->xlfd_fontname))
    {
      save_string (data, GbStyleFont, gbstyle->xlfd_fontname);
    }

  /* Colors */
  default_style = gb_widget_default_gb_style->style;
  save_colors (data, save_all, style->fg, default_style->fg, "fg");
  save_colors (data, save_all, style->bg, default_style->bg, "bg");
  save_colors (data, save_all, style->text, default_style->text, "text");
  save_colors (data, save_all, style->base, default_style->base, "base");

  /* Background pixmaps */
  for (i = 0; i < GB_NUM_STYLE_STATES; i++)
    {
      /* Can't use '[' or ']' in XML element names */
      sprintf (buffer, "GtkStyle::%s-%s", GbBgPixmapName, GbStateNames[i]);
      if (gbstyle->bg_pixmap_filenames[i])
	{
	  if (save_all
	      || gb_widget_default_gb_style->bg_pixmap_filenames[i] == NULL
	      || strcmp (gbstyle->bg_pixmap_filenames[i],
			 gb_widget_default_gb_style->bg_pixmap_filenames[i]))
	    {
	      save_string (data, buffer, gbstyle->bg_pixmap_filenames[i]);
	    }
	}
    }

  data->indent--;
  save_tag (data, "/style");
}


static void
save_colors (GbWidgetGetArgData * data, gboolean save_all, GdkColor colors[],
	     GdkColor default_colors[], gchar * name)
{
  gint state;
  gchar buf[128];

  for (state = 0; state < GB_NUM_STYLE_STATES; state++)
    {
      if (save_all
	  || colors[state].red != default_colors[state].red
	  || colors[state].green != default_colors[state].green
	  || colors[state].blue != default_colors[state].blue)
	{
	  /* Can't use ':', '[' or ']' in XML element names */
	  sprintf (buf, "GtkStyle::%s-%s", name, GbStateNames[state]);
	  save_color (data, buf, &colors[state]);
	}
    }
}


static void
save_accelerators (GtkWidget * widget, GbWidgetGetArgData * data)
{
  GList *item;
  GbAccelerator *accel;
  gchar key_buffer[128];

  if (data->widget_data == NULL)
    return;
  item = data->widget_data->accelerators;
  while (item)
    {
      accel = (GbAccelerator *) item->data;

      save_tag (data, "accelerator");
      data->indent++;
      save_string (data, "modifiers",
		   create_modifiers_string (accel->modifiers));

      /* Make sure we don't overflow the buffer. */
      if (strlen (accel->key) < 100)
	{
	  strcpy (key_buffer, "GDK_");
	  strcat (key_buffer, accel->key);
	  save_string (data, "key", key_buffer);
	}

      save_string (data, "signal", accel->signal);
      data->indent--;
      save_tag (data, "/accelerator");
      item = item->next;
    }
}

/* This creates a string representing the modifiers flags, e.g.
   "GDK_CONTROL_MASK | GDK_SHIFT_MASK". It is used for saving the XML, and
   also for writing the source code. It uses a static buffer. */
static gchar*
create_modifiers_string (guint8 modifier_flags)
{
  static gchar modifiers[128];

  modifiers[0] = '\0';
  if (modifier_flags == 0)
    {
      strcat (modifiers, "0");
      return modifiers;
    }

  if (modifier_flags & GDK_CONTROL_MASK)
    {
      strcat (modifiers, "GDK_CONTROL_MASK");
    }
  if (modifier_flags & GDK_SHIFT_MASK)
    {
      if (modifiers[0] != '\0')
	strcat (modifiers, " | ");
      strcat (modifiers, "GDK_SHIFT_MASK");
    }
  if (modifier_flags & GDK_MOD1_MASK)
    {
      if (modifiers[0] != '\0')
	strcat (modifiers, " | ");
      strcat (modifiers, "GDK_MOD1_MASK");
    }
  return modifiers;
}

static void
save_signals (GtkWidget * widget, GbWidgetGetArgData * data)
{
  GList *item;
  GbSignal *signal;

  if (data->widget_data == NULL)
    return;
  item = data->widget_data->signals;
  while (item)
    {
      signal = (GbSignal *) item->data;

      save_tag (data, "signal");
      data->indent++;
      if (signal->name)
	save_string (data, "name", signal->name);
      if (signal->handler)
	save_string (data, "handler", signal->handler);
      if (signal->object)
	save_string (data, "object", signal->object);
      if (signal->after)
	save_bool (data, "after", signal->after);
      if (signal->data)
	save_string (data, "data", signal->data);
      if (signal->last_modification_time != 0)
	save_date (data, "last_modification_time",
		   &signal->last_modification_time);
      data->indent--;
      save_tag (data, "/signal");

      item = item->next;
    }
}



/*************************************************************************
 * Functions for writing C source code
 *************************************************************************/

void
gb_widget_write_source (GtkWidget * widget,
			GbWidgetWriteSourceData * data)
{
  GtkWidget *parent;
  GbWidget *gbwidget;
  GbWidgetData *widget_data;
  gchar *class_name, *child_name;
  gchar *component_key;

  component_key = gtk_object_get_data (GTK_OBJECT (widget), GB_COMPONENT_KEY);
  if ((component_key) && (!strcmp (component_key, "True")))
  {
    source_write_subcomponent_create (widget, data);

    printf("not writing source to create component\n");
    return;
  }

  /* This is a temporary kludge so that the code GtkDialogs is OK.
     We stop the source code for the action_area from being written. */
  child_name = gtk_object_get_data (GTK_OBJECT (widget), GB_CHILD_NAME_KEY);
  if (child_name && !strcmp (child_name, "Dialog:action_area")
      && data->create_widget)
    return;

  parent = data->parent;

  class_name = gtk_type_name (GTK_OBJECT_TYPE (widget));

  widget_data = gtk_object_get_data (GTK_OBJECT (widget), GB_WIDGET_DATA_KEY);
  /* If this isn't a gbwidget, skip it. */
  if (widget_data)
    {
      gbwidget = (GbWidget *) g_hash_table_lookup (gb_widget_table, class_name);
      g_return_if_fail (gbwidget != NULL);

      data->standard_widget
	= (GtkWidget *) g_hash_table_lookup (data->standard_widgets, class_name);
      if (data->standard_widget == NULL)
	{
	  data->standard_widget = GTK_WIDGET (gtk_object_newv (gtk_type_from_name (class_name), 0, NULL));
	  g_hash_table_insert (data->standard_widgets, class_name,
			       data->standard_widget);
	}

      data->wname = source_create_valid_identifier (gtk_widget_get_name (widget));
      data->widget_data = widget_data;
      data->write_children = TRUE;
      if (gbwidget->gb_widget_write_source)
	(gbwidget->gb_widget_write_source) (widget, data);
      else
	source_add (data, "  /* Skipping %s: unimplemented. */\n", class_name);

      /* Make sure there is a blank line after each widget, for readability. */
      if (data->buffer.pos > 2
	  && (data->buffer.ptr[data->buffer.pos - 1] != '\n'
	      || data->buffer.ptr[data->buffer.pos - 2] != '\n'))
	source_add (data, "\n");
      g_free (data->wname);
      data->wname = NULL;
      data->parent = widget;
    }
  else if (GB_IS_PLACEHOLDER (widget) && parent)
    {
      if (GTK_IS_NOTEBOOK (parent))
	{
	  /* SPECIAL CODE: If notebook pages are empty (i.e. placeholders),
	     we create dummy widgets instead, so it still runs OK. */
	  if (child_name == NULL)
	    {
	      gchar *wname, *parent_name;
	  
	      wname = "empty_notebook_page";
	  
	      /* We need to make sure that the dummy widget declaration is only
		 ever output once. */
	      if (!glade_util_strstr (data->decl_buffer.ptr, wname))
		source_add_decl (data, "  GtkWidget *%s;\n", wname);
	  
	      parent_name = gtk_widget_get_name (parent);
	      parent_name = source_create_valid_identifier (parent_name);
	      source_add (data,
			  "  %s = gtk_vbox_new (FALSE, 0);\n"
			  "  gtk_widget_show (%s);\n"
			  "  gtk_container_add (GTK_CONTAINER (%s), %s);\n"
			  "\n",
			  wname, wname, parent_name, wname);
	      g_free (parent_name);
	    }
	  else
	    {
	      /* For empty notebook tabs, we increment the 'last_child' written
		 value. */
	      gint col = GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT (parent),
							       "last_child"));
	      gtk_object_set_data (GTK_OBJECT (parent), "last_child",
				   GINT_TO_POINTER (col + 1));
	    }
	}
      else if (GTK_IS_CLIST (parent))
	{
	  /* For empty clist/ctree titles, we increment the 'last_child'
	     written value. */
	  if (child_name && (!strcmp (child_name, "CList:title")
			     || !strcmp (child_name, "CTree:title")))
	    {
	      gint col = GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT (parent),
							       "last_child"));
	      gtk_object_set_data (GTK_OBJECT (parent), "last_child",
				   GINT_TO_POINTER (col + 1));
	    }
	}
    }


  /* Recursively write source for children.
     We need to reset the parent after the children have been written. */
  data->create_widget = TRUE;
  if (data->write_children)
    gb_widget_children_foreach (widget, (GtkCallback) gb_widget_write_source, data);

  /* We need to reset some of the members of the GbWidgetWriteSourceData struct
     so that they work OK for the next sibling. */
  data->parent = parent;
}


/* This is called by each GbWidget's write_source function to write all the
   common source code, including code to add the widget to its parent. */
void
gb_widget_write_standard_source (GtkWidget * widget,
				 GbWidgetWriteSourceData * data)
{
  GbWidgetData *wdata = data->widget_data;
  gint i, width, height, can_focus, can_default;

  source_add_decl (data, "  GtkWidget *%s;\n", data->wname);

  if (data->set_widget_names)
    source_add (data, "  gtk_widget_set_name (%s, \"%s\");\n", data->wname,
		data->wname);

  if (data->use_widget_hash)
    source_add (data, "  gtk_object_set_data (GTK_OBJECT (%s), %s, %s);\n",
		data->component_name, source_make_string (data->wname, FALSE),
		data->wname);

  /* SPECIAL CODE: menus are attached to the parent menu items. */
  if (GTK_IS_MENU (widget))
    {
      if (gtk_menu_get_attach_widget (GTK_MENU (widget)))
	{
	  MSG1 ("widget %s is a menu with a parent", data->wname);
	  gb_widget_write_add_child_source (widget, data);
	}
    }
  else if (widget->parent)
    {
      if (wdata->flags & GB_VISIBLE)
	source_add (data, "  gtk_widget_show (%s);\n", data->wname);

      /* Output code to add widget to parent. */
      gb_widget_write_add_child_source (widget, data);
    }

  if (wdata->flags & (GB_X_SET | GB_Y_SET))
    {
      source_add (data, "  gtk_widget_set_uposition (%s, %i, %i);\n",
		  data->wname,
		  wdata->flags & GB_X_SET ? wdata->x : -2,
		  wdata->flags & GB_Y_SET ? wdata->y : -2);
    }

  if (wdata->flags & (GB_WIDTH_SET | GB_HEIGHT_SET))
    {
      width = wdata->flags & GB_WIDTH_SET ? wdata->width : -2;
      height = wdata->flags & GB_HEIGHT_SET ? wdata->height : -2;

      /* GTK BUG WORKAROUND - a combo should manage the size of its entry. */
      if (GTK_IS_COMBO (widget))
	source_add(data,
		   "  gtk_widget_set_usize (GTK_COMBO (%s)->entry, %i, %i);\n",
		   data->wname, width - 16 < 0 ? -2 : width - 16, height);

      source_add (data, "  gtk_widget_set_usize (%s, %i, %i);\n",
		  data->wname, width, height);
    }


  if (GTK_IS_CONTAINER (widget))
    {
      if (GTK_CONTAINER (widget)->border_width != 0)
	{
	  source_add (data,
		 "  gtk_container_border_width (GTK_CONTAINER (%s), %i);\n",
		      data->wname, GTK_CONTAINER (widget)->border_width);
	}
    }


  if (!(wdata->flags & GB_SENSITIVE))
    source_add (data, "  gtk_widget_set_sensitive (%s, FALSE);\n", data->wname);

  can_focus = GTK_WIDGET_CAN_FOCUS (widget);
  if (can_focus != GTK_WIDGET_CAN_FOCUS (data->standard_widget))
    {
      if (can_focus)
	source_add (data, "  GTK_WIDGET_SET_FLAGS (%s, GTK_CAN_FOCUS);\n",
		    data->wname);
      /* SPECIAL CODE: toolbar widgets can't get focus by default. */
      else if (!gb_widget_is_toolbar_button (widget))
	source_add (data, "  GTK_WIDGET_UNSET_FLAGS (%s, GTK_CAN_FOCUS);\n",
		    data->wname);
    }
  can_default = GTK_WIDGET_CAN_DEFAULT (widget);
  if (can_default != GTK_WIDGET_CAN_DEFAULT (data->standard_widget))
    {
      if (can_default)
	source_add (data, "  GTK_WIDGET_SET_FLAGS (%s, GTK_CAN_DEFAULT);\n",
		    data->wname);
      else
	source_add (data, "  GTK_WIDGET_UNSET_FLAGS (%s, GTK_CAN_DEFAULT);\n",
		    data->wname);
    }

  if (wdata->flags & GB_GRAB_FOCUS)
    source_add (data, "  gtk_widget_grab_focus (%s);\n", data->wname);
  if (wdata->flags & GB_GRAB_DEFAULT)
    source_add (data, "  gtk_widget_grab_default (%s);\n", data->wname);

  /* SPECIAL CODE: toolbar children use the toolbar's tooltips. */
  if (wdata->tooltip && (!widget->parent || !GTK_IS_TOOLBAR (widget->parent)))
    {
      data->need_tooltips = TRUE;
      source_add (data, "  gtk_tooltips_set_tip (tooltips, %s, %s, NULL);\n",
		  data->wname,
		  source_make_string (wdata->tooltip, data->use_gettext));
    }

  if (!GTK_WIDGET_NO_WINDOW (widget))
    {
      GdkExtensionMode ext_mode = gtk_widget_get_extension_events (widget);
      gboolean written_first = FALSE;

      if (wdata->events)
	{
	  source_add (data, "  gtk_widget_set_events (%s, ", data->wname);
	  for (i = 0; i < GB_EVENT_MASKS_COUNT; i++)
	    {
	      if (wdata->events & GbEventMaskValues[i])
		{
		  if (!written_first)
		    source_add (data, "%s", GbEventMaskSymbols[i]);
		  else
		    source_add (data, " | %s", GbEventMaskSymbols[i]);
		  written_first = TRUE;
		}
	    }
	  source_add (data, ");\n");
	}

      if (ext_mode != GDK_EXTENSION_EVENTS_NONE)
	{
	  for (i = 0; GbExtensionModeChoices[i]; i++)
	    {
	      if (GbExtensionModeValues[i] == ext_mode)
		source_add (data, "  gtk_widget_set_extension_events (%s, %s);\n",
			    data->wname, GbExtensionModeSymbols[i]);
	    }
	}
    }

  gb_widget_write_signals_source (widget, data);
  gb_widget_write_accelerators_source (widget, data);
}


void
gb_widget_write_add_child_source (GtkWidget * widget,
				  GbWidgetWriteSourceData * data)
{
  GbWidgetData *wdata = data->widget_data;
  GtkWidget *parent;
  gchar *parent_name;

  /* If the widget is created automatically by its parent, we don't need
     to add source to add it. */
  if (!data->create_widget)
    return;

  /* SPECIAL CODE: to handle menus. */
  if (GTK_IS_MENU (widget))
    parent = gtk_menu_get_attach_widget (GTK_MENU (widget));
  else
    parent = data->parent;

  /* Return if no parent exists */
  if (!parent)
    return;

  parent_name = gtk_widget_get_name (parent);
  parent_name = source_create_valid_identifier (parent_name);

  MSG2 ("Adding %s to %s", data->wname, parent_name);

  if (GTK_IS_TABLE (parent))
    {
      gchar xoptions[48], yoptions[48];
      GtkTableChild *tchild = glade_util_find_table_child (GTK_TABLE (parent),
							   widget);
      g_return_if_fail (tchild != NULL);

      xoptions[0] = yoptions[0] = '\0';
      if (tchild->xexpand)
	strcpy (xoptions, "GTK_EXPAND");
      if (tchild->xshrink)
	{
	  if (xoptions[0] != '\0')
	    strcat (xoptions, " | ");
	  strcat (xoptions, "GTK_SHRINK");
	}
      if (tchild->xfill)
	{
	  if (xoptions[0] != '\0')
	    strcat (xoptions, " | ");
	  strcat (xoptions, "GTK_FILL");
	}
      if (xoptions[0] == '\0')
	strcpy (xoptions, "0");

      if (tchild->yexpand)
	strcpy (yoptions, "GTK_EXPAND");
      if (tchild->yshrink)
	{
	  if (yoptions[0] != '\0')
	    strcat (yoptions, " | ");
	  strcat (yoptions, "GTK_SHRINK");
	}
      if (tchild->yfill)
	{
	  if (yoptions[0] != '\0')
	    strcat (yoptions, " | ");
	  strcat (yoptions, "GTK_FILL");
	}
      if (yoptions[0] == '\0')
	strcpy (yoptions, "0");

      source_add (data,
		  "  gtk_table_attach (GTK_TABLE (%s), %s, %i, %i, %i, %i,\n"
		  "                    (GtkAttachOptions) %s, (GtkAttachOptions) %s, %i, %i);\n",
		  parent_name, data->wname,
		  tchild->left_attach, tchild->right_attach,
		  tchild->top_attach, tchild->bottom_attach,
		  xoptions, yoptions, tchild->xpadding, tchild->ypadding);
    }
  else if (GTK_IS_BOX (parent) && !GTK_IS_BUTTON_BOX (parent))
    {
#ifdef GTK_HAVE_FEATURES_1_1_0
      gboolean expand, fill;
      guint padding;
#else
      gint expand, fill, padding;
#endif
      GtkPackType pack_type;

      gtk_box_query_child_packing (GTK_BOX (parent), widget,
				   &expand, &fill, &padding, &pack_type);
      if (pack_type == GTK_PACK_START)
	{
	  source_add (data,
		   "  gtk_box_pack_start (GTK_BOX (%s), %s, %s, %s, %i);\n",
		      parent_name, data->wname,
	       expand ? "TRUE" : "FALSE", fill ? "TRUE" : "FALSE", padding);
	}
      else
	{
	  source_add (data,
		      "  gtk_box_pack_end (GTK_BOX (%s), %s, %s, %s, %i);\n",
		      parent_name, data->wname,
	       expand ? "TRUE" : "FALSE", fill ? "TRUE" : "FALSE", padding);
	}
    }
#ifdef GTK_HAVE_FEATURES_1_1_0
  else if (GTK_IS_PACKER (parent))
    {
      gchar *side = NULL, *anchor = NULL, options[48];
      gint i;
      GtkPackerChild *pchild = glade_util_find_packer_child (GTK_PACKER (parent),
							     widget);
      g_return_if_fail (pchild != NULL);

      for (i = 0; GbPackerSideChoices[i]; i++)
	{
	  if (GbPackerSideValues[i] == pchild->side)
	    side = GbPackerSideSymbols[i];
	}
      for (i = 0; GbPackerAnchorChoices[i]; i++)
	{
	  if (GbPackerAnchorValues[i] == pchild->anchor)
	    anchor = GbPackerAnchorSymbols[i];
	}
      options[0] = '\0';
      if (pchild->options & GTK_PACK_EXPAND)
	strcpy (options, "GTK_PACK_EXPAND");
      if (pchild->options & GTK_FILL_X)
	{
	  if (options[0] != '\0')
	    strcat (options, " | ");
	  strcat (options, "GTK_FILL_X");
	}
      if (pchild->options & GTK_FILL_Y)
	{
	  if (options[0] != '\0')
	    strcat (options, " | ");
	  strcat (options, "GTK_FILL_Y");
	}
      if (options[0] == '\0')
	strcpy (options, "0");

      source_add (data,
		  "  gtk_packer_add (GTK_PACKER (%s), %s, %s,\n"
		  "                  %s, %s,\n"
		  "                  %i, %i, %i, %i, %i);\n",
		  parent_name, data->wname, side, anchor, options,
		  pchild->border_width, pchild->pad_x, pchild->pad_y,
		  pchild->i_pad_x, pchild->i_pad_y);
    }
#endif
  else if (GTK_IS_FIXED (parent))
    {
      source_add (data, "  gtk_fixed_put (GTK_FIXED (%s), %s, %i, %i);\n",
		  parent_name, data->wname,
		  widget->allocation.x, widget->allocation.y);

    }
  else if (GTK_IS_CLIST (parent))
    {
      /* See if this is a title widget. */
      gchar *child_name = gtk_object_get_data (GTK_OBJECT (widget),
					       GB_CHILD_NAME_KEY);
      MSG ("Source:parent is clist");
      /* FIXME: This should use CListTitle from gbclist.c */
      if (child_name && (!strcmp (child_name, "CList:title")
			 || !strcmp (child_name, "CTree:title")))
	{
	  /* We store the last column title written in 'last_child' */
	  gint col = GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT (parent),
							   "last_child"));
	  source_add (data,
		"  gtk_clist_set_column_widget (GTK_CLIST (%s), %i, %s);\n",
		      parent_name, col + 1, data->wname);

	  gtk_object_set_data (GTK_OBJECT (parent), "last_child",
			       GINT_TO_POINTER (col + 1));
	}
      else
	{
	  source_add (data, "  gtk_container_add (GTK_CONTAINER (%s), %s);\n",
		      parent_name, data->wname);
	}
    }
  else if (GTK_IS_NOTEBOOK (parent))
    {
      /* See if this is a notebook tab widget. */
      gchar *child_name = gtk_object_get_data (GTK_OBJECT (widget),
					       GB_CHILD_NAME_KEY);
      MSG ("Source:parent is notebook");
      /* FIXME: This should use NotebookTab from gbnotebook.c */
      if (child_name && !strcmp (child_name, "Notebook:tab"))
	{
	  /* We store the last tab written in 'last_child' */
	  gint col = GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT (parent),
							   "last_child"));
	  /* We use our own function to add the tab to the notebook. */
	  source_add (data,
		      "  set_notebook_tab (%s, %i, %s);\n",
		      parent_name, col + 1, data->wname);

	  gtk_object_set_data (GTK_OBJECT (parent), "last_child",
			       GINT_TO_POINTER (col + 1));
	}
      else
	{
	  source_add (data, "  gtk_container_add (GTK_CONTAINER (%s), %s);\n",
		      parent_name, data->wname);
	}
    }
  else if (GTK_IS_MENU_ITEM (parent) && GTK_IS_MENU (widget))
    {
      MSG ("Source:parent is a menu item and widget is a menu");
      source_add (data,
		  "  gtk_menu_item_set_submenu (GTK_MENU_ITEM (%s), %s);\n",
		  parent_name, data->wname);

    }
  else if (GTK_IS_TOOLBAR (parent))
    {
      /* For toolbar buttons, the widgets are added to the toolbar when
	 created, so we skip this. For standard widgets, we add them here. */
      if (!gb_widget_is_toolbar_button (widget))
	{
	  source_add (data,
		      "  gtk_toolbar_append_widget (GTK_TOOLBAR (%s), %s, %s, NULL);\n",
		      parent_name, data->wname,
		      wdata->tooltip ? source_make_string (wdata->tooltip, data->use_gettext) : "NULL");
	}
    }
  else
    {
      source_add (data, "  gtk_container_add (GTK_CONTAINER (%s), %s);\n",
		  parent_name, data->wname);
    }
  g_free (parent_name);
}


void
gb_widget_write_signals_source (GtkWidget * widget,
				GbWidgetWriteSourceData * data)
{
  GList *item = data->widget_data->signals;
  GbSignal *signal;
  guint signal_id;
  GtkSignalQuery *query_info;
  gint param, i;
  gchar buffer[1024];
  gchar *ret_type, *pos, *type_name, *arg_name, *object_name, *object_arg;
  gchar *object_arg_start;
  gboolean is_pointer;
  gint param_num, widget_num, event_num, callback_num;
  gint handler_len, object_name_len, type_name_len, num_spaces;
  gint *arg_num;
  gchar *handler, **arg_names;
  gchar signal_name[128];
  gboolean skip_handler_code;

  while (item)
    {
      signal = (GbSignal *) item->data;
      item = item->next;
      param_num = 1;
      widget_num = event_num = callback_num = 0;

      /* If we're appending new/changed signals and this is an old one, then
	 skip it. Note that we also skip it if the last mod time is 0, which
	 will happen for XML files created before the last mod tag was added.*/
      MSG2 ("Creating Signals:%s Handler: %s",
	    data->creating_signals_files ? "TRUE" : "FALSE", signal->handler);
      MSG1 ("LastMod:%s", ctime (&signal->last_modification_time));
      MSG1 ("LastWrite:%s", ctime (&data->last_write_time));

      if (!data->creating_signals_files
	  && (signal->last_modification_time <= data->last_write_time))
	{
	  MSG1 ("Skipping signal: %s", signal->handler);
	  skip_handler_code = TRUE;
	}
      else
	skip_handler_code = FALSE;

      if (!signal->name)
	{
	  /* FIXME: store warnings */
	  g_warning ("Signal name not set");
	  continue;
	}

      if (!signal->handler)
	{
	  /* FIXME: store warnings */
	  g_warning ("Signal handler not set");
	  continue;

	}

      signal_id = gtk_signal_lookup (signal->name, GTK_OBJECT_TYPE (widget));
      if (signal_id == 0)
	{
	  /* FIXME: store warnings */
	  g_warning ("Signal unknown");
	  continue;
	}

      /* Make sure handler function name is valid. */
      handler = source_create_valid_identifier (signal->handler);

      /* Output code to connect signal. */
      if (signal->object)
	{
	  if (signal->after)
	    {
	      source_add (data,
	      "  gtk_signal_connect_object_after (GTK_OBJECT (%s), \"%s\",\n"
		"                                   GTK_SIGNAL_FUNC (%s),\n"
		   "                                   GTK_OBJECT (%s));\n",
			data->wname, signal->name, handler, signal->object);
	    }
	  else
	    {
	      source_add (data,
		   "  gtk_signal_connect_object (GTK_OBJECT (%s), \"%s\",\n"
		      "                             GTK_SIGNAL_FUNC (%s),\n"
			  "                             GTK_OBJECT (%s));\n",
			data->wname, signal->name, handler, signal->object);
	    }
	}
      else
	{
	  if (signal->after)
	    {
	      source_add (data,
		    "  gtk_signal_connect_after (GTK_OBJECT (%s), \"%s\",\n"
		       "                            GTK_SIGNAL_FUNC (%s),\n"
			  "                            %s);\n",
			  data->wname, signal->name, handler,
			  signal->data ? signal->data : "NULL");
	    }
	  else
	    {
	      source_add (data,
			  "  gtk_signal_connect (GTK_OBJECT (%s), \"%s\",\n"
			  "                      GTK_SIGNAL_FUNC (%s),\n"
			  "                      %s);\n",
			  data->wname, signal->name, handler,
			  signal->data ? signal->data : "NULL");
	    }
	}

      /* We don't need to output code for standard GTK functions. */
      if (!strcmp(handler, "gtk_widget_destroy"))
 	{
	  /*printf("gtk_widget_destroy is a reserved function name.\n");*/
	  g_free (handler);
	  continue;
	}

      /* If we're appending new/changed signals and this is an old one, we
	 don't want to output the source code for it. */
      if (skip_handler_code)
	{
	  g_free (handler);
	  continue;
	}

      /* Output empty signal handler. */
      query_info = gtk_signal_query (signal_id);
      if (query_info == NULL)
	{
	  g_warning ("Couldn't query signal");
	  g_free (handler);
	  continue;
	}

      /* Output the return type and function name. */
      ret_type = get_type_name (query_info->return_val, &is_pointer);
      pos = buffer;
      sprintf (pos, "%s%s\n%s", ret_type, is_pointer ? "*" : "", handler);
      pos += strlen (pos);

      handler_len = strlen (handler);
      if (handler_len >= GB_PARAM_INDENT - 1)
	{
	  *pos++ = '\n';
	  for (i = 0; i < GB_PARAM_INDENT; i++)
	    *pos++ = ' ';
	}
      else
	{
	  num_spaces = GB_PARAM_INDENT - handler_len - 1;
	  for (i = 0; i < num_spaces; i++)
	    *pos++ = ' ';
	}

      /* Output the signal object type and the argument name. We assume the
	 type is a pointer - I think that is OK. We remove "Gtk" and convert
	 to lower case for the argument name. */
      object_name = gtk_type_name (query_info->object_type);
      sprintf (pos, "(%s ", object_name);
      pos += strlen (pos);
      object_name_len = strlen (object_name);
      if (object_name_len + 1 < GB_PARAM_TYPE_WIDTH)
	{
	  num_spaces = GB_PARAM_TYPE_WIDTH - object_name_len - 1;
	  for (i = 0; i < num_spaces; i++)
	    *pos++ = ' ';
	}
      object_arg = (!strncmp (object_name, "Gtk", 3)) ? object_name + 3
	: object_name;
      object_arg_start = pos;
      sprintf (pos, "*%s", object_arg);
      pos += strlen (pos);
      g_strdown (object_arg_start);
      if (!strcmp (object_arg_start, "widget"))
	widget_num++;
      sprintf (pos, ",\n");
      pos += strlen (pos);

      /* Convert signal name to use underscores rather than dashes '-'. */
      strcpy (signal_name, query_info->signal_name);
      for (i = 0; signal_name[i]; i++)
	{
	  if (signal_name[i] == '-')
	    signal_name[i] = '_';
	}

      /* Output the signal parameters. */
      arg_names = lookup_signal_arg_names (object_name, signal_name);

      for (param = 0; param < query_info->nparams; param++)
	{
	  for (i = 0; i < GB_PARAM_INDENT; i++)
	    *pos++ = ' ';

	  if (arg_names)
	    {
	      sprintf (pos, "%s,\n", arg_names[param]);
	      pos += strlen (pos);
	    }
	  else
	    {
	      type_name = get_type_name (query_info->params[param], &is_pointer);
	      /* Most arguments to the callback are called "arg1", "arg2", etc.
	         GtkWidgets are called "widget", "widget2", ...
	         GdkEvents are called "event", "event2", ...
	         GtkCallbacks are called "callback", "callback2", ... */
	      if (!strcmp (type_name, "GtkWidget"))
		{
		  arg_name = "widget";
		  arg_num = &widget_num;
		}
	      else if (!strcmp (type_name, "GdkEvent"))
		{
		  type_name = get_gdk_event (signal->name);
		  arg_name = "event";
		  arg_num = &event_num;
		  is_pointer = TRUE;
		}
	      else if (!strcmp (type_name, "GtkCallback"))
		{
		  arg_name = "callback";
		  arg_num = &callback_num;
		}
	      else
		{
		  arg_name = "arg";
		  arg_num = &param_num;
		}
	      sprintf (pos, "%s ", type_name);
	      pos += strlen (pos);
	      type_name_len = strlen (type_name);
	      if (type_name_len + 1 < GB_PARAM_TYPE_WIDTH)
		{
		  num_spaces = GB_PARAM_TYPE_WIDTH - type_name_len - 1;
		  for (i = 0; i < num_spaces; i++)
		    *pos++ = ' ';
		}
	      if (!arg_num || *arg_num == 0)
		sprintf (pos, "%s%s,\n", is_pointer ? "*" : " ", arg_name);
	      else
		sprintf (pos, "%s%s%i,\n", is_pointer ? "*" : " ", arg_name,
			 *arg_num);
	      pos += strlen (pos);

	      if (arg_num)
		*arg_num += 1;
	    }
	}
      for (i = 0; i < GB_PARAM_INDENT; i++)
	*pos++ = ' ';
      sprintf (pos, "gpointer         user_data)");
      pos += strlen (pos);

      /* return FALSE from the handler if the return type is bool. */
      fprintf (data->sigcfp, "\n%s\n{\n\n%s}\n\n", buffer,
	       query_info->return_val == GTK_TYPE_BOOL ? "  return FALSE;\n"
	       : "");
      fprintf (data->sighfp, "\n%s;\n", buffer);

      g_free (query_info);
      g_free (handler);
    }
}


static gchar *
get_type_name (GtkType type, gboolean * is_pointer)
{
  static gchar *GbTypeNames[] =
  {
    "char", "gchar",
    "bool", "gboolean",
    "int", "gint",
    "uint", "guint",
    "long", "glong",
    "ulong", "gulong",
    "float", "gfloat",
    "double", "gdouble",
    "string", "gchar",
    "enum", "gint",
    "flags", "gint",
    "boxed", "gpointer",
    "foreign", "gpointer",
    "callback", "GtkCallback",	/* ?? */
    "args", "gpointer",

    "pointer", "gpointer",
    "signal", "gpointer",
    "c_callback", "GtkCallback",	/* ?? */

    NULL
  };

  GtkType parent_type;
  gchar *type_name, *parent_type_name;
  gint i;

  *is_pointer = FALSE;
  type_name = gtk_type_name (type);
  for (i = 0; GbTypeNames[i]; i += 2)
    {
      if (!strcmp (type_name, GbTypeNames[i]))
	{
	  if (!strcmp (type_name, "string"))
	    *is_pointer = TRUE;
	  return GbTypeNames[i + 1];
	}
    }

  for (;;)
    {
      parent_type = gtk_type_parent (type);
      if (parent_type == 0)
	break;
      type = parent_type;
    }
  parent_type_name = gtk_type_name (type);
  if (!strcmp (parent_type_name, "GtkObject")
      || !strcmp (parent_type_name, "boxed")
      || !strcmp (parent_type_name, "GtkBoxed"))
    *is_pointer = TRUE;

  return type_name;
}


static gchar *
get_gdk_event (gchar * signal_name)
{
  static gchar *GbGDKEvents[] =
  {
    "button_press_event", "GdkEventButton",
    "button_release_event", "GdkEventButton",
    "motion_notify_event", "GdkEventMotion",
    "delete_event", "GdkEvent",
    "destroy_event", "GdkEvent",
    "expose_event", "GdkEventExpose",
    "key_press_event", "GdkEventKey",
    "key_release_event", "GdkEventKey",
    "enter_notify_event", "GdkEventCrossing",
    "leave_notify_event", "GdkEventCrossing",
    "configure_event", "GdkEventConfigure",
    "focus_in_event", "GdkEventFocus",
    "focus_out_event", "GdkEventFocus",
    "map_event", "GdkEvent",
    "unmap_event", "GdkEvent",
    "property_notify_event", "GdkEventProperty",
    "selection_clear_event", "GdkEventSelection",
    "selection_request_event", "GdkEventSelection",
    "selection_notify_event", "GdkEventSelection",
    "proximity_in_event", "GdkEventProximity",
    "proximity_out_event", "GdkEventProximity",
    "drag_begin_event", "GdkEventDragBegin",
    "drag_request_event", "GdkEventDragRequest",
    "drag_end_event", "GdkEventDragRequest",
    "drop_enter_event", "GdkEventDropEnter",
    "drop_leave_event", "GdkEventDropLeave",
    "drop_data_available_event", "GdkEventDropDataAvailable",
    "other_event", "GdkEventOther",
    "client_event", "GdkEventClient",
    "no_expose_event", "GdkEventNoExpose",
    NULL
  };

  gint i;

  for (i = 0; GbGDKEvents[i]; i += 2)
    {
      if (!strcmp (signal_name, GbGDKEvents[i]))
	return GbGDKEvents[i + 1];
    }
  return "GdkEvent";
}


/* This returns argument names to use for some known GTK signals. */
/* Note: Could possibly return the types to use as well, since several of
   them could be better, e.g. "GtkOrientation" instead of "gint". */
static gchar **
lookup_signal_arg_names (gchar * type, gchar * signal_name)
{
  /* Each arg array starts with the object type name and the signal name,
     and then signal arguments follow. Note that the spacing is hardcoded. */
  static gchar *GbArgTable[][8] =
  {
    {"GtkCList", "select_row",
     "gint             row",
     "gint             column",
     "GdkEvent        *event"},
    {"GtkCList", "unselect_row",
     "gint             row",
     "gint             column",
     "GdkEvent        *event"},
    {"GtkCList", "click_column",
     "gint             column"},

    {"GtkCList", "resize_column",
     "gint             column",
     "gint             width"},

    {"GtkCList", "extend_selection",
     "GtkScrollType    scroll_type",
     "gfloat           position",
     "gboolean         auto_start_selection"},
    {"GtkCList", "scroll_vertical",
     "GtkScrollType    scroll_type",
     "gfloat           position"},
    {"GtkCList", "scroll_horizontal",
     "GtkScrollType    scroll_type",
     "gfloat           position"},
    {"GtkContainer", "focus",
     "GtkDirectionType direction"},
    {"GtkCTree", "tree_select_row",
     "GList           *node",
     "gint             column"},
    {"GtkCTree", "tree_unselect_row",
     "GList           *node",
     "gint             column"},

    {"GtkCTree", "tree_expand",
     "GList           *node"},
    {"GtkCTree", "tree_collapse",
     "GList           *node"},
    {"GtkCTree", "tree_move",
     "GList           *node",
     "GList           *new_parent",
     "GList           *new_sibling"},
    {"GtkCTree", "change_focus_row_expansion",
     "GtkCTreeExpansionType expansion"},

    {"GtkEditable", "insert_text",
     "gchar           *new_text",
     "gint             new_text_length",
     "gint            *position"},
    {"GtkEditable", "delete_text",
     "gint             start_pos",
     "gint             end_pos"},
    {"GtkEditable", "set_editable",
     "gboolean         is_editable"},
    {"GtkEditable", "move_cursor",
     "gint             x",
     "gint             y"},
    {"GtkEditable", "move_word",
     "gint             num_words"},
    {"GtkEditable", "move_page",
     "gint             x",
     "gint             y"},
    {"GtkEditable", "move_to_row",
     "gint             row"},
    {"GtkEditable", "move_to_column",
     "gint             column"},

    {"GtkEditable", "kill_char",
     "gint             direction"},
    {"GtkEditable", "kill_word",
     "gint             direction"},
    {"GtkEditable", "kill_line",
     "gint             direction"},


    {"GtkInputDialog", "enable_device",
     "gint             deviceid"},
    {"GtkInputDialog", "disable_device",
     "gint             deviceid"},

    {"GtkListItem", "extend_selection",
     "GtkScrollType    scroll_type",
     "gfloat           position",
     "gboolean         auto_start_selection"},
    {"GtkListItem", "scroll_vertical",
     "GtkScrollType    scroll_type",
     "gfloat           position"},
    {"GtkListItem", "scroll_horizontal",
     "GtkScrollType    scroll_type",
     "gfloat           position"},

    {"GtkMenuShell", "move_current",
     "GtkMenuDirectionType direction"},
    {"GtkMenuShell", "activate_current",
     "gboolean         force_hide"},


    {"GtkNotebook", "switch_page",
     "GtkNotebookPage *page",
     "gint             page_num"},
    {"GtkStatusbar", "text_pushed",
     "guint            context_id",
     "gchar           *text"},
    {"GtkStatusbar", "text_popped",
     "guint            context_id",
     "gchar           *text"},
    {"GtkTipsQuery", "widget_entered",
     "GtkWidget       *widget",
     "gchar           *tip_text",
     "gchar           *tip_private"},
    {"GtkTipsQuery", "widget_selected",
     "GtkWidget       *widget",
     "gchar           *tip_text",
     "gchar           *tip_private",
     "GdkEventButton  *event"},
    {"GtkToolbar", "orientation_changed",
     "GtkOrientation   orientation"},
    {"GtkToolbar", "style_changed",
     "GtkToolbarStyle  style"},
    {"GtkWidget", "draw",
     "GdkRectangle    *area"},
    {"GtkWidget", "size_request",
     "GtkRequisition  *requisition"},
    {"GtkWidget", "size_allocate",
     "GtkAllocation   *allocation"},
    {"GtkWidget", "state_changed",
     "GtkStateType     state"},
    {"GtkWidget", "style_set",
     "GtkStyle        *previous_style"},

    {"GtkWidget", "install_accelerator",
     "gchar           *signal_name",
     "gchar            key",
     "gint             modifiers"},

    {"GtkWidget", "add_accelerator",
     "guint            accel_signal_id",
     "GtkAccelGroup   *accel_group",
     "guint            accel_key",
     "GdkModifierType  accel_mods",
     "GtkAccelFlags    accel_flags"},

    {"GtkWidget", "parent_set",
     "GtkObject       *old_parent"},

    /* This is different in GTK 1.0.x and 1.1.x */
#ifdef GTK_HAVE_FEATURES_1_1_0
    {"GtkWidget", "remove_accelerator",
     "GtkAccelGroup   *accel_group",
     "guint            accel_key",
     "GdkModifierType  accel_mods"},
#else
    {"GtkWidget", "remove_accelerator",
     "gchar           *signal_name"},
#endif

    {"GtkWidget", "debug_msg",
     "gchar           *message"},
    {"GtkWindow", "move_resize",
     "gint            *x",
     "gint            *y",
     "gint             width",
     "gint             height"},
    {"GtkWindow", "set_focus",
     "GtkWidget       *widget"},

    {"GtkWidget", "selection_get",
     "GtkSelectionData *data",
     "guint            info",
     "guint            time"},
    {"GtkWidget", "selection_received",
     "GtkSelectionData *data",
     "guint            time"},

    {"GtkWidget", "drag_begin",
     "GdkDragContext  *drag_context"},
    {"GtkWidget", "drag_end",
     "GdkDragContext  *drag_context"},
    {"GtkWidget", "drag_data_delete",
     "GdkDragContext  *drag_context"},
    {"GtkWidget", "drag_leave",
     "GdkDragContext  *drag_context",
     "guint            time"},
    {"GtkWidget", "drag_motion",
     "GdkDragContext  *drag_context",
     "gint             x",
     "gint             y",
     "guint            time"},
    {"GtkWidget", "drag_drop",
     "GdkDragContext  *drag_context",
     "gint             x",
     "gint             y",
     "guint            time"},
    {"GtkWidget", "drag_data_get",
     "GdkDragContext  *drag_context",
     "GtkSelectionData *data",
     "guint            info",
     "guint            time"},
    {"GtkWidget", "drag_data_received",
     "GdkDragContext  *drag_context",
     "gint             x",
     "gint             y",
     "GtkSelectionData *data",
     "guint            info",
     "guint            time"},

    {NULL}
  };

  gint i;

  for (i = 0; GbArgTable[i][0]; i++)
    {
      if (!strcmp (type, GbArgTable[i][0])
	  && !strcmp (signal_name, GbArgTable[i][1]))
	return &GbArgTable[i][2];
    }
  return NULL;
}


void
gb_widget_write_accelerators_source (GtkWidget * widget,
				     GbWidgetWriteSourceData * data)
{
  GList *item = data->widget_data->accelerators;
  GbAccelerator *accel;

  while (item)
    {
      accel = (GbAccelerator *) item->data;
      item = item->next;

      /* Don't add accelerators to toplevel menuitems in menu bars. */
      /* The problem turned out to be somewhere else, so we may not need this.
      if (GTK_IS_MENU_ITEM (widget) && widget->parent
	  && GTK_IS_MENU_BAR (widget->parent))
	continue;
      */

      /* If this is the first accelerator, create the accel group/table. */
      if (data->need_accel_group == FALSE)
	{
	  data->need_accel_group = TRUE;

#ifdef GTK_HAVE_FEATURES_1_1_0
	  source_add (data, "  accel_group = gtk_accel_group_new ();\n");
	  if (GTK_IS_MENU (data->component))
	    source_add (data, 
			"  gtk_menu_set_accel_group (GTK_MENU (%s), accel_group);\n",
			data->component_name);
	  else
	    source_add (data, 
			"  gtk_window_add_accel_group (GTK_WINDOW (%s), accel_group);\n",
			data->component_name);
#else
	  source_add (data,
		      "  accelerator_table = gtk_accelerator_table_new ();\n");
	  if (GTK_IS_MENU (data->component))
	    source_add (data,
			"  gtk_menu_set_accelerator_table (GTK_MENU (%s), accelerator_table);\n",
			data->component_name);
	  else
	    source_add (data,
			"  gtk_window_add_accelerator_table (GTK_WINDOW (%s), accelerator_table);\n",
			data->component_name);
#endif
	}

#ifdef GTK_HAVE_FEATURES_1_1_0
      source_add (data,
		  "  gtk_widget_add_accelerator (%s, \"%s\", accel_group,\n"
		  "                              GDK_%s, %s, GTK_ACCEL_VISIBLE);\n",
		  data->wname, accel->signal, accel->key,
		  create_modifiers_string (accel->modifiers));
#else
      source_add (data,
		  "  gtk_widget_install_accelerator (%s, accelerator_table, \"%s\",\n"
		  "                                  GDK_%s, %s);\n",
		  data->wname, accel->signal, accel->key,
		  create_modifiers_string (accel->modifiers));
#endif


    }
}



/*************************************************************************
 * Functions for replacing widgets in the user interface
 *************************************************************************/

/*
 * Replace a child with a new child. Used to replace placeholders with
 * a widget when adding widgets, and also to replace widgets with
 * placeholders when deleting.
 * NOTE: gbwidgets do not currently have their own function for this,
 * as there are only a few special cases (I hope!)
 */
void
gb_widget_replace_child (GtkWidget * widget,
			 GtkWidget * current_child,
			 GtkWidget * new_child)
{
  gchar *child_name, *new_child_name;

  /* Stop resizes until child is replaced */
#ifndef GTK_HAVE_FEATURES_1_1_4
  gtk_container_block_resize (GTK_CONTAINER (widget));
#endif

  /* We move the child name to the new widget, except for toolbar buttons,
     in which case we leave the new child's child_name as it is. */
  child_name = gtk_object_get_data (GTK_OBJECT (current_child),
				    GB_CHILD_NAME_KEY);
  new_child_name = gtk_object_get_data (GTK_OBJECT (new_child),
					GB_CHILD_NAME_KEY);
  if (new_child_name == NULL || strcmp (new_child_name, "Toolbar:button"))
    {
      g_free (new_child_name);
      if (child_name && strcmp (child_name, "Toolbar:button"))
	gtk_object_set_data (GTK_OBJECT (new_child), GB_CHILD_NAME_KEY,
			     g_strdup (child_name));
      else
	gtk_object_set_data (GTK_OBJECT (new_child), GB_CHILD_NAME_KEY, NULL);
    }

  if (GTK_IS_BIN (widget))
    {
      /* For a bin, we just delete the existing child and add the new one. */
      gtk_container_remove (GTK_CONTAINER (widget), current_child);
      gtk_container_add (GTK_CONTAINER (widget), new_child);

    }
#ifdef GTK_HAVE_FEATURES_1_1_4
  else if (GTK_IS_SCROLLED_WINDOW (widget))
    {
      /* For a scrolled window in GTK 1.1.5, we just delete the existing child
	 and add the new one, just like a GtkBin (which a GtkScrolledWindow is
	 now). */
      gtk_container_remove (GTK_CONTAINER (widget), current_child);
      gtk_container_add (GTK_CONTAINER (widget), new_child);
    }
#endif
  else if (GTK_IS_BOX (widget))
    {
      /* For a box, we find out the position of the current child, delete it,
         add the new one, and move it into position with reorder_child().
	 If the existing child is a placeholder and the new one is a menubar
	 or toolbar, we set the packing so it doesn't expand, as that is
	 probably what the user wants. */
#ifdef GTK_HAVE_FEATURES_1_1_0
      gboolean expand, fill;
      guint padding;
#else
      gint expand, fill, padding;
#endif
      GtkPackType pack_type;
      gint pos = glade_util_get_box_pos (GTK_BOX (widget), current_child);
      g_return_if_fail (pos != -1);
      gtk_box_query_child_packing (GTK_BOX (widget), current_child,
				   &expand, &fill, &padding, &pack_type);
      if (GB_IS_PLACEHOLDER (current_child)
	  && (GTK_IS_MENU_BAR (new_child)
	      || GTK_IS_TOOLBAR (new_child)
	      || GTK_IS_STATUSBAR (new_child)))
	expand = FALSE;

      gtk_container_remove (GTK_CONTAINER (widget), current_child);
      gtk_container_add (GTK_CONTAINER (widget), new_child);

      gtk_box_set_child_packing (GTK_BOX (widget), new_child, expand, fill,
				 padding, pack_type);
      gtk_box_reorder_child (GTK_BOX (widget), new_child, pos);

    }
  else if (GTK_IS_TOOLBAR (widget))
    {
      /* For a toolbar, we find out the position of the current child, then
	 remove it. We call a special function to insert the new child in the
	 same position. */
      GtkToolbarChild *tbchild;
      gint pos;

      tbchild = glade_util_find_toolbar_child (widget, current_child, &pos,
					       NULL);
      g_return_if_fail (tbchild != NULL);
      gtk_container_remove (GTK_CONTAINER (widget), current_child);
      gb_widget_insert_toolbar_child (GTK_TOOLBAR (widget), new_child, pos);
    }
  else if (GTK_IS_LIST (widget))
    {
      /* For a list, we find out the position of the current child, delete it,
         and add the new one at the same position. */
      gint pos = gtk_list_child_position (GTK_LIST (widget), current_child);
      GList glist =
      {NULL, NULL, NULL};
      glist.data = current_child;
      gtk_list_remove_items (GTK_LIST (widget), &glist);
      glist.data = new_child;
      gtk_list_insert_items (GTK_LIST (widget), &glist, pos);

    }
  else if (GTK_IS_NOTEBOOK (widget))
    {
      /* For a notebook, we find out the position of the current child, delete
         it, and add the new one at the same position. If the current_child is
         a notebook tab, just replace it. */
      GtkWidget *tab_label, *notebook_page;
      gint pos;
      GtkNotebookPage *page = find_notebook_page (widget, current_child, &pos);
      g_return_if_fail (page != NULL);
      if (page->child == current_child)
	{
	  tab_label = page->tab_label;
	  gtk_widget_ref (tab_label);
	  gtk_notebook_remove_page (GTK_NOTEBOOK (widget), pos);
	  gtk_notebook_insert_page (GTK_NOTEBOOK (widget), new_child, tab_label,
				    pos);
	  gtk_notebook_set_page (GTK_NOTEBOOK (widget), pos);
	  gtk_widget_unref (tab_label);
	}
      else
	{
	  notebook_page = page->child;
	  gtk_widget_ref (notebook_page);
	  gtk_notebook_remove_page (GTK_NOTEBOOK (widget), pos);
	  gtk_notebook_insert_page (GTK_NOTEBOOK (widget), notebook_page, new_child,
				    pos);
	  gtk_notebook_set_page (GTK_NOTEBOOK (widget), pos);
	  gtk_widget_unref (notebook_page);
	}

    }
  else if (GTK_IS_PANED (widget))
    {
      /* For paned, we find out the position of the current child, delete it,
         and add the new one at the same position. */
      gint pos = (GTK_PANED (widget)->child1 == current_child) ? 1 : 2;
      gtk_container_remove (GTK_CONTAINER (widget), current_child);
      if (pos == 1)
	gtk_paned_add1 (GTK_PANED (widget), new_child);
      else
	gtk_paned_add2 (GTK_PANED (widget), new_child);

    }
  else if (GTK_IS_TABLE (widget))
    {
      /* For a table, we find out the position & size of the current child,
         delete it, and add the new one in the same place. */
      gint left, right, top, bottom;
      GtkTableChild *tchild = glade_util_find_table_child (GTK_TABLE (widget),
							   current_child);
      g_return_if_fail (tchild != NULL);
      left = tchild->left_attach;
      right = tchild->right_attach;
      top = tchild->top_attach;
      bottom = tchild->bottom_attach;
      gtk_container_remove (GTK_CONTAINER (widget), current_child);
      gtk_table_attach_defaults (GTK_TABLE (widget), new_child,
				 left, right, top, bottom);

    }
  else if (GTK_IS_TREE (widget))
    {
      /* For a tree, we find out the position of the current child, delete it,
         and add the new one at the same position. */
      gint pos = gtk_tree_child_position (GTK_TREE (widget), current_child);
      GList glist =
      {NULL, NULL, NULL};
      glist.data = current_child;
      gtk_tree_remove_items (GTK_TREE (widget), &glist);
      gtk_tree_insert (GTK_TREE (widget), new_child, pos);

    }
  else if (GTK_IS_CLIST (widget))
    {
      /* For a clist, we check if the widget is a column title and if it is
	 we replace it. */
	gint pos;
	GtkCList *clist;

	clist = GTK_CLIST(widget);

	for (pos = 0; pos < clist->columns; pos++)
	{
	    if (clist->column[pos].button == current_child)
	    {
		gtk_clist_set_column_widget(clist, pos, new_child);
 	    }
	}
    }
  else if (GTK_IS_BUTTON (widget))
    {
      /* For a button, we just delete the existing child and add the new one.*/
      gtk_container_remove (GTK_CONTAINER (widget), current_child);
      gtk_container_add (GTK_CONTAINER (widget), new_child);

    }
  else if (GTK_IS_CONTAINER (widget))
    {
      /* General code for container - has to remove all children and add back
         NOTE: this may not work for specialised containers.
         NOTE: need to ref widgets? */
      g_warning (_("replacing child of container - not implemented yet\n"));
    }

  /* Now allow a resize */
#ifndef GTK_HAVE_FEATURES_1_1_4
  gtk_container_unblock_resize (GTK_CONTAINER (widget));
#endif
}


GtkWidget *
gb_widget_new_gtk_pixmap (GtkWidget * widget, gchar * class_name)
{
  GdkColormap *colormap;
  GtkWidget * pixmap;
  GbWidget * gbwidget = g_hash_table_lookup (gb_widget_table, class_name);

  g_return_val_if_fail (gbwidget != NULL,NULL);
  g_return_val_if_fail (gbwidget->pixmap_struct != NULL,NULL);

  /* Create the pixmap if it hasn't already been.  Don't unref since we want
     it to stick around. */
  if (!(gbwidget->gdkpixmap))
    {
      colormap = gtk_widget_get_colormap (widget);
      gbwidget->gdkpixmap = gdk_pixmap_colormap_create_from_xpm_d (NULL, colormap, &(gbwidget->mask), NULL, gbwidget->pixmap_struct);
    }

  pixmap = gtk_pixmap_new (gbwidget->gdkpixmap, gbwidget->mask);
  return pixmap;
}

/*************************************************************************
 * Functions for showing/hiding tooltips of the widgets
 *************************************************************************/

gboolean
gb_widget_get_show_tooltips ()
{
  return GTK_TOOLTIPS (gb_widget_tooltips)->enabled;
}


void
gb_widget_set_show_tooltips (gboolean show)
{
  if (show)
    gtk_tooltips_enable (gb_widget_tooltips);
  else
    gtk_tooltips_disable (gb_widget_tooltips);
}


void
gb_widget_reset_tooltips ()
{
  gtk_object_destroy (GTK_OBJECT (gb_widget_tooltips));
  gb_widget_tooltips = gtk_tooltips_new ();
}



/*************************************************************************
 * Misc. Functions
 *************************************************************************/


/* This is a GTK bug workaround for combo widgets. They should manage the
   size of their GtkEntry, but they don't at present, so we have to set its
   size explicitly (and also in the source code output). */
void
gb_widget_set_usize (GtkWidget *widget,
		     gint w,
		     gint h)
{
  if (GTK_IS_COMBO (widget))
    gtk_widget_set_usize (GTK_COMBO (widget)->entry,
			  w - 16 < 0 ? -1 : w - 16, h);
  gtk_widget_set_usize (widget, w, h);
}

/* This function is used to iterate through the table children in reverse.
   It is needed so we output the XML file in the same order each time. */
static void
table_foreach (GtkTable * table,
	       GtkCallback callback,
	       gpointer callback_data)
{
  GList *children;
  GtkTableChild *child;

  children = g_list_last (table->children);
  while (children)
    {
      child = children->data;
      children = children->prev;

      (*callback) (child->widget, callback_data);
    }
}


/* This iterates over a notebook's tabs. */
static void
notebook_tabs_foreach (GtkNotebook *notebook,
		       GtkCallback callback,
		       gpointer callback_data)
{
  GList *children;
  GtkNotebookPage *page;

  children = notebook->children;
  while (children)
    {
      page = children->data;
      (*callback) (page->tab_label, callback_data);
      children = children->next;
    }
}


static GtkNotebookPage *
find_notebook_page (GtkWidget * widget, GtkWidget * current_child, gint * pos)
{
  GList *children;
  GtkNotebookPage *page;
  gint page_num = 0;

  children = GTK_NOTEBOOK (widget)->children;
  while (children)
    {
      page = children->data;
      if (page->child == current_child || page->tab_label == current_child)
	{
	  *pos = page_num;
	  return page;
	}
      children = children->next;
      page_num++;
    }
  return NULL;
}

/* This calls the given callback for each child of a widget. It gets round
   some of the quirks of the different versions of GTK, and descends menus
   as well. */
void
gb_widget_children_foreach (GtkWidget *widget,
			    GtkCallback callback,
			    gpointer data)
{
  /* SPECIAL CODE: for table, so we output in the reverse order. */
  if (GTK_IS_TABLE (widget))
    table_foreach (GTK_TABLE (widget), callback, data);
  else if (GTK_IS_CONTAINER (widget))
    gtk_container_foreach (GTK_CONTAINER (widget), callback, data);

  /* SPECIAL CODE: output notebook tabs. */
  if (GTK_IS_NOTEBOOK (widget))
    notebook_tabs_foreach (GTK_NOTEBOOK (widget), callback, data);

  /* SPECIAL CODE: output menuitem submenus. */
  if (GTK_IS_MENU_ITEM (widget) && GTK_MENU_ITEM (widget)->submenu)
    (*callback) (GTK_MENU_ITEM (widget)->submenu, data);

  /* SPECIAL CODE: output CList (& CTree) titles in GTK 1.1.
     The foreach method has been replaced with forall. */
#ifdef GTK_HAVE_FEATURES_1_1_0
  if (GTK_IS_CLIST (widget))
    gtk_container_forall (GTK_CONTAINER (widget), callback, data);
#endif
}


/*************************************************************************
 * Common popup menu callbacks
 *************************************************************************/
void
gb_widget_remove_label (GtkWidget * menuitem,
			GtkWidget * widget)
{
  GtkWidget *child;

  if (GTK_IS_BIN (widget))
    child = GTK_BIN (widget)->child;
  else if (GTK_IS_BUTTON (widget))
    child = GTK_BUTTON (widget)->child;
  else
    return;
  if (child && GTK_IS_LABEL (child))
    editor_delete_widget (child);
}

static void
gb_widget_add_alignment (GtkWidget * menuitem,
			 GtkWidget * widget)
{
  GtkWidget *alignment, *parent;

  parent = widget->parent;
  alignment = gb_widget_new ("GtkAlignment", parent);

  gtk_widget_ref (widget);
  gb_widget_replace_child (parent, widget, alignment);
  if (GTK_BIN (alignment)->child)
    gtk_container_remove (GTK_CONTAINER (alignment),
			  GTK_BIN (alignment)->child);
  gtk_container_add (GTK_CONTAINER (alignment), widget);
  gtk_widget_unref (widget);
#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_insert_widget_parent (alignment, widget);
#endif
}


static void
gb_widget_remove_alignment (GtkWidget * menuitem,
			    GtkWidget * widget)
{
  GtkWidget *alignment, *parent;

  alignment = widget->parent;
  g_return_if_fail (GTK_IS_ALIGNMENT (alignment));
#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_remove_widget_parent (alignment, widget);
#endif
  parent = alignment->parent;

  gtk_widget_ref (widget);
  gtk_container_remove (GTK_CONTAINER (alignment), widget);
  gb_widget_replace_child (parent, alignment, widget);
  gtk_widget_unref (widget);
}


static void
gb_widget_add_event_box (GtkWidget * menuitem,
			 GtkWidget * widget)
{
  GtkWidget *event_box, *parent;

  parent = widget->parent;
  event_box = gb_widget_new ("GtkEventBox", parent);

  gtk_widget_ref (widget);
  gb_widget_replace_child (parent, widget, event_box);
  if (GTK_BIN (event_box)->child)
    gtk_container_remove (GTK_CONTAINER (event_box),
			  GTK_BIN (event_box)->child);
  gtk_container_add (GTK_CONTAINER (event_box), widget);
  gtk_widget_unref (widget);
#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_insert_widget_parent (event_box, widget);
#endif
}

static void
gb_widget_remove_event_box (GtkWidget * menuitem,
			    GtkWidget * widget)
{
  GtkWidget *event_box, *parent;

  event_box = widget->parent;
  g_return_if_fail (GTK_IS_EVENT_BOX (event_box));
#ifdef GTK_HAVE_FEATURES_1_1_0
  tree_remove_widget_parent (event_box, widget);
#endif
  parent = event_box->parent;

  gtk_widget_ref (widget);
  gtk_container_remove (GTK_CONTAINER (event_box), widget);
  gb_widget_replace_child (parent, event_box, widget);
  gtk_widget_unref (widget);
}


/*************************************************************************
 * Common functions used by gbwidgets.
 *************************************************************************/

/* This gets the child label for buttons or menuitems or subclasses.
   This is for showing or saving. */
void
gb_widget_output_child_label (GtkWidget * widget, GbWidgetGetArgData * data,
			      gchar * Label)
{
  GtkWidget *child;
  gboolean output_label = FALSE;

  /* FIXME: always use GTK_BIN for GTK 1.1 */
  if (GTK_IS_BUTTON (widget))
    child = GTK_BUTTON (widget)->child;
  else
    child = GTK_BIN (widget)->child;

  /* Note that we don't want to save the child label if it is a GbWidget,
     since it will be saved as a separate widget. */
  if (child && GTK_IS_LABEL (child) && !GB_IS_GB_WIDGET (child))
    output_label = TRUE;
  else if (gb_widget_is_toolbar_button (widget))
    {
      if (!gb_widget_get_toolbar_button_widgets (widget, NULL, &child))
	return;
      output_label = TRUE;
    }

  if (output_label)
    {
      gchar *label_text;
      gtk_label_get (GTK_LABEL (child), &label_text);
      gb_widget_output_text (data, Label, label_text);
      if (data->action == GB_SHOWING)
	property_set_sensitive (Label, TRUE);
    }
  else
    {
      if (data->action == GB_SHOWING)
	{
	  gb_widget_output_text (data, Label, "");
	  property_set_sensitive (Label, FALSE);
	}
    }
}


/* This sets the child label for buttons/items/menuitems or subclasses.
   This is for applying or loading. */
void
gb_widget_input_child_label (GtkWidget * widget, GbWidgetSetArgData * data,
			     gchar * Label)
{
  GtkWidget *child, *label;
  gchar *label_text = gb_widget_input_text (data, Label);

  MSG ("IN input_child_label");

  if (GTK_IS_BUTTON (widget))
    child = GTK_BUTTON (widget)->child;
  else
    child = GTK_BIN (widget)->child;

  if (data->apply)
    {
      if (child && GTK_IS_LABEL (child))
	{
	  gtk_label_set (GTK_LABEL (child), label_text);
	}
      else if (gb_widget_is_toolbar_button (widget))
	{
	  if (gb_widget_get_toolbar_button_widgets (widget, NULL, &child))
	    gtk_label_set (GTK_LABEL (child), label_text);
	}
      else
	{
	  if (child != NULL)
	    gtk_container_remove (GTK_CONTAINER (widget), child);
#ifdef GTK_HAVE_FEATURES_1_1_0
	  if (GTK_IS_MENU_ITEM (widget))
	    label = gtk_accel_label_new (label_text);
	  else
	    label = gtk_label_new (label_text);
#else
	    label = gtk_label_new (label_text);
#endif
	  gtk_widget_show (label);
	  gtk_container_add (GTK_CONTAINER (widget), label);

	  /* Simple child labels are given different alignments according to
	     the parent. GtkButton and GtkToggleButton are centred. All the
	     others are aligned left. See the GTK _new_with_label() fns. */
	  if (data->action == GB_LOADING)
	    {
	      if (GTK_IS_CHECK_BUTTON (widget)
		  || GTK_IS_ITEM (widget))
		gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	    }
	}
    }
  /* This isn't very nice. When a text property is got from the property
     editor (i.e. when action is GB_APPLYING) it needs to be freed after. */
  if (data->action == GB_APPLYING)
    g_free (label_text);
}


/* This gets the child icon filename for toolbar buttons.
   This is for showing or saving. If the widget is not actually a toolbar
   button and we are showing the properties, then we clear the property and
   make it insensitive. */
void
gb_widget_output_child_icon (GtkWidget * widget,
			     GbWidgetGetArgData * data,
			     gchar * property_name)
{
  if (gb_widget_is_toolbar_button (widget))
    {
      gchar *filename = gtk_object_get_data (GTK_OBJECT (widget),
					     GbToolbarIconFilenameKey);
      gb_widget_output_pixmap_filename (data, property_name, filename);
      if (data->action == GB_SHOWING)
	property_set_sensitive (property_name, TRUE);
    }
  else
    {
      if (data->action == GB_SHOWING)
	{
	  gb_widget_output_pixmap_filename (data, property_name, "");
	  property_set_sensitive (property_name, FALSE);
	}
    }
}


/* This sets the child icon filename for toolbar buttons.
   This is for applying or loading. */
void
gb_widget_input_child_icon (GtkWidget * widget,
			    GbWidgetSetArgData * data,
			    gchar * property_name)
{
  gchar *filename, *old_filename;
  GtkWidget *icon;
  GdkColormap *colormap;
  GdkPixmap *gdkpixmap;
  GdkBitmap *mask;

  if (!gb_widget_is_toolbar_button (widget))
    return;
  if (!gb_widget_get_toolbar_button_widgets (widget, &icon, NULL))
    return;

  filename = gb_widget_input_pixmap_filename (data, property_name);
  if (data->apply)
    {
      if (filename && filename[0] == '\0')
	filename = NULL;

      old_filename = gtk_object_get_data (GTK_OBJECT (widget),
					  GbToolbarIconFilenameKey);
      if (old_filename)
	{
	  glade_project_remove_pixmap (data->project, old_filename);
	  g_free (old_filename);
	}

      gtk_object_set_data (GTK_OBJECT (widget), GbToolbarIconFilenameKey,
			   g_strdup (filename));
      if (filename)
	{
	  glade_project_add_pixmap (data->project, filename);
	  colormap = gtk_widget_get_colormap (widget);
	  gdkpixmap = gdk_pixmap_colormap_create_from_xpm (NULL, colormap,
							   &mask, NULL,
							   filename);
	  if (!gdkpixmap)
	    {
	      glade_util_show_message_box (_("Couldn't create pixmap from file\n"));
	    }
	  else
	    {
	      gtk_pixmap_set (GTK_PIXMAP (icon), gdkpixmap, mask);
	      gdk_pixmap_unref (gdkpixmap);
	      gdk_bitmap_unref (mask);
	      editor_refresh_widget (icon);
	    }
	}
      else
	{
	  GbWidget * pixmap_gbwidget = g_hash_table_lookup (gb_widget_table,
							    "GtkPixmap");
	  g_return_if_fail (pixmap_gbwidget != NULL);

	  gtk_pixmap_set (GTK_PIXMAP (icon), pixmap_gbwidget->gdkpixmap,
			  pixmap_gbwidget->mask);
	  editor_refresh_widget (icon);
	}
    }
  if (data->action == GB_LOADING)
    g_free (filename);
}


void
gb_widget_destroy_child_icon (GtkWidget * widget, GbWidgetDestroyData * data)
{
  gchar *filename;

  if (gb_widget_is_toolbar_button (widget))
    {
      filename = gtk_object_get_data (GTK_OBJECT (widget),
				      GbToolbarIconFilenameKey);
      if (filename)
	{
	  glade_project_remove_pixmap (data->project, filename);
	  g_free (filename);
	}
    }
}


void
gb_widget_create_child_label_popup_menu (GtkWidget	       *widget,
					 GbWidgetCreateMenuData *data)
{
  GtkWidget *menuitem;

  if (GTK_IS_LABEL (GTK_BUTTON (widget)->child))
    {
      menuitem = gtk_menu_item_new_with_label (_("Remove Label"));
      gtk_widget_show (menuitem);
      gtk_menu_append (GTK_MENU (data->menu), menuitem);
      gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
			  GTK_SIGNAL_FUNC (gb_widget_remove_label), widget);
    }
}


void
gb_widget_create_toolbar_button_popup_menu (GtkWidget	           *widget,
					    GbWidgetCreateMenuData *data)
{
  GtkWidget *menuitem;

  if (gb_widget_is_toolbar_button (widget))
    {
      menuitem = gtk_menu_item_new_with_label (_("Convert to ordinary button"));
      gtk_widget_show (menuitem);
      gtk_menu_append (GTK_MENU (data->menu), menuitem);
      gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
			  GTK_SIGNAL_FUNC (gb_widget_convert_toolbar_button),
			  widget);
    }
}


static void
gb_widget_convert_toolbar_button (GtkWidget * menuitem,
				  GtkWidget * widget)
{
  GtkWidget *button;
  GtkToolbarChild *child;

  g_return_if_fail (widget->parent != NULL);
  child = glade_util_find_toolbar_child (widget->parent, widget, NULL, NULL);
  g_return_if_fail (child != NULL);

  /* We set the parent to NULL when calling gb_widget_new, otherwise it
     creates a new toolbar button! */
  if (child->type == GTK_TOOLBAR_CHILD_BUTTON)
    button = gb_widget_new ("GtkButton", NULL);
  else if (child->type == GTK_TOOLBAR_CHILD_TOGGLEBUTTON)
    button = gb_widget_new ("GtkToggleButton", NULL);
  else if (child->type == GTK_TOOLBAR_CHILD_RADIOBUTTON)
    button = gb_widget_new ("GtkRadioButton", NULL);
  else
    {
      g_warning ("Invalid toolbar button");
      return;
    }

  /* Set properties widget to NULL, in case the widget or parent is replaced */
  property_set_widget (NULL);

  gb_widget_replace_child (widget->parent, widget, button);
}


/* This updates the boxes size to the given value, adding placeholders or
   deleting widgets as necessary. */
void
gb_widget_update_box_size (GtkWidget * widget, gint size)
{
  GtkWidget *new_child;
  gint current_size = g_list_length (GTK_BOX (widget)->children);
  gint i;

  if (current_size < size)
    {
      /* This avoids any problems with redrawing the selection. */
      editor_clear_selection (NULL);

      for (i = 0; i < size - current_size; i++)
	{
	  if (GTK_IS_BUTTON_BOX (widget))
	    {
	      new_child = gb_widget_new ("GtkButton", widget);
	    }
	  else
	    {
	      new_child = editor_new_placeholder ();
	      /*if (GTK_IS_HBOX(widget)) gtk_widget_set_usize(new_child, 60, 80);
	         else gtk_widget_set_usize(new_child, 80, 60); */
	    }
	  gtk_box_pack_start (GTK_BOX (widget), new_child, TRUE, TRUE, 0);
	}
    }
  else if (current_size > size)
    {
      gint data[2];
      data[0] = 0;
      data[1] = size;
      gtk_container_foreach (GTK_CONTAINER (widget),
			     (GtkCallback) update_box_size_callback, data);
    }
}


/* This removes widgets from the end of the box. data[0] is the current index,
   data[1] is the desired size of the box. */
static void
update_box_size_callback (GtkWidget * widget, gint data[2])
{
  if (data[0] >= data[1])
    gtk_container_remove (GTK_CONTAINER (widget->parent), widget);
  data[0]++;
}


/*************************************************************************
 * Toolbar helper functions.
 *************************************************************************/

GtkWidget *
gb_widget_new_toolbar_button (GbWidgetNewData *data,
			      GtkToolbarChildType type)
{
  GtkToolbar *toolbar = GTK_TOOLBAR (data->parent);
  GtkWidget *new_widget, *vbox, *icon, *label;

  /* If we are loading, we check that a toolbar button is needed. */
  MSG ("In gb_widget_new_toolbar_button");
  if (data->action == GB_LOADING)
    {
      gchar *child_name = load_get_value (data->loading_data, "child_name");
      if (!child_name || strcmp (child_name, "Toolbar:button"))
	return NULL;
    }

  if (type == GTK_TOOLBAR_CHILD_BUTTON)
    {
      new_widget = gtk_button_new ();
    }
  else if (type == GTK_TOOLBAR_CHILD_TOGGLEBUTTON)
    {
      new_widget = gtk_toggle_button_new ();
      gtk_toggle_button_set_mode (GTK_TOGGLE_BUTTON (new_widget), FALSE);
    }
  else
    {
      new_widget = gtk_radio_button_new (NULL);
      gtk_toggle_button_set_mode (GTK_TOGGLE_BUTTON (new_widget), FALSE);
    }

#ifdef GTK_HAVE_FEATURES_1_1_0
  gtk_button_set_relief (GTK_BUTTON (new_widget), toolbar->relief);
#endif

  GTK_WIDGET_UNSET_FLAGS (new_widget, GTK_CAN_FOCUS);
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (new_widget), vbox);
  gtk_widget_show (vbox);

  label = gtk_label_new (data->name);
  gtk_box_pack_end (GTK_BOX (vbox), label, FALSE, FALSE, 0);
  if (toolbar->style != GTK_TOOLBAR_ICONS)
    gtk_widget_show (label);

  icon = gb_widget_new_gtk_pixmap (data->parent->parent, "GtkPixmap");
  gtk_box_pack_end (GTK_BOX (vbox), icon, FALSE, FALSE, 0);
  if (toolbar->style != GTK_TOOLBAR_TEXT)
    gtk_widget_show (icon);

  /* Remember it is a special widget. */
  gtk_object_set_data (GTK_OBJECT (new_widget), GB_CHILD_NAME_KEY,
		       g_strdup ("Toolbar:button"));

  return new_widget;
}


/* SPECIAL CODE: This inserts a proper toolbar button (with an icon and label)
   into a toolbar at the given position. We need it since GtkToolbar creates
   and inserts the widget all in one go, which isn't good enough for us,
   as we want to set the events mask to add signal handlers to the buttons.
   This also sets the radio button group, if it is a radio button. */
void
gb_widget_insert_toolbar_child (GtkToolbar *toolbar,
				GtkWidget *new_child,
				gint position)
{
  GtkToolbarChild *child;
  GtkWidget *icon, *label;
  GtkToolbarChildType type;

  /* Determine if the new child is a proper toolbar button, i.e. it has
     a vbox with a pixmap and a label in it. If it isn't we just add it as
     a standard widget. */
  if (!gb_widget_is_toolbar_button (new_child))
    {
      gtk_toolbar_insert_widget (toolbar, new_child, NULL, NULL, position);
      return;
    }

  if (!gb_widget_get_toolbar_button_widgets (new_child, &icon, &label))
    return;
  if (GTK_IS_RADIO_BUTTON (new_child))
    type = GTK_TOOLBAR_CHILD_RADIOBUTTON;
  else if (GTK_IS_TOGGLE_BUTTON (new_child))
    type = GTK_TOOLBAR_CHILD_TOGGLEBUTTON;
  else
    type = GTK_TOOLBAR_CHILD_BUTTON;

  child = g_new (GtkToolbarChild, 1);
  child->widget = new_child;
  child->icon = icon;
  child->label = label;
  child->type = type;

  /* This is mainly copied from gtktoolbar.c */
  toolbar->children = g_list_insert (toolbar->children, child, position);
  toolbar->num_children++;

  if (type != GTK_TOOLBAR_CHILD_SPACE)
    gtk_widget_set_parent (child->widget, GTK_WIDGET (toolbar));

  if ((type != GTK_TOOLBAR_CHILD_SPACE) && GTK_WIDGET_VISIBLE (toolbar))
    {
      if (GTK_WIDGET_REALIZED (toolbar)
	  && !GTK_WIDGET_REALIZED (child->widget))
	gtk_widget_realize (child->widget);
	
      if (GTK_WIDGET_MAPPED (toolbar)
	  && !GTK_WIDGET_MAPPED (child->widget))
	gtk_widget_map (child->widget);
    }

  if (GTK_WIDGET_VISIBLE (toolbar) &&
      ((type == GTK_TOOLBAR_CHILD_SPACE) ||
       GTK_WIDGET_VISIBLE (child->widget)))
    gtk_widget_queue_resize (GTK_WIDGET (toolbar));

  if (type == GTK_TOOLBAR_CHILD_RADIOBUTTON)
    gb_widget_check_toolbar_radio_group (new_child);
}


/* This makes sure that if a toolbar widget is a radio button it is in the
   correct group. */
void
gb_widget_check_toolbar_radio_group (GtkWidget * widget)
{
  GtkWidget *group_widget;
  GSList *current_group, *new_group;

  if (!GTK_IS_RADIO_BUTTON (widget))
    return;

  MSG1 ("Checking radio group of %s", gtk_widget_get_name (widget));
  /* We don't want the current widget returned, since its group may not have
     been set yet, as it is being added to its parent now. */
  group_widget = gb_widget_get_toolbar_radio_group_widget (widget, TRUE,
							   FALSE);
  if (group_widget)
    {
      MSG1 ("  group widget: %s", gtk_widget_get_name (group_widget));
    }
  current_group = gtk_radio_button_group (GTK_RADIO_BUTTON (widget));

  if (group_widget)
    {
      new_group = gtk_radio_button_group (GTK_RADIO_BUTTON (group_widget));
      if (current_group != new_group)
	{
	  if (current_group->data == widget)
	    current_group = current_group->next;
	  gtk_radio_button_set_group (GTK_RADIO_BUTTON (widget), new_group);
	  gb_widget_update_radio_group (current_group);
	  gb_widget_update_radio_group (gtk_radio_button_group (GTK_RADIO_BUTTON (widget)));
	}
    }
  else if (g_slist_length (current_group) != 1)
    {
      current_group = gb_widget_reset_radio_group (widget);
      gb_widget_update_radio_group (current_group);
      gb_widget_update_radio_group (gtk_radio_button_group (GTK_RADIO_BUTTON (widget)));
    }
}


/* This resets the radio button's group so it is in a group of its own.
   Unfortunately the radio button widget hasn't got a function to do this. */
GSList*
gb_widget_reset_radio_group (GtkWidget * widget)
{
  GtkRadioButton *radio_button = GTK_RADIO_BUTTON (widget);
  GSList *old_group, *slist;

  old_group = g_slist_remove (radio_button->group, radio_button);
  /* Update the pointer to the group in all the other radio buttons. */
  for (slist = old_group; slist; slist = slist->next)
    GTK_RADIO_BUTTON (slist->data)->group = old_group;
  radio_button->group = g_slist_prepend (NULL, radio_button);
  return old_group;
}


/* This tries to ensure that one and only one radio button in the given
   group is selected. It's used because GtkRadioButton doesn't handle this
   when groups are changed. */
void
gb_widget_update_radio_group (GSList * group)
{
  GtkRadioButton *button;
  GSList *slist;
  gboolean found_selected = FALSE;

  if (group == NULL)
    return;

  for (slist = group; slist; slist = slist->next)
    {
      button = GTK_RADIO_BUTTON (slist->data);
      if (GTK_TOGGLE_BUTTON (button)->active)
	{
	  if (found_selected)
	    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (button), FALSE);
	  else
	    found_selected = TRUE;
	}
    }
  /* If none are currently selected, select the first. */
  if (!found_selected)
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (group->data), TRUE);
}



/* This returns TRUE if the widget is a special toolbar button, i.e. it
   has an icon and label. */
gboolean
gb_widget_is_toolbar_button (GtkWidget *widget)
{
  gchar *child_name = gtk_object_get_data (GTK_OBJECT (widget),
					   GB_CHILD_NAME_KEY);
  return (child_name && !strcmp (child_name, "Toolbar:button"));
}


/* This finds the icon and/or label widgets within a toolbar button.
   If icon_return and label_return are non-NULL the pointers are placed there.
   it returns TRUE if it finds them OK. */
gboolean
gb_widget_get_toolbar_button_widgets (GtkWidget *widget,
				      GtkWidget **icon_return,
				      GtkWidget **label_return)
{
  GtkWidget *child, *label, *icon;
  GList *vbox_children;

  g_return_val_if_fail (GTK_IS_BUTTON (widget), FALSE);

#ifdef GTK_HAVE_FEATURES_1_1_0
  child = GTK_BIN (widget)->child;
#else
  child = GTK_BUTTON (widget)->child;
#endif

  g_return_val_if_fail (GTK_IS_VBOX (child), FALSE);

  vbox_children = GTK_BOX (child)->children;
  g_return_val_if_fail (g_list_length (vbox_children) == 2, FALSE);

  /* Note that the label is added first in gtktoolbar.c which is why
     it is first in the list. */
  label = ((GtkBoxChild*)vbox_children->data)->widget;
  icon = ((GtkBoxChild*)vbox_children->next->data)->widget;

  g_return_val_if_fail (GTK_IS_LABEL (label), FALSE);
  g_return_val_if_fail (!GB_IS_GB_WIDGET (label), FALSE);
  g_return_val_if_fail (GTK_IS_PIXMAP (icon), FALSE);
  g_return_val_if_fail (!GB_IS_GB_WIDGET (icon), FALSE);

  if (label_return)
    *label_return = label;
  if (icon_return)
    *icon_return = icon;
  return TRUE;
}


gboolean
gb_widget_get_new_toolbar_group	(GtkWidget	       *toolbar,
				 GtkWidget	       *widget)
{
  GtkToolbarChild *tbchild, *prev_tbchild;
  GList *elem;

  tbchild = glade_util_find_toolbar_child (toolbar, widget, NULL, &elem);
  g_return_val_if_fail (tbchild != NULL, FALSE);

  if (elem->prev == NULL)
    return FALSE;

  prev_tbchild = (GtkToolbarChild*) elem->prev->data;
  if (prev_tbchild->type == GTK_TOOLBAR_CHILD_SPACE)
    return TRUE;

  return FALSE;
}


/* This sets whether or not a toolbar child starts a new group on the toolbar,
   i.e. there is a space before it. */
void
gb_widget_set_new_toolbar_group (GtkWidget	       *toolbar,
				 GtkWidget	       *widget,
				 gboolean		new_group)
{
  GtkToolbarChild *tbchild, *prev_tbchild;
  gint pos;
  GList *elem;
  GSList *current_group;
  gboolean check_groups = FALSE;
  GtkWidget *first_radio_widget = NULL;

  tbchild = glade_util_find_toolbar_child (toolbar, widget, &pos, &elem);
  g_return_if_fail (tbchild != NULL);

  if (elem->prev == NULL)
    return;

  prev_tbchild = (GtkToolbarChild*) elem->prev->data;

  if (new_group)
    {
      if (prev_tbchild->type != GTK_TOOLBAR_CHILD_SPACE)
	{
	  gtk_toolbar_insert_space (GTK_TOOLBAR (toolbar), pos);
	  /* We try to find the first radio item in the toolbar group, which is
	     using the default (NULL) radio group. If found we reset its radio
	     group, and indicate that the rest of the toolbar radio items in
	     the group should be checked. */
	  first_radio_widget = gb_widget_get_toolbar_radio_group_widget (widget, FALSE, TRUE);
	  if (first_radio_widget)
	    {
	      current_group = gb_widget_reset_radio_group (first_radio_widget);
	      gb_widget_update_radio_group (current_group);
	      gb_widget_update_radio_group (gtk_radio_button_group (GTK_RADIO_BUTTON (first_radio_widget)));
	      check_groups = TRUE;
	    }
	}
    }
  else
    {
      if (prev_tbchild->type == GTK_TOOLBAR_CHILD_SPACE)
	{
	  /* FIXME: This is naughty, but GtkToolbar has no way for removing
	     spaces. */
	  GTK_TOOLBAR (toolbar)->children = g_list_remove (GTK_TOOLBAR (toolbar)->children, prev_tbchild);
	  GTK_TOOLBAR (toolbar)->num_children--;
	  gtk_widget_queue_resize (toolbar);
	  check_groups = TRUE;
	}
    }

  /* If a space has been added or removed, we need to check all the radio
     groups until the next space on the toolbar. We don't check the first
     radio widget since its group has just been reset. */
  if (check_groups)
    {
      while (elem)
	{
	  tbchild = (GtkToolbarChild*) elem->data;
	  if (tbchild->type == GTK_TOOLBAR_CHILD_SPACE)
	    break;
	  if (tbchild->widget != first_radio_widget)
	    gb_widget_check_toolbar_radio_group (tbchild->widget);
	  elem = elem->next;
	}
    }
}


/* This writes the source code for creating toolbar buttons.
   Note that these are added to the toolbar here, so the standard code for
   adding them to the parent has to be skipped. */
void
gb_widget_write_toolbar_button_source (GtkWidget * widget,
				       GbWidgetWriteSourceData * data)
{
  GbWidgetData *wdata = data->widget_data;
  GtkWidget *icon, *label, *radio_widget;
  gchar *parent_name, *label_text, *type, *radio_widget_name = NULL;
  gboolean has_icon;
  gchar *icon_filename = gtk_object_get_data (GTK_OBJECT (widget),
					      GbToolbarIconFilenameKey);

  if (!gb_widget_get_toolbar_button_widgets (widget, &icon, &label))
    return;

  g_return_if_fail (widget->parent != NULL);
  parent_name = gtk_widget_get_name (widget->parent);
  parent_name = source_create_valid_identifier (parent_name);

  gtk_label_get (GTK_LABEL (label), &label_text);

  if (gb_widget_get_new_toolbar_group (widget->parent, widget))
    source_add (data, "  gtk_toolbar_append_space (GTK_TOOLBAR (%s));\n\n",
		parent_name);

  if (GTK_IS_RADIO_BUTTON (widget))
    {
      type = "GTK_TOOLBAR_CHILD_RADIOBUTTON";
      radio_widget = gb_widget_get_toolbar_radio_group_widget (widget, FALSE,
							       FALSE);
      if (radio_widget && radio_widget != widget)
	radio_widget_name = source_create_valid_identifier (gtk_widget_get_name (radio_widget));
    }
  else if (GTK_IS_TOGGLE_BUTTON (widget))
    type = "GTK_TOOLBAR_CHILD_TOGGLEBUTTON";
  else
    type = "GTK_TOOLBAR_CHILD_BUTTON";

  if (!icon_filename)
    {
      MSG ("Pixmap filename is not set");
      has_icon = FALSE;
    }
  else
    {
      has_icon = TRUE;

      /* Make sure the tmp_toolbar_icon is declared.*/
      if (!glade_util_strstr (data->decl_buffer.ptr, "tmp_toolbar_icon"))
	source_add_decl (data, "  GtkWidget *tmp_toolbar_icon;\n");

      source_create_pixmap (data, "tmp_toolbar_icon", icon_filename);
    }

  label_text = g_strdup (source_make_string (label_text, data->use_gettext));
  source_add (data,
	      "  %s = gtk_toolbar_append_element (GTK_TOOLBAR (%s),\n"
	      "                                %s,\n"
	      "                                %s,\n"
	      "                                %s,\n"
	      "                                %s, NULL,\n"
	      "                                %s, NULL, NULL);\n",
	      data->wname, parent_name, type,
	      radio_widget_name ? radio_widget_name : "NULL",
	      label_text,
	      wdata->tooltip ? source_make_string (wdata->tooltip, data->use_gettext) : "NULL",
	      has_icon ? "tmp_toolbar_icon" : "NULL");

  g_free (label_text);
  g_free (radio_widget_name);
  g_free (parent_name);
}


/* This returns the first radio button widget in a toolbar with the same
   group as the given widget, or NULL if there isn't any.
   If the given widget has a group name set then it steps through
   the entire toolbar to find the first radio button with that group.
   If the given widget has its group set to NULL (i.e. using the default group)
   then it steps through the toolbar only until it finds a toolbar space
   item. */
static GtkWidget*
gb_widget_get_toolbar_radio_group_widget (GtkWidget *widget,
					  gboolean skip_widget,
					  gboolean find_default_group)
{
  GtkToolbarChild *tbchild;
  GList *elem;
  GtkWidget *group_widget = NULL;
  gchar *group = NULL, *tmp_group;

  g_return_val_if_fail (widget->parent != NULL, NULL);
  g_return_val_if_fail (GTK_IS_TOOLBAR (widget->parent), NULL);

  if (!find_default_group)
    /* FIXME: should use Group from gbradiobutton.c */
    group = gtk_object_get_data (GTK_OBJECT (widget), "GtkRadioButton::group");

  tbchild = glade_util_find_toolbar_child (widget->parent, widget, NULL,
					   &elem);
  g_return_val_if_fail (tbchild != NULL, NULL);

  /* Find the start of the toolbar group, or the first widget on the toolbar
     if the radio button has a group set. */
  while (elem->prev)
    {
      tbchild = (GtkToolbarChild*) elem->prev->data;
      if (group == NULL && tbchild->type == GTK_TOOLBAR_CHILD_SPACE)
	break;
      elem = elem->prev;
    }

  while (elem)
    {
      tbchild = (GtkToolbarChild*) elem->data;

      /* Skip the widget we are adding, if requested. */
      if (!skip_widget || tbchild->widget != widget)
	{
	  if (tbchild->type == GTK_TOOLBAR_CHILD_RADIOBUTTON
	      || (tbchild->type == GTK_TOOLBAR_CHILD_WIDGET
		  && GTK_IS_RADIO_BUTTON (tbchild->widget)))
	    {
	      tmp_group = gtk_object_get_data (GTK_OBJECT (tbchild->widget),
					       "GtkRadioButton::group");
	      if ((group == NULL && tmp_group == NULL)
		  || (group && tmp_group && !strcmp (group, tmp_group)))
		{
		  group_widget = tbchild->widget;
		  break;
		}
	    }
	  else if (group == NULL && tbchild->type == GTK_TOOLBAR_CHILD_SPACE)
	    break;
	}
      elem = elem->next;
    }

  return group_widget;
}


/*************************************************************************
 * Adjustment convenience functions - handles all 6 values.
 *************************************************************************/
void
gb_widget_output_adjustment (GbWidgetGetArgData * data, gchar * Values[],
			     GtkAdjustment * adjustment)
{
  gb_widget_output_float (data, Values[0], adjustment->value);
  gb_widget_output_float (data, Values[1], adjustment->lower);
  gb_widget_output_float (data, Values[2], adjustment->upper);
  gb_widget_output_float (data, Values[3], adjustment->step_increment);
  gb_widget_output_float (data, Values[4], adjustment->page_increment);
  gb_widget_output_float (data, Values[5], adjustment->page_size);
}


gboolean
gb_widget_input_adjustment (GbWidgetSetArgData * data, gchar * Values[],
			    GtkAdjustment * adjustment)
{
  gfloat value, lower, upper, step_inc, page_inc, page_size;
  gboolean myApply;

  value = gb_widget_input_float (data, Values[0]);
  myApply = data->apply;
  lower = gb_widget_input_float (data, Values[1]);
  myApply |= data->apply;
  upper = gb_widget_input_float (data, Values[2]);
  myApply |= data->apply;
  step_inc = gb_widget_input_float (data, Values[3]);
  myApply |= data->apply;
  page_inc = gb_widget_input_float (data, Values[4]);
  myApply |= data->apply;
  page_size = gb_widget_input_float (data, Values[5]);
  myApply |= data->apply;

  if (myApply)
    {
      adjustment->value = value;
      adjustment->lower = lower;
      adjustment->upper = upper;
      adjustment->step_increment = step_inc;
      adjustment->page_increment = page_inc;
      adjustment->page_size = page_size;
      return TRUE;
    }
  return FALSE;
}

void gb_box_clist_auto_policy(GtkWidget *vbox, GtkWidget *clist)
{   
#ifdef GTK_HAVE_FEATURES_1_1_4
  GtkWidget *scrolled_win;

  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), clist);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start (GTK_BOX (vbox), scrolled_win, TRUE, TRUE, 0);
  gtk_widget_show(scrolled_win);
#else
  gtk_box_pack_start (GTK_BOX (vbox), clist, TRUE, TRUE, 0);
  gtk_clist_set_policy (GTK_CLIST (clist), GTK_POLICY_AUTOMATIC,
			GTK_POLICY_AUTOMATIC);
#endif
}
